(*
Copyright 2011 Christian Sternagel, René Thiemann

This file is part of IsaFoR/CeTA.

IsaFoR/CeTA is free software: you can redistribute it and/or modify it under the
terms of the GNU Lesser General Public License as published by the Free Software
Foundation, either version 3 of the License, or (at your option) any later
version.

IsaFoR/CeTA is distributed in the hope that it will be useful, but WITHOUT ANY
WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR A
PARTICULAR PURPOSE.  See the GNU Lesser General Public License for more details.

You should have received a copy of the GNU Lesser General Public License along
with IsaFoR/CeTA. If not, see <http://www.gnu.org/licenses/>.
*)
theory Argument_Filter
imports
  Term_Order_Impl
  Trs_Impl
begin

datatype af_entry = Collapse nat | AFList "nat list"
type_synonym 'f afs = "'f \<Rightarrow> nat \<Rightarrow> af_entry"
datatype 'f filtered = FPair 'f  nat
(* the main issue to not use pairs is that we can provide an alternative show function *)

fun
  filtered_fun :: "'f filtered \<Rightarrow> 'f"
where
  "filtered_fun (FPair f n) = f"

instantiation filtered :: ("show") "show" begin
  definition "shows_prec (d::nat) ff = shows (filtered_fun ff)"
  definition "shows_list_filtered (ffs::('f::show) filtered list) = shows_list_aux shows ffs"
  instance by (intro_classes, auto simp: shows_prec_filtered_def shows_list_filtered_def)
end

fun apply_af_entry :: "'f \<Rightarrow> af_entry \<Rightarrow> ('f,'v)term list \<Rightarrow> ('f,'v)term" where
  "apply_af_entry _ (Collapse i) ts = ts ! i"
| "apply_af_entry f (AFList is) ts = Fun f (map (\<lambda> i. ts ! i) is)"

fun wf_af_entry :: "nat \<Rightarrow> af_entry \<Rightarrow> bool"
where "wf_af_entry n (Collapse i) = (i < n)"
  |   "wf_af_entry n (AFList is) = list_all (\<lambda> i. i < n) is"

definition wf_afs :: "'f afs \<Rightarrow> bool"
where "wf_afs \<pi> \<equiv> \<forall> f n. wf_af_entry n (\<pi> f n)"

fun afs_term :: "'f afs \<Rightarrow> ('f,'v)term \<Rightarrow> ('f filtered,'v)term" where 
  "afs_term \<pi> (Fun f ts) = (let l = length ts in apply_af_entry (FPair f l) (\<pi> f l) (map (afs_term \<pi>) ts))"
| "afs_term \<pi> (Var x) = Var x"

definition afs_rule :: "'f afs \<Rightarrow> ('f,'v)rule \<Rightarrow> ('f filtered,'v)rule"
where "afs_rule \<pi> lr \<equiv> (afs_term \<pi> (fst lr), afs_term \<pi> (snd lr))"

definition afs_rules :: "'f afs \<Rightarrow> ('f,'v)rules \<Rightarrow> ('f filtered,'v)rules"
where "afs_rules \<pi> R = map (afs_rule \<pi>) R"

definition
  afs_check ::
    "shows \<Rightarrow> 'f::show afs \<Rightarrow> (('f filtered,'v::show)rule \<Rightarrow> shows check) \<Rightarrow> ('f,'v)rule  \<Rightarrow> shows check"
where
  "afs_check r \<pi> g lr \<equiv> let pl = afs_term \<pi> (fst lr); pr = afs_term \<pi> (snd lr) in g (pl,pr) <+? (\<lambda>s. 
  ''could not orient '' +#+ shows(fst lr) +@+ '' '' +#+ r +@+ '' '' +#+ shows(snd lr) +@+ shows_nl +@+
  ''pi( '' +#+ shows(fst lr) +@+ '' ) = '' +#+ shows pl +@+ shows_nl +@+
  ''pi( '' +#+ shows(snd lr) +@+ '' ) = '' +#+ shows pr +@+ shows_nl +@+ s
)"
 
locale afs =
  fixes \<pi> :: "'f afs"
  assumes wf_afs: "wf_afs \<pi>"
begin

abbreviation afs_subst :: "('f,'v)subst \<Rightarrow> ('f filtered,'v)subst"
where "afs_subst \<sigma> \<equiv> (\<lambda>x. afs_term \<pi> (\<sigma> x))"

definition afs_rel :: "('f filtered,'v)term ars \<Rightarrow> ('f,'v)term ars"
where "afs_rel R = {lr | lr. afs_rule \<pi> lr \<in> R}"

lemma afs_subst[simp]: "afs_term \<pi> (t \<cdot> \<sigma>) = afs_term \<pi> t \<cdot> afs_subst \<sigma>"
proof -
  have "afs_term \<pi> (t \<cdot> \<sigma>) = afs_term \<pi> t \<cdot> afs_subst (\<sigma>)"
  proof (induct t)
    case (Fun f ss)
    let ?l = "length ss"
    let ?e = "\<pi> f ?l"
    from wf_afs have wf_e: "wf_af_entry ?l ?e" unfolding wf_afs_def by auto
    show ?case
    proof (cases ?e)
      case (Collapse i)
      with wf_e have "i < ?l" by simp
      with Collapse Fun(1)[of "ss ! i"] show ?thesis
	by (auto simp: Let_def)
    next
      case (AFList "is")
      with wf_e have wf: "\<forall> i \<in> set is. i < ?l" by (auto simp: list_all_iff)
      with AFList Fun(1) show ?thesis 
	by auto
    qed
  qed simp
  thus ?thesis by simp
qed

lemma afs_subst_closed: 
  assumes "subst_closed R"
  shows "subst_closed (afs_rel R)"
unfolding subst_closed_def
proof (rule subsetI, simp only: split_paired_all)
  fix ss ts
  assume "(ss,ts) \<in> CS (afs_rel R)"
  thus "(ss,ts) \<in> afs_rel R"
  proof (induct)
    case (subst_closureI s t \<sigma>)
    from this have "(afs_term \<pi> s, afs_term \<pi> t) \<in> R" unfolding afs_rel_def afs_rule_def by simp
    with assms have "(afs_term \<pi> s \<cdot> afs_subst \<sigma>, afs_term \<pi> t \<cdot> afs_subst \<sigma>) \<in> R" unfolding subst_closed_def 
      using subst_closure.subst_closureI by auto    
    thus ?case unfolding afs_rel_def afs_rule_def by auto
  qed
qed

lemma afs_SN: 
  assumes "SN R"
  shows "SN (afs_rel R)"
proof (rule ccontr)
  assume "\<not> ?thesis"
  from this obtain f where nSN: "\<forall> i. (f i, f (Suc i)) \<in> afs_rel R" unfolding SN_defs by blast
  obtain g where "g = (\<lambda> i. afs_term \<pi> (f i))" by blast
  with nSN have "\<forall> i. (g i, g (Suc i)) \<in> R" unfolding afs_rel_def afs_rule_def by auto
  hence "\<not> (SN R)" unfolding SN_defs by auto
  with assms show False by blast
qed

lemma afs_compat: 
  assumes "NS O S \<subseteq> S"
  shows "afs_rel NS O afs_rel S \<subseteq> afs_rel S"
using assms
  by (unfold afs_rel_def afs_rule_def, clarify, auto)

lemma afs_compat2: 
  assumes "S O NS \<subseteq> S"
  shows "afs_rel S O afs_rel NS \<subseteq> afs_rel S"
using assms
  by (unfold afs_rel_def afs_rule_def, clarify, auto)

lemma afs_refl: 
  assumes "refl NS"
  shows "refl (afs_rel NS)"
using assms unfolding refl_on_def
  by (unfold afs_rel_def afs_rule_def, auto)

lemma afs_trans: 
  assumes "trans r"
  shows "trans (afs_rel r)"
unfolding trans_def
proof (unfold afs_rel_def afs_rule_def, clarify, unfold fst_conv snd_conv)
  fix a b c
  assume gt: "(afs_term \<pi> a, afs_term \<pi> b) \<in> r" "(afs_term \<pi> b, afs_term \<pi> c) \<in> r"
  show "\<exists> lr. (a,c) = lr \<and> (afs_term \<pi> (fst lr), afs_term \<pi> (snd lr)) \<in> r"
    by (intro exI conjI, rule refl, insert assms[unfolded trans_def, THEN spec, THEN spec, THEN spec, THEN mp[OF _ gt(1)], THEN mp[OF _ gt(2)]], simp)
qed

lemma afs_NS_mono: 
  assumes acc: "all_ctxt_closed R"
  and refl: "refl R"
  shows "one_step_ctxt_closed (afs_rel R)"
unfolding one_step_ctxt_closed_def
proof (intro allI impI)
  fix f bef s t aft
  assume st: "(s,t) \<in> afs_rel R"
  let ?ps = "afs_term \<pi> s"
  let ?pt = "afs_term \<pi> t"
  let ?bsa = "bef @ s # aft"
  let ?bta = "bef @ t # aft"
  from st have pst: "(?ps,?pt) \<in> R" unfolding afs_rel_def afs_rule_def by auto
  let ?l = "Suc (length bef + length aft)"
  let ?ls = "length (bef @ s # aft)"
  let ?lt = "length (bef @ t # aft)"
  let ?af = "\<pi> f ?l"
  from wf_afs have wf_af: "wf_af_entry ?l ?af" unfolding wf_afs_def by simp
  let ?msa = "map (afs_term \<pi>) ?bsa"
  let ?mta = "map (afs_term \<pi>) ?bta"
  let ?pfs = "apply_af_entry (FPair f ?l) ?af ?msa"
  let ?pft = "apply_af_entry (FPair f ?l) ?af ?mta"
  have all2: "\<forall> i < ?l. (?msa ! i, ?mta ! i) \<in> R"
  proof (intro allI impI)
    fix i
    assume "i < ?l"
    hence i_s: "i < ?ls" and i_t: "i < ?lt" by simp_all
    {
      assume bef: "i < length bef"
      with refl have "(?msa ! i, ?mta ! i) \<in> R" unfolding refl_on_def 
	by (simp only: nth_map[OF i_s], simp add: nth_append)
    }
    moreover
    {
      assume "i = length bef"
      with pst have "(?msa ! i, ?mta ! i) \<in> R" unfolding refl_on_def
	by (simp only: nth_map[OF i_s], simp add: nth_append)
    }
    moreover
    {
      assume i: "i > length bef"
      from this obtain k where k1: "i - length bef = Suc k" by (cases "i - length bef", arith+)
      with i_s have k2: "k < length aft" by auto
      from i refl have "(?msa ! i, ?mta ! i) \<in> R" unfolding refl_on_def 
	by (simp only: nth_map[OF i_s] nth_map[OF i_t], simp add: nth_append k1)
    }
    moreover
    {
      have "i < length bef \<or> i = length bef \<or> i > length bef" by auto
    }      
    ultimately
    show "(?msa ! i, ?mta ! i) \<in> R" by blast
  qed
  have "(?pfs,?pft) \<in> R"
  proof (cases ?af)
    case (Collapse i)
    with wf_af have "i < ?l"  by simp
    with all2 Collapse show ?thesis by auto
  next
    case (AFList ids)
    with wf_af have ids: "\<forall> i \<in> set ids. i < ?l" by (simp add: list_all_iff)
    show ?thesis 
    proof (simp only: AFList apply_af_entry.simps)
      let ?msi = "map (op ! (map (afs_term \<pi>) ?bsa)) ids"
      let ?mti = "map (op ! (map (afs_term \<pi>) ?bta)) ids"
      from ids all2 have "\<forall> i < length ids. (?msi ! i, ?mti ! i) \<in> R" by auto
      with acc show "(Fun (FPair f ?l) ?msi, Fun (FPair f ?l) ?mti) \<in> R"
	unfolding all_ctxt_closed_def by force
    qed
  qed
  thus "(Fun f ?bsa, Fun f ?bta) \<in> afs_rel R" unfolding afs_rel_def afs_rule_def by simp
qed
end  

locale afs_redtriple = afs \<pi> + redtriple_order S NS NST
  for \<pi> :: "('f::show)afs" and S NS NST :: "('f filtered,'v::show)trs" 

locale afs_redtriple_impl = afs_redtriple \<pi> S NS NST + redtriple_impl S NS NST cS cNS cNST
  for \<pi> :: "('f::show)afs" and S NS NST :: "('f filtered,'v::show)trs" and cS cNS cNST


locale mono_afs_redtriple_impl = afs_redtriple_impl \<pi> S NS NST cS cNS cNST + mono_redtriple_impl S NS NST cS cNS cNST cm
  for \<pi> :: "('f::show)afs" and S NS NST :: "('f filtered,'v::show)trs" and cS cNS cNST cm

sublocale afs_redtriple \<subseteq> redtriple "afs_rel S" "afs_rel NS" "afs_rel NST"
proof -
  let ?S = "afs_rel S"
  let ?NS = "afs_rel NS"
  let ?NST = "afs_rel NST"
  show "redtriple ?S ?NS ?NST"
  proof
    show "SN ?S" 
      by (rule afs_SN, rule SN)
  next
    show "?NS O ?S \<subseteq> ?S"
      by (rule afs_compat, rule compat_NS_S)
  next
    show "?S O ?NS \<subseteq> ?S"
      by (rule afs_compat2, rule compat_S_NS)
  next
    show "subst_closed ?S"
      by (rule afs_subst_closed, rule subst_S)
  next
    show "subst_closed ?NS"
      by (rule afs_subst_closed, rule subst_NS)
  next
    show "subst_closed ?NST"
      by (rule afs_subst_closed, rule subst_NST)
  next
    show "one_step_ctxt_closed ?NS"
      by (rule afs_NS_mono, rule all_ctxt_closed, rule refl_NS)
  next
    show "?NST O ?S \<subseteq> ?S"
      by (rule afs_compat[OF compat_NST])
  qed
qed

fun mono_af_entry :: "nat \<Rightarrow> af_entry \<Rightarrow> bool"
where "mono_af_entry n (Collapse i) = (n \<le> 1)"
  | "mono_af_entry n (AFList ids) = list_all (\<lambda> i. i \<in> set ids) [0 ..< n]"

definition mono_afs :: "'f afs \<Rightarrow> bool"
where "mono_afs \<pi> \<equiv> \<forall> f n. mono_af_entry n (\<pi> f n)"

context afs_redtriple
begin

lemma one_step_ctxt_closed: assumes mono_pi: "mono_afs \<pi>"
  and monoS: "one_step_ctxt_closed S"
  shows "one_step_ctxt_closed (afs_rel S)"
proof -
  let ?S = "afs_rel S"
  let ?pi = "afs_term \<pi>"
  {
    fix f bef s t aft
    assume st: "(?pi s, ?pi t) \<in> S"
    let ?bsa = "bef @ s # aft"
    let ?bta = "bef @ t # aft"
    let ?fs = "Fun f ?bsa"
    let ?ft = "Fun f ?bta"
    let ?l = "Suc (length bef + length aft)"
    let ?lba = "length bef + length aft"
    let ?e = "\<pi> f ?l"
    from mono_pi have mono_e: "mono_af_entry ?l ?e" unfolding mono_afs_def by auto
    from wf_afs have wf_e: "wf_af_entry ?l ?e" unfolding wf_afs_def by auto
    have "(?pi ?fs, ?pi ?ft) \<in> S"
    proof (cases ?e)
      case (Collapse i)
      with mono_e wf_e show ?thesis using st by auto
    next
      case (AFList ids)
      thus ?thesis 
      proof simp
	let ?ids = "{n | n. n < ?l}"
	let ?f = "FPair f ?l"
	obtain l where l: "?l = l" by auto
	from AFList wf_e have ids1: "set ids \<subseteq> ?ids" by (auto simp: list_all_iff)
	from AFList mono_e have ids2: "?ids \<subseteq> set ids" 	
	  by (simp only: l, auto simp: list_all_iff)
	from ids1 ids2 have ids: "set ids = ?ids" by auto
	let ?eq = "\<lambda> i. ?pi (?bsa ! i) = ?pi (?bta ! i)"
	let ?gr = "\<lambda> i. (?pi (?bsa ! i), ?pi (?bta ! i)) \<in> S"
	let ?lb = "length bef"
	have geq: "\<forall> i. ?eq i \<or> ?gr i"
	proof
	  fix i
	  {
	    assume "i < ?lb" hence "?eq i" by (simp add: nth_append)
	  }
	  moreover
	  {
	    assume "i = ?lb"
	    with st have "?gr i" by (simp add: nth_append)
	  }
	  moreover 
	  {
	    assume i2: "i > ?lb"
	    then obtain k where k1: "i - length bef = Suc k" by (cases "i - length bef", arith+)
	    have "?eq i" unfolding nth_append using i2 k1 by auto
	  }
	  moreover
	  {
	    have "i < ?lb \<or> i = ?lb \<or> i > ?lb" by arith
	  }
	  ultimately show "?eq i \<or> ?gr i" by blast
	qed
	from ids have lb: "?lb \<in> set ids" by auto
	from st have "?gr ?lb" by (simp add: nth_append)
	with lb have gr: "\<exists> i \<in> set ids. ?gr i \<and> i = ?lb" by auto
	obtain ss ts pi g where g: "g = ?f" and ss: "ss = ?bsa" and ts: "ts = ?bta" and pi: "pi = (?pi :: ('f,'v)term \<Rightarrow> ('f filtered,'v)term)" by auto
	let ?eq = "\<lambda> i. pi (ss ! i) = pi (ts ! i)"
	let ?gr = "\<lambda> i. (pi (ss ! i), pi (ts ! i)) \<in> S"
	let ?ss = "map (op ! (map pi ss))"
	let ?ts = "map (op ! (map pi ts))"
	let ?both = "\<lambda> n. ?ts (take n ids) @ ?ss (drop n ids)"
	let ?gss = "Fun g (?ss ids)"
	let ?gboth = "\<lambda> n. Fun g (?both n)"
	from geq ss ts pi have geq: "\<forall> i. ?eq i \<or> ?gr i" by auto
	from gr ss ts pi have "\<exists> i \<in> set ids. ?gr i \<and> i = ?lb" by auto
	from this obtain i where i: "?gr i" and iids: "i \<in> set ids" and idef: "i = ?lb" by auto
	have "\<forall> n. (i \<in> set (take n ids) \<longrightarrow> (?gss, ?gboth n) \<in> S) \<and> (i \<notin> set (take n ids) \<longrightarrow> ?gss = ?gboth n)" (is "\<forall> n. ?p n") 
	proof (intro allI)
	  fix n
	  show "?p n"
	  proof (induct n, simp)
	    case (Suc n)
	    show ?case
	    proof (cases "n \<ge> length ids")
	      case True
	      with Suc show ?thesis by simp
	    next
	      case False 
	      hence n: "n < length ids" by auto
	      let ?bef = "?ts (take n ids)"
	      let ?aft = "?ss (drop (Suc n) ids)"
	      let ?ssn = "pi (ss ! (ids ! n))"
	      let ?tsn = "pi (ts ! (ids ! n))"
	      from n ids have ids_n: "ids ! n < ?l" by force
	      have gn: "?both n = ?bef @ ?ssn # ?aft"
		by (simp only: drop_Suc_conv_tl[OF n, symmetric], simp, simp add: nth_map[of "ids ! n" ?bsa pi, simplified, OF ids_n] ss)
	      have gsn: "?both (Suc n) = ?bef @ ?tsn # ?aft"
		by (simp only: take_Suc_conv_app_nth[OF n], simp, simp add: nth_map[of "ids ! n" ?bta pi, simplified, OF ids_n] ts)
	      {
		assume "?ssn \<noteq> ?tsn \<or> ids ! n = i"
		with geq i have "(?ssn, ?tsn) \<in> S" by blast		
		with monoS[unfolded one_step_ctxt_closed_def] 
		have gnsn: "(?gboth n, ?gboth (Suc n)) \<in> S"
		  by (simp only: gn gsn)
	      } note noteq_i = this
	      {
		assume "ids ! n \<noteq> i"
		{
		  assume "ids ! n < ?lb"
		  hence "?gboth (Suc n) = ?gboth n" by (simp add: gn gsn, simp add: ss ts, simp add: nth_append)
		}
		moreover
		{
		  assume i2: "ids ! n > ?lb"
		  have "?gboth (Suc n) = ?gboth n" unfolding term.simps gn gsn 
                    using i2 ss ts by (simp add: nth_append)
		}
		moreover
		{
		  from `ids ! n \<noteq> i` have "ids ! n < ?lb \<or> ids ! n > ?lb" by (simp only: idef, arith)
		}
		ultimately have "?gboth (Suc n) = ?gboth n" by blast
	      } note eq = this	      
	      show ?thesis 
	      proof (cases "i \<in> set (take (Suc n) ids)")
		case False 
		hence i_not: "i \<notin> set (take n ids)" and i_not2: "ids ! n \<noteq> i" using take_Suc_conv_app_nth[OF n] by auto
		from i_not Suc have rec: "?gss = ?gboth n" by blast 
		with False eq[OF i_not2] rec show ?thesis by simp 
	      next
		case True note oTrue = this
		show ?thesis
		proof (cases "i \<in> set (take n ids)")
		  case False
		  with True have id: "ids ! n = i" using take_Suc_conv_app_nth[OF n] by auto
		  from False Suc have rec: "?gss = ?gboth n" by blast
		  from noteq_i id have "(?gboth n, ?gboth (Suc n)) \<in> S" by blast
		  with True rec show ?thesis by simp
		next
		  case True
		  with Suc have rec: "(?gss, ?gboth n) \<in> S" by blast
		  show ?thesis
		  proof (cases "ids ! n = i")
		    case False
		    from rec eq[OF False] oTrue show ?thesis by simp
		  next
		    case True
		    with noteq_i have "(?gboth n, ?gboth (Suc n)) \<in> S" by blast
		    from rec this have "(?gss, ?gboth (Suc n)) \<in> S" using trans_S unfolding trans_def by blast
		    with oTrue show ?thesis by simp
		  qed
		qed
	      qed
	    qed
	  qed
	qed
	with spec[OF this, of "length ids"] iids have "(Fun g (?ss ids), Fun g (?ts ids)) \<in> S" by auto
	with g ss ts pi show "(Fun ?f (map (op ! (map ?pi bef @ ?pi s # map ?pi aft)) ids),
	  Fun ?f (map (op ! (map ?pi bef @ ?pi t # map ?pi aft)) ids)) \<in> S"
	  by auto 
      qed
    qed
  }
  thus "one_step_ctxt_closed ?S"
    unfolding afs_rel_def afs_rule_def one_step_ctxt_closed_def
    by auto
qed
end

sublocale afs_redtriple \<subseteq> redtriple_order "afs_rel S" "afs_rel NS" "afs_rel NST"
proof -
  let ?S = "afs_rel S"
  let ?NS = "afs_rel NS"
  let ?NST = "afs_rel NST"
  show "redtriple_order ?S ?NS ?NST"
  proof
    show "refl ?NS"
      by (rule afs_refl[OF refl_NS])
  next
    show "refl ?NST"
      by (rule afs_refl[OF refl_NST])
  next
    show "trans ?S"
      by (rule afs_trans[OF trans_S])
  next
    show "trans ?NS"
      by (rule afs_trans[OF trans_NS])
  next
    show "trans ?NST"
      by (rule afs_trans[OF trans_NST])
  qed 
qed

sublocale afs_redtriple_impl \<subseteq> redtriple_impl "afs_rel S" "afs_rel NS" "afs_rel NST" "afs_check rS \<pi> cS" "afs_check rNS \<pi> cNS" "afs_check rNST \<pi> cNST"
proof -
  let ?S = "afs_rel S"
  let ?NS = "afs_rel NS"
  let ?NST = "afs_rel NST"
  let ?cS = "afs_check rS \<pi> cS"
  let ?cNS = "afs_check rNS \<pi> cNS"
  let ?cNST = "afs_check rNST \<pi> cNST"
  show "redtriple_impl ?S ?NS ?NST ?cS ?cNS ?cNST"
  proof
    fix lr
    assume "isOK (afs_check rS \<pi> cS lr)"
    thus "lr \<in> ?S" unfolding afs_check_def afs_rel_def 
      using checkSSound by (auto simp: split_paired_all Let_def afs_rule_def)
  next
    fix lr
    assume "isOK (afs_check rNS \<pi> cNS lr)"
    thus "lr \<in> ?NS" unfolding afs_check_def afs_rel_def 
      using checkNSSound by (auto simp: split_paired_all Let_def afs_rule_def)
  next
    fix lr
    assume "isOK (afs_check rNST \<pi> cNST lr)"
    thus "lr \<in> ?NST" unfolding afs_check_def afs_rel_def 
      using checkNSTSound by (auto simp: split_paired_all Let_def afs_rule_def)
  qed
qed


sublocale mono_afs_redtriple_impl \<subseteq> mono_redtriple_impl "afs_rel S" "afs_rel NS" "afs_rel NST" "afs_check rS \<pi> cS" "afs_check rNS \<pi> cNS" "afs_check rNST \<pi> cNST" "cm \<guillemotright> check (mono_afs \<pi>) mess"
proof
  assume "isOK (cm \<guillemotright> check (mono_afs \<pi>) mess)"
  with checkMonoSound have monoS: "one_step_ctxt_closed S" and mono_pi: "mono_afs \<pi>" by auto
  from one_step_ctxt_closed[OF mono_pi monoS] show "one_step_ctxt_closed (afs_rel S)" .
qed
  

type_synonym 'f afs_list = "(('f \<times> nat) \<times> af_entry)list"

definition default_af_entry :: "nat \<Rightarrow> af_entry"
  where "default_af_entry n \<equiv> AFList [0..<n]"

definition afs_list_to_afs :: "'f afs_list \<Rightarrow> 'f afs"
  where "afs_list_to_afs \<pi> f n \<equiv> case lookup (f,n) \<pi> of Some entry \<Rightarrow> entry | None \<Rightarrow> default_af_entry n"

definition afs_to_unused_arity :: "'f afs_list \<Rightarrow> nat"
  where "afs_to_unused_arity \<pi> \<equiv> foldr (\<lambda> ((f,n),e) m. max n m) \<pi> 0"

lemma afs_to_unused_arity_sound: shows "\<forall> e \<in> set pi. snd (fst e) \<le> afs_to_unused_arity pi"
proof (induct pi)
  case (Cons e es)
  {
    fix f n en
    assume "((f,n),en) \<in> set (e # es)"
    hence "((f,n),en) = e \<or> ((f,n),en) \<in> set es" (is "?one \<or> ?two")
      by simp
    hence "n \<le> afs_to_unused_arity (e # es)"
    proof
      assume ?one thus ?thesis unfolding afs_to_unused_arity_def by auto
    next
      assume ?two with Cons have "n \<le> afs_to_unused_arity es" by auto
      also have "\<dots> \<le> afs_to_unused_arity (e # es)" unfolding afs_to_unused_arity_def by (induct es arbitrary: e, auto)
      finally show ?thesis .
    qed
  }
  thus ?case by force 
qed simp

lemma default_af_entry_id: 
  assumes "\<pi> f (length ts) = default_af_entry (length ts)"
  shows "afs_term \<pi> (Fun f ts) = Fun (FPair f (length ts)) (map (afs_term \<pi>) ts)"
  by (simp add: default_af_entry_def assms Let_def, simp only: list_eq_iff_nth_eq, simp)

lemma wf_default_af_entry: 
  "wf_af_entry n (default_af_entry n)"
  unfolding default_af_entry_def by (auto simp: list_all_iff)

definition check_wf_afs :: "('f::show)afs_list \<Rightarrow> shows check"
where "check_wf_afs pi = check_allm (\<lambda>((f,n),e). (case e of
  Collapse i \<Rightarrow> check (i < n) i
| AFList ids \<Rightarrow> check_all (\<lambda> i. i < n) ids
) <+? (\<lambda>i. ''index '' +#+ shows i +@+ '' too high for AF-entry of'' +#+ shows(f,n) +@+ shows_nl)
) pi"

lemma check_wf_afs_sound: assumes check: "isOK(check_wf_afs pi)"
  shows "afs (afs_list_to_afs pi)"
proof (intro afs.intro, unfold wf_afs_def afs_list_to_afs_def, intro allI impI)
  fix f n
  show "wf_af_entry n (case lookup (f, n) pi of None \<Rightarrow> default_af_entry n | Some en \<Rightarrow> en)"
  proof (cases "lookup (f,n) pi", simp add: wf_default_af_entry)
    case (Some en)
    hence mem: "((f,n),en) \<in> set pi" by (rule lookup_sound)
    show ?thesis 
    proof (simp add: Some, cases en)
      case (Collapse i)
      with Some mem check show "wf_af_entry n en" unfolding check_wf_afs_def by auto
    next
      case (AFList ids)
      with Some mem check show "wf_af_entry n en" unfolding check_wf_afs_def by (auto simp: list_all_iff)
    qed
  qed
qed

fun check_mono_afs_help where "check_mono_afs_help ((f,a),e) = (''entry for '' +#+ shows f +@+ '' not monotonic'' +#+ shows_nl)"

definition
  check_mono_afs :: "('f::show)afs_list \<Rightarrow> shows check"
where
  "check_mono_afs pi = (check_all (\<lambda>x . mono_af_entry (snd(fst x)) (snd x)) pi <+? check_mono_afs_help)"

lemma check_mono_afs_sound: assumes check: "isOK(check_mono_afs pi)"
  and check2: "wf_afs (afs_list_to_afs pi)"
  shows "mono_afs (afs_list_to_afs pi)"
proof (unfold mono_afs_def afs_list_to_afs_def, intro allI impI)
  fix f n
  show "mono_af_entry n (case lookup (f, n) pi of None \<Rightarrow> default_af_entry n | Some en \<Rightarrow> en)"
  proof (cases "lookup (f,n) pi")
    case None thus ?thesis by (simp add: wf_default_af_entry default_af_entry_def list_all_iff)
  next
    case (Some en)
    hence mem: "((f,n),en) \<in> set pi" by (rule lookup_sound)
    show ?thesis 
    proof (simp add: Some, cases en)
      case (Collapse i)
      with Some mem check check2 show "mono_af_entry n en" unfolding check_mono_afs_def mono_af_entry_def wf_afs_def by auto
    next
      case (AFList ids)
      with Some mem check check2 show "mono_af_entry n en" unfolding check_mono_afs_def mono_af_entry_def wf_afs_def by auto
    qed
  qed
qed

definition afs_list_to_af :: "'f afs_list \<Rightarrow> 'f af" where
  "afs_list_to_af pi fn \<equiv> case lookup fn pi of
    Some e \<Rightarrow> (case e of Collapse j \<Rightarrow> (op = j) | AFList ids \<Rightarrow> (\<lambda> i. i \<in> set ids))
  | None \<Rightarrow> \<lambda> _ .True"

context afs
begin

lemma apply_af_entry_const_is_id: "apply_af_entry (FPair f (0 :: nat)) (\<pi> f 0) [] = Fun (FPair f 0) []"
proof -
  let ?e = "\<pi> f 0"
  from wf_afs have "wf_af_entry 0 ?e" unfolding wf_afs_def by auto
  thus ?thesis by (cases ?e, simp_all add: list_all_iff)
qed
end

lemma af_compatible:
  fixes pi :: "'f afs_list"
  assumes wf_afs: "wf_afs (afs_list_to_afs pi)" (is "wf_afs ?pi")
  and refl_NS: "refl NS"
  and all_ctxt_closed: "all_ctxt_closed NS"
  shows "af_compatible (afs_list_to_af pi) (afs.afs_rel ?pi (NS :: ('f filtered,'v)term ars))" (is "af_compatible ?af ?NS")
  unfolding af_compatible_def
proof (intro allI)      
  fix f :: 'f and bef :: "('f,'v)term list" and s t :: "('f,'v)term" and  aft
  let ?pi = "afs_list_to_afs pi"
  let ?bsa = "bef @ s # aft"
  let ?bta = "bef @ t # aft"
  let ?lb = "length bef"
  let ?l = "Suc (?lb + length aft)"    
  let ?s = "Fun f ?bsa"
  let ?t = "Fun f ?bta"
  show "?af (f, ?l) ?lb \<or> (?s,?t) \<in> ?NS"
  proof (cases "?af (f, ?l) ?lb")
    case False
    let ?e = "?pi f ?l"
    from wf_afs have wf: "wf_af_entry ?l ?e" unfolding wf_afs_def by auto
    from wf_afs have afs: "afs ?pi" by (rule afs.intro) 
    show ?thesis
    proof (cases "lookup (f,?l) pi")
      case None
      with False show ?thesis unfolding afs_list_to_af_def by auto
    next
      case (Some e)
      hence e: "?e = e" by (simp add: afs_list_to_afs_def)
      let ?ls = "length (bef @ s # aft)"
      let ?lt = "length (bef @ t # aft)"
      let ?msa = "map (afs_term ?pi) ?bsa"
      let ?mta = "map (afs_term ?pi) ?bta"
      let ?pfs = "apply_af_entry (FPair f ?l) ?e ?msa"
      let ?pft = "apply_af_entry (FPair f ?l) ?e ?mta"
      have all2: "\<forall> i < ?l. i \<noteq> ?lb \<longrightarrow> (?msa ! i, ?mta ! i) \<in> NS" 
      proof (intro allI impI)
	fix i
	assume "i < ?l" and ib: "i \<noteq> ?lb"
	hence i_s: "i < ?ls" and i_t: "i < ?lt" by simp_all
	{
	  assume bef: "i < length bef"
	  with refl_NS have "(?msa ! i, ?mta ! i) \<in> NS" unfolding refl_on_def 
	    by (simp only: nth_map[OF i_s], simp add: nth_append)
	}
	moreover
	{
	  assume i: "i > length bef"
	  from this obtain k where k1: "i - length bef = Suc k" by (cases "i - length bef", arith+)
	  with i_s have k2: "k < length aft" by auto
	  from i refl_NS have "(?msa ! i, ?mta ! i) \<in> NS" unfolding refl_on_def 
	    by (simp only: nth_map[OF i_s] nth_map[OF i_t], simp add: nth_append k1)
	} 
	moreover
	{
	  have "i < length bef \<or> i = length bef \<or> i > length bef" by auto
	}      
	ultimately
	show "(?msa ! i, ?mta ! i) \<in> NS" using ib by blast
      qed 
      have "(?pfs,?pft) \<in> NS" 
      proof (cases ?e)
	case (Collapse i)
	with wf have len: "i < ?l"  by simp
	from Collapse[simplified] Some False have "i \<noteq> ?lb" unfolding afs_list_to_afs_def afs_list_to_af_def by auto
	with all2 len Collapse show ?thesis by auto
      next
	case (AFList ids)
	with wf have ids: "\<forall> i \<in> set ids. i < ?l" by (simp add: list_all_iff)
	show ?thesis 
	proof (simp only: AFList apply_af_entry.simps)
	  let ?msi = "map (op ! (map (afs_term ?pi) ?bsa)) ids"
	  let ?mti = "map (op ! (map (afs_term ?pi) ?bta)) ids"
	  from ids False AFList[simplified] Some have "\<forall> i \<in> set ids. i \<noteq> ?lb" unfolding afs_list_to_afs_def afs_list_to_af_def 
	    by (auto)
	  with ids all2 have "\<forall> i < length ids. (?msi ! i, ?mti ! i) \<in> NS" by auto
	  with all_ctxt_closed show "(Fun (FPair f ?l) ?msi, Fun (FPair f ?l) ?mti) \<in> NS"
	    unfolding all_ctxt_closed_def by force
	qed
      qed
      thus "afs_list_to_af pi (f, ?l) ?lb \<or> (Fun f ?bsa, Fun f ?bta) \<in> afs.afs_rel ?pi NS" unfolding afs.afs_rel_def[OF afs] afs_rule_def by (simp add: afs)
    qed
  qed simp
qed

lemma af_ce_compat: fixes pi :: "('f :: show) afs_list"
  and NS :: "('f filtered,'v :: show)term ars"
  assumes wf_afs: "wf_afs (afs_list_to_afs pi)" 
  and ce_compat: "\<exists> n. \<forall> m \<ge> n. \<forall> c d. ce_trs (c,d,m) \<subseteq> NS"
  shows "\<exists> n. \<forall> m \<ge> n. \<forall> c d. ce_trs (c,d,m) \<subseteq> afs.afs_rel (afs_list_to_afs pi) NS"
proof -
  let ?pi = "afs_list_to_afs pi"
  let ?NS = "afs.afs_rel ?pi NS"
  let ?cS = "afs_check rS ?pi cS"
  let ?cNS = "afs_check rNS ?pi cNS"
  let ?ar = "afs_to_unused_arity pi"
  let ?sar = "Suc (Suc ?ar)"
  let ?af = "afs_list_to_af pi"
  have pi2: "\<And> m c. m \<ge> ?sar \<Longrightarrow> ?pi c m = default_af_entry m"
  proof -
    fix m c
    assume m: "m \<ge> ?sar"
    show "?pi c m = default_af_entry m"
    proof (unfold afs_list_to_afs_def, cases "lookup (c, m) pi")
      case (Some foo)
      hence "((c,m),foo) \<in> set pi" by (rule lookup_sound)
      from bspec[OF afs_to_unused_arity_sound this] m
      show "(case lookup (c, m) pi of None \<Rightarrow> default_af_entry m | Some foo \<Rightarrow> foo) = (default_af_entry m)" by simp
    qed simp
  qed
  from wf_afs have afs: "afs ?pi" by (rule afs.intro) 
  from ce_compat obtain n where ce_NS: "\<And> m c d . m \<ge> n \<Longrightarrow> ce_trs (c,d,m) \<subseteq> NS" by blast
  let ?n = "max n ?sar"
  show "\<exists> n. \<forall> m \<ge> n. \<forall> c d. ce_trs (c,d,m) \<subseteq> ?NS"
  proof (rule exI[of _ ?n], intro allI impI)
    fix m c d
    assume m: "m \<ge> ?n"
    hence mn: "m \<ge> n" by simp
    {
      fix t s :: "('f,'v)term"
      let ?list = "t # s # replicate m (Fun d [])"
      from pi2 m have id: "?pi c (length ?list) = default_af_entry (length ?list)" by simp
      have "(afs_term ?pi (Fun c ?list), afs_term ?pi t) \<in> NS"
 	using ce_NS[OF mn, simplified ce_trs.simps, of "FPair c (Suc (Suc m))" "FPair d 0"]  
        by (simp only: default_af_entry_id[of ?pi c ?list, OF id], auto simp: afs.apply_af_entry_const_is_id[OF afs])
    }
    moreover
    {
      fix t s :: "('f,'v)term"
      let ?list = "t # s # replicate m (Fun d [])"
      from pi2 m have id: "?pi c (length ?list) = default_af_entry (length ?list)" by simp
      have "(afs_term ?pi (Fun c ?list), afs_term ?pi s) \<in> NS"
 	using ce_NS[OF mn, simplified ce_trs.simps, of "FPair c (Suc (Suc m))" "FPair d 0"]  
        by (simp only: default_af_entry_id[of ?pi c ?list, OF id], auto simp: afs.apply_af_entry_const_is_id[OF afs])
    }
    ultimately
    show "ce_trs (c,d,m) \<subseteq> ?NS"
      by (auto simp: ce_trs.simps afs.afs_rel_def[OF afs] afs_rule_def pi2)      
  qed
qed



locale afs_ce_redtriple_impl = afs_redtriple_impl "afs_list_to_afs pi" S NS NST cS cNS cNST + filtered: redtriple_impl S NS NST cS cNS cNST + ce: ce_redtriple S NS NST + order: redtriple_order S NS NST
  for pi and S NS NST :: "(('f::show)filtered,'v::show)trs" and cS cNS cNST

locale afs_ce_mono_redtriple_impl = mono_afs_redtriple_impl "afs_list_to_afs pi" S NS NST cS cNS cNST cm + filtered: ce_mono_redtriple_impl S NS NST cS cNS cNST cm + order: redtriple_order S NS NST
  for pi and S NS NST :: "(('f::show)filtered,'v::show)trs" and cS cNS cNST cm

sublocale afs_ce_mono_redtriple_impl \<subseteq> afs_ce_redtriple_impl ..

sublocale afs_ce_redtriple_impl \<subseteq> ce_af_redtriple_impl "afs_rel S" "afs_rel NS" "afs_rel NST" "afs_check rS (afs_list_to_afs pi) cS" "afs_check rNS (afs_list_to_afs pi) cNS" "afs_check rNST (afs_list_to_afs pi) cNST" "afs_list_to_af pi"
proof -
  let ?pi = "afs_list_to_afs pi"
  let ?S = "afs_rel S"
  let ?NS = "afs_rel NS"
  let ?NST = "afs_rel NST"
  let ?cS = "afs_check rS ?pi cS"
  let ?cNS = "afs_check rNS ?pi cNS"
  let ?cNST = "afs_check rNST ?pi cNST"
  let ?af = "afs_list_to_af pi"
  show "ce_af_redtriple_impl ?S ?NS ?NST ?cS ?cNS ?cNST ?af"
  proof
    have rp1: "pre_order_pair S NS" ..
    have rp2: "redpair_order S NS" ..
    show "af_compatible ?af ?NS"
      by (rule af_compatible[OF wf_afs pre_order_pair.refl_NS[OF rp1] redpair_order.all_ctxt_closed[OF rp2]])
  next
    show "\<exists> n. \<forall> m \<ge> n. \<forall> c d. ce_trs (c,d,m) \<subseteq> ?NS"
      by (rule af_ce_compat[OF wf_afs NS_ce_compat])
  qed
qed


sublocale afs_ce_mono_redtriple_impl \<subseteq> ce_mono_redtriple_impl "afs_rel S" "afs_rel NS" "afs_rel NST" "afs_check rS (afs_list_to_afs pi) cS" "afs_check rNS (afs_list_to_afs pi) cNS" "afs_check rNST (afs_list_to_afs pi) cNST" "cm \<guillemotright> check_mono_afs pi"
proof 
  assume "isOK (cm \<guillemotright> check_mono_afs pi)"
  hence c1: "isOK(cm)" and c2: "isOK(check_mono_afs pi)" by auto
  hence check: "isOK(cm \<guillemotright> check (mono_afs (afs_list_to_afs pi)) mess)"
    using check_mono_afs_sound[OF c2 wf_afs] by auto 
  show "one_step_ctxt_closed (afs.afs_rel (afs_list_to_afs pi) S)"
    by (rule checkMonoSound[OF check])
next
  let ?S = "afs_rel S"
  show "\<exists> n. \<forall> m \<ge> n. \<forall> c d. ce_trs (c,d,m) \<subseteq> ?S"
    by (rule af_ce_compat[OF wf_afs S_ce_compat])
qed


fun shows_afs :: "('f::show)afs_list \<Rightarrow> shows"
where "shows_afs af = shows_map (\<lambda>((f,n),e). ''pi('' +#+ shows f +@+ ''/'' +#+ shows n +@+ '') = '' +#+ (case e of
  Collapse i \<Rightarrow> shows(Suc i)
| AFList ids \<Rightarrow> shows(map Suc ids)) +@+ shows_nl) af"


definition af_redtriple :: "'f afs_list \<Rightarrow> ('f filtered,'v)redtriple \<Rightarrow> ('f :: show,'v :: show)redtriple"
where "af_redtriple pi rp \<equiv> let af = afs_list_to_af pi; afs = afs_list_to_afs pi in \<lparr>
  redtriple.valid = do {check_wf_afs pi; redtriple.valid rp}, 
  s = afs_check (shows_string ''>'') afs (redtriple.s rp),
  ns = afs_check (shows_string ''>='') afs (redtriple.ns rp),
  nst = afs_check (shows_string ''>='') afs (redtriple.nst rp),
  af = af,
  mono = do {check_mono_afs pi; redtriple.mono rp} ,
  desc = (shows_string ''Argument Filter: '') +@+ shows_nl +@+ shows_afs pi +@+ shows_nl +@+ redtriple.desc rp
  \<rparr>"

lemma af_redtriple: 
  fixes pi :: "('f :: show) afs_list" and
        rp :: "'a \<Rightarrow> ('f filtered, 'v :: show)redtriple"
  assumes rp: "generic_redtriple_impl rp" shows 
  "generic_redtriple_impl (\<lambda> I. (af_redtriple pi (rp I)))"
proof
  fix I s_list ns_list nst_list
  let ?rp = "af_redtriple pi (rp I) :: ('f :: show, 'v:: show)redtriple"
  assume valid: "isOK(redtriple.valid ?rp)" and stri: "isOK (check_allm (redtriple.s ?rp) s_list)" and nstri: "isOK (check_allm (redtriple.ns ?rp) ns_list)"
    and nstrit: "isOK (check_allm (redtriple.nst ?rp) nst_list)"
  let ?pi = "afs_list_to_afs pi"
  let ?p = "afs_list_to_af pi"
  let ?pi' = "redtriple.af (rp I)"
  let ?pi'' = "redtriple.af (af_redtriple pi (rp I))"
  have pi'': "?pi'' = ?p"  unfolding af_redtriple_def Let_def by simp
  let ?ar = "afs_to_unused_arity pi"
  let ?sar = "Suc (Suc ?ar)"
  let ?af = "afs_list_to_af pi"
  let ?af_term = "afs_term ?pi"
  let ?af_list = "map (\<lambda> (s,t). (?af_term s, ?af_term t))"
  let ?as_list = "?af_list s_list"
  let ?ans_list = "?af_list ns_list"
  let ?anst_list = "?af_list nst_list"
  from valid have valid: "isOK(redtriple.valid (rp I))" and wf_afs: "isOK(check_wf_afs pi)" unfolding af_redtriple_def by (auto simp: Let_def)
  from wf_afs interpret afs ?pi by (rule check_wf_afs_sound)
  from stri have stri: "isOK(check_allm (redtriple.s (rp I)) ?as_list)" by (auto simp: Let_def af_redtriple_def afs_check_def)
  from nstri have nstri: "isOK(check_allm (redtriple.ns (rp I)) ?ans_list)" by (auto simp: Let_def af_redtriple_def afs_check_def)
  from nstrit have nstrit: "isOK(check_allm (redtriple.nst (rp I)) ?anst_list)" by (auto simp: Let_def af_redtriple_def afs_check_def)
  from generic_redtriple_impl.generate_redtriple[of rp, OF rp valid stri nstri nstrit]
  obtain S NS NST where redtriple: "ce_af_redtriple_order S NS NST ?pi'"
     and stri: "set ?as_list \<subseteq> S"
     and nstri: "set ?ans_list \<subseteq> NS"
     and nstrit: "set ?anst_list \<subseteq> NST"
     and cmono: "isOK (redtriple.mono (rp I)) \<longrightarrow> mono_ce_af_redtriple_order S NS NST ?pi' \<and> one_step_ctxt_closed S" by auto
  let ?aS = "afs_rel S"
  let ?aNS = "afs_rel NS"
  let ?aNST = "afs_rel NST"
  from redtriple interpret ce_af_redtriple_order S NS NST ?pi' .
  have rp1: "pre_order_pair S NS" ..
  have rp2: "redpair_order S NS" ..
  interpret afs_redtriple ?pi S NS NST ..
  interpret ce_af_redtriple_order ?aS ?aNS ?aNST ?p
  proof(unfold_locales, rule af_ce_compat[OF wf_afs NS_ce_compat])
    show "af_compatible ?p ?aNS"
      by (rule af_compatible[OF wf_afs pre_order_pair.refl_NS[OF rp1] redpair_order.all_ctxt_closed[OF rp2]])
  qed
  show "\<exists>S NS NST. ce_af_redtriple_order S NS NST (redtriple.af (af_redtriple pi (rp I))) \<and>
                 set s_list \<subseteq> S \<and>
                 set ns_list \<subseteq> NS \<and>
                 set nst_list \<subseteq> NST \<and>
                 (isOK (redtriple.mono (af_redtriple pi (rp I))) \<longrightarrow>
                  mono_ce_af_redtriple_order S NS NST ?pi'' \<and> one_step_ctxt_closed S)" unfolding pi''
  proof (intro exI conjI)
    show "set s_list \<subseteq> ?aS" using stri unfolding afs_rel_def afs_rule_def by auto
  next
    show "set ns_list \<subseteq> ?aNS" using nstri unfolding afs_rel_def afs_rule_def by auto
  next
    show "set nst_list \<subseteq> ?aNST" using nstrit unfolding afs_rel_def afs_rule_def by auto
  next
    show "ce_af_redtriple_order ?aS ?aNS ?aNST ?p" ..
  next
    show "isOK (redtriple.mono (af_redtriple pi (rp I))) \<longrightarrow>
                  mono_ce_af_redtriple_order ?aS ?aNS ?aNST ?p \<and> one_step_ctxt_closed ?aS" (is "?prem \<longrightarrow> ?mono \<and> ?ctxt")
    proof
      assume ?prem
      with cmono have mono: "mono_ce_af_redtriple_order S NS NST ?pi'" and cmono: "isOK(check_mono_afs pi)" and ctxt: "one_step_ctxt_closed S" unfolding af_redtriple_def by (auto simp: Let_def)
      from mono interpret mono_ce_af_redtriple_order S NS NST ?pi' .
      have ?mono 
        by (unfold_locales, rule af_ce_compat[OF wf_afs S_ce_compat])
      have ?ctxt
        by (rule one_step_ctxt_closed[OF check_mono_afs_sound[OF cmono wf_afs] ctxt])
      from `?mono` `?ctxt` show "?mono \<and> ?ctxt" ..
    qed
  qed
qed

end
