(*
Copyright 2011 Christian Sternagel, René Thiemann

This file is part of IsaFoR/CeTA.

IsaFoR/CeTA is free software: you can redistribute it and/or modify it under the
terms of the GNU Lesser General Public License as published by the Free Software
Foundation, either version 3 of the License, or (at your option) any later
version.

IsaFoR/CeTA is distributed in the hope that it will be useful, but WITHOUT ANY
WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR A
PARTICULAR PURPOSE.  See the GNU Lesser General Public License for more details.

You should have received a copy of the GNU Lesser General Public License along
with IsaFoR/CeTA. If not, see <http://www.gnu.org/licenses/>.
*)
theory Critical_Pairs_Impl
imports 
  Critical_Pairs
  "Q-Restricted_Rewriting_Impl"
begin

definition critical_pairs_impl :: "('f,string)rules \<Rightarrow> ('f,string)rules \<Rightarrow> (bool \<times> ('f,string)rule)list" 
  where "critical_pairs_impl P R \<equiv> concat (map (\<lambda> (l,r). concat (map (\<lambda> p. let C = ctxt_of_pos_term p l; l'' = l |_ p; b = (C = \<box>) in 
  if is_Var l'' then [] else concat (map (\<lambda> (l',r'). case mgu_var_disjoint_string l'' l' of Some (\<sigma>,\<tau>) \<Rightarrow> [(b, r \<cdot> \<sigma>, (C \<cdot>c \<sigma>)\<langle>r' \<cdot> \<tau>\<rangle>)] | None \<Rightarrow> []) R)) (poss_list l))) P)"

lemma critical_pairs_impl[simp]: "set (critical_pairs_impl P R) = critical_pairs (set P) (set R)" (is "?l = ?r")
proof -
  note cpdefs = critical_pairs_impl_def critical_pairs_def set_concat
    set_map Let_def poss_list_sound
  {
    fix b s t
    assume "(b,s,t) \<in> ?r"
    from this[unfolded cpdefs]
    obtain l r l' r' l'' C \<sigma> \<tau> where
      b: "b = (C = \<box>)" and s: "s = r \<cdot> \<sigma>" and t: "t = (C \<cdot>c \<sigma>)\<langle>r' \<cdot> \<tau>\<rangle>"
      and P: "(l,r) \<in> set P" and R: "(l', r') \<in> set R"
      and l: "l = C\<langle>l''\<rangle>" and l'': "is_Fun l''"
      and mgu: "mgu_var_disjoint_string l'' l' = Some (\<sigma>, \<tau>)"
      by auto
    let ?p = "hole_pos C"
    from l have p: "?p \<in> poss l" by auto
    from l p have C: "C = ctxt_of_pos_term ?p l" by auto
    from l p have l: "l |_ ?p = l''" by auto
    have "(b,s,t) \<in> ?l" unfolding cpdefs b s t
      by (rule, rule, rule P, unfold cpdefs, rule, rule imageI[OF p],
      insert C l R l'' mgu, force)
  } 
  hence "?r \<subseteq> ?l" by auto
  moreover
  {
    fix b s t
    assume "(b,s,t) \<in> ?l"
    from this[unfolded cpdefs]
    obtain l r p l' r' \<sigma> \<tau> where
      P: "(l,r) \<in> set P" and p: "p \<in> poss l" and lp: "is_Fun (l |_ p)" 
      and R: "(l',r') \<in> set R" and mgu: "Some (\<sigma>,\<tau>) = mgu_var_disjoint_string (l |_ p) l'" and b: "b = (ctxt_of_pos_term p l = \<box>)" and s: "s = r \<cdot> \<sigma>" and t: "t = (ctxt_of_pos_term p l \<cdot>c \<sigma>)\<langle>r' \<cdot> \<tau>\<rangle>"  by force
    have "(b,s,t) \<in> ?r"
      by (rule critical_pairsI[OF P R _ lp mgu[symmetric] s t b], insert ctxt_supt_id[OF p], simp)
  }
  hence "?l \<subseteq> ?r" by auto
  ultimately show ?thesis by auto
qed

definition critical_pairs_top_impl :: "('f,string)rules \<Rightarrow> ('f,string)rules \<Rightarrow> (('f,string)rule)list" 
  where "critical_pairs_top_impl P R \<equiv> concat (map (\<lambda> (l,r).  
  if is_Var l then [] else concat (map (\<lambda> (l',r'). case mgu_var_disjoint_string l l' of Some (\<sigma>,\<tau>) \<Rightarrow> [(r \<cdot> \<sigma>, r' \<cdot> \<tau>)] | None \<Rightarrow> []) R)) P)"

lemma critical_pairs_top_impl[simp]: "set (critical_pairs_top_impl P R) = {(s,t). (True,s,t) \<in> critical_pairs (set P) (set R)}" (is "?l = ?r")
proof -
  note cpdefs = critical_pairs_top_impl_def critical_pairs_def set_concat
    set_map
  {
    fix s t
    assume "(s,t) \<in> ?r"
    from this[unfolded cpdefs]
    obtain l r l' r' l'' C \<sigma> \<tau> where
      C: "C = \<box>" and s: "s = r \<cdot> \<sigma>" and t: "t = (C \<cdot>c \<sigma>)\<langle>r' \<cdot> \<tau>\<rangle>"
      and P: "(l,r) \<in> set P" and R: "(l', r') \<in> set R"
      and l: "l = C\<langle>l''\<rangle>" and l'': "is_Fun l''"
      and mgu: "mgu_var_disjoint_string l'' l' = Some (\<sigma>, \<tau>)"
      by auto
    have "(s,t) \<in> ?l" unfolding cpdefs s t 
      by (rule, rule, rule P, insert l[unfolded C] l'' mgu[symmetric] R C, force)
  }
  hence "?r \<subseteq> ?l" by auto
  moreover
  {
    fix s t
    assume st: "(s,t) \<in> ?l"
    have id: "\<And> p b t e. p \<in> set (if b then t else e) = (b \<and> p \<in> set t \<or> \<not> b \<and> p \<in> set e)" by auto
    from st[unfolded cpdefs]
    obtain l r l' r' \<sigma> \<tau> where
      P: "(l,r) \<in> set P" and lp: "is_Fun l" 
      and R: "(l',r') \<in> set R" and mgu: "Some (\<sigma>,\<tau>) = mgu_var_disjoint_string l l'" and s: "s = r \<cdot> \<sigma>" and t: "t = r' \<cdot> \<tau>" by (auto simp: id)
    have "(s,t) \<in> ?r" unfolding s t
      by (rule, unfold split, rule critical_pairsI[OF P R _ lp mgu[symmetric] refl, of \<box>], auto)
  }
  hence "?l \<subseteq> ?r" by auto
  ultimately show ?thesis by auto
qed

definition show_crit_pair :: "('f :: show,'v :: show)rule \<Rightarrow> shows"
  where "show_crit_pair lr \<equiv> 
  shows ''('' +@+ shows (fst lr) +@+ shows '', '' +@+ 
  shows  (snd lr) +@+ shows '')''"

definition check_critical_pairs_innermost :: "('f :: show,string)rules \<Rightarrow> shows check"
  where "check_critical_pairs_innermost R \<equiv> 
    check_allm (\<lambda> (l,r). check (l = r) (shows ''there is a non-trivial critical pair '' +@+ show_crit_pair (l,r))) (critical_pairs_top_impl R R)"

lemma check_critical_pairs_innermost: assumes ok: "isOK(check_critical_pairs_innermost R)" and NF: "NF_terms Q \<subseteq> NF_trs (set R)" 
  and var: "\<And> l r. (l,r) \<in> set R \<Longrightarrow> is_Fun l"
  shows "CR (qrstep Q (set R))"
proof (rule critical_pairs_innermost[OF _ NF var])
  fix l r
  assume "(True,l,r) \<in> critical_pairs (set R) (set R)"
  with ok[unfolded check_critical_pairs_innermost_def]
  show "l = r" by simp
qed

definition check_join :: "('f :: show,'v :: show)rules \<Rightarrow> ('f,'v)term \<Rightarrow> ('f,'v)rseq \<Rightarrow> ('f,'v)term \<Rightarrow> ('f,'v)rseq \<Rightarrow> shows check"
  where "check_join R s sseq t tseq = do {
           let u = last (s # map (\<lambda> (_,_,u). u) sseq);
           let v = last (t # map (\<lambda> (_,_,u). u) tseq);
           check (u = v) (shows ''the rewrite sequences end in different terms '' +@+ shows u +@+ shows '' and '' +@+ shows v);
           let chk = check_qrsteps (\<lambda> _. True) R;
           chk sseq s u;
           chk tseq t v
         }"

lemma check_join: assumes ok: "isOK(check_join R s sseq t tseq)"
  shows "\<exists> u. (s,u) \<in> (rstep (set R))^* \<and> (t,u) \<in> (rstep (set R))^*"
  by (rule exI[of _ "last (s # map (\<lambda> (_,_,u). u) sseq)"],
  insert ok[unfolded check_join_def],
  auto simp: Let_def intro: rel_pow_imp_rtrancl[OF check_qrsteps_sound[of "\<lambda> _. True" Nil R], simplified])

type_synonym ('f,'v)crit_pair_joins = "(('f,'v)term \<times> ('f,'v)rseq \<times> ('f,'v)term \<times> ('f,'v)rseq)list"

definition check_critical_pairs_guided :: "('f :: show,'v :: show)rules \<Rightarrow> (bool \<times> ('f,'v)rule) list \<Rightarrow> ('f,'v)crit_pair_joins \<Rightarrow> shows check"
  where "check_critical_pairs_guided R cp joins \<equiv> do {
    check_allm (\<lambda> (s,sseq,t,tseq). check_join R s sseq t tseq) joins;
    let cp' = map (\<lambda> (s,sseq,t,tseq). (s,t)) joins;
    check_allm (\<lambda> (b,s,t). check (s = t \<or> (\<exists> st' \<in> set cp'. instance_rule (s,t) st' \<or> instance_rule (t,s) st')) (shows ''critical pair '' +@+ show_crit_pair (s,t) +@+ shows '' missing'')) cp
  }"
     
lemma check_critical_pairs_guided: assumes ok: "isOK(check_critical_pairs_guided R (critical_pairs_impl R R) joins)"
  shows "WCR (rstep (set R))"
proof (rule critical_pairs)
  fix s t b
  assume cp: "(b,s,t) \<in> critical_pairs (set R) (set R)" and neq: "s \<noteq> t"
  let ?f = "\<lambda> (s,sseq,t,tseq). (s,t)"
  note ok = ok[unfolded check_critical_pairs_guided_def, unfolded Let_def, simplified]
  from ok cp neq have "\<exists> st' \<in> set joins. instance_rule (s,t) (?f st') \<or> instance_rule (t,s) (?f st')" by auto
  then obtain st' where join: "st' \<in> set joins" and inst: "instance_rule (s,t) (?f st') \<or> instance_rule (t,s) (?f st')" by auto
  obtain s' sseq t' tseq where st': "st' = (s',sseq,t',tseq)" by (cases st', auto)
  note join = join[unfolded st']
  let ?R = "(rstep (set R))^*"
  from ok join have "isOK (check_join R s' sseq t' tseq)" by auto
  from check_join[OF this] obtain u where su: "(s',u) \<in> ?R" and tu: "(t',u) \<in> ?R" by auto
  from inst[unfolded st' split_def fst_conv snd_conv]
  show "\<exists> l' r' u. instance_rule (s,t) (l',r') \<and> (l',u) \<in> ?R \<and> (r', u) \<in> ?R"
  proof
    assume inst: "instance_rule (s,t) (s',t')"
    show ?thesis
      by (intro exI conjI, rule inst, rule su, rule tu)
  next
    assume "instance_rule (t,s) (s',t')"
    hence inst: "instance_rule (s,t) (t',s')" unfolding instance_rule_def by auto
    show ?thesis
      by (intro exI conjI, rule inst, rule tu, rule su)
  qed
qed

definition check_critical_pairs_NF :: "('f :: show,'v :: show)rules \<Rightarrow> (bool \<times> ('f,'v)rule) list \<Rightarrow> shows check"
  where "check_critical_pairs_NF R cp \<equiv> do {
    check_allm (\<lambda> (_,s,t). 
       if (s = t) then OK else
       check_join_NF R s t
           <+? (\<lambda> e. shows ''problem when joining critical pair '' +@+ show_crit_pair (s,t) +@+ shows_nl +@+ e)
    ) cp
  }"


lemma check_critical_pairs_NF: fixes R :: "('f :: show, string)rules"
  assumes ok: "isOK(check_critical_pairs_NF R (critical_pairs_impl R R))"
  shows "WCR (rstep (set R))"
proof (rule critical_pairs)
  fix t1 t2 b
  assume cp: "(b,t1,t2) \<in> critical_pairs (set R) (set R)" and neq: "t1 \<noteq> t2"
  note ok = ok[unfolded check_critical_pairs_NF_def, unfolded Let_def, simplified]
  let ?Rel = "(rstep (set R))^*"
  from cp neq ok have "isOK(check_join_NF R t1 t2)"
    by auto
  from check_join_NF_sound[OF this] have "(t1,t2) \<in> ?Rel O ?Rel^-1" by auto
  then obtain s where t1: "(t1,s) \<in> ?Rel" and t2: "(t2,s) \<in> ?Rel" by auto
  show "\<exists> l' r' u. instance_rule (t1,t2) (l',r') \<and> (l',u) \<in> ?Rel \<and> (r', u) \<in> ?Rel"
  proof (intro exI conjI)
    show "instance_rule (t1,t2) (t1,t2)" using eq_rule_mod_var_refl[of "(t1,t2)"] unfolding eq_rule_mod_vars_def by simp
  qed (insert t1 t2, auto)
qed    

definition check_critical_pairs :: "('f :: show,string)rules \<Rightarrow> (bool \<times> ('f,string)rule) list \<Rightarrow> ('f,string)crit_pair_joins option \<Rightarrow> shows check"
  where "check_critical_pairs R cp joins_o \<equiv> case joins_o of Some joins \<Rightarrow>
            check_critical_pairs_guided R cp joins 
          | None \<Rightarrow> check_critical_pairs_NF R cp"

lemma check_critical_pairs: 
  fixes R :: "('f :: show, string)rules"
  assumes ok: "isOK(check_critical_pairs R (critical_pairs_impl R R) joins_o)"
  shows "WCR (rstep (set R))"
proof -
  note ok = ok[unfolded check_critical_pairs_def]
  show ?thesis
  proof (cases joins_o)
    case None
    show ?thesis
      by (rule check_critical_pairs_NF, insert None ok, auto)
  next
    case (Some joins)
    show ?thesis
      by (rule check_critical_pairs_guided, insert Some ok, auto)
  qed
qed

    
end


