(*
Copyright 2011 Christian Sternagel, René Thiemann

This file is part of IsaFoR/CeTA.

IsaFoR/CeTA is free software: you can redistribute it and/or modify it under the
terms of the GNU Lesser General Public License as published by the Free Software
Foundation, either version 3 of the License, or (at your option) any later
version.

IsaFoR/CeTA is distributed in the hope that it will be useful, but WITHOUT ANY
WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR A
PARTICULAR PURPOSE.  See the GNU Lesser General Public License for more details.

You should have received a copy of the GNU Lesser General Public License along
with IsaFoR/CeTA. If not, see <http://www.gnu.org/licenses/>.
*)
theory Linear_Poly_Carrier
imports
  SN_Order_Carrier
  Linear_Poly_Order
  Matrix_Poly 
begin

instantiation arctic :: "show" begin
fun shows_arctic :: "arctic \<Rightarrow> shows"
  where "shows_arctic(Num_arc i) = shows i"
  | "shows_arctic(MinInfty)  = shows_string ''-inf''"
definition "shows_prec (d::nat) ai = shows_arctic ai"
definition "shows_list (as::arctic list) \<equiv> shows_list_aux shows as"
lemma assoc_arctic: "shows_prec d (a::arctic) r @ s = shows_prec d a (r @ s)" by (cases a) (auto simp: shows_prec_arctic_def)
lemma assoc_arctic_elt: "\<forall>(a::arctic)\<in>set as. shows_prec d a r @ s = shows_prec d a (r @ s)"
  using assoc_arctic by auto
instance 
  by (intro_classes,rule assoc_arctic,unfold shows_list_arctic_def)
     (rule shows_list_aux_assoc[OF assoc_arctic_elt])
end

instantiation arctic_rat :: "show" begin
fun shows_arctic_rat :: "arctic_rat \<Rightarrow> shows"
  where "shows_arctic_rat (Num_arc_rat i) = shows i"
  | "shows_arctic_rat(MinInfty_rat)  = shows_string ''-inf''"
definition "shows_prec (d::nat) ari = shows_arctic_rat ari"
definition "shows_list (as::arctic_rat list) \<equiv> shows_list_aux shows as"
lemma assoc_arctic_rat: "shows_prec d (a::arctic_rat) r @ s = shows_prec d a (r @ s)" by (cases a) (auto simp: shows_prec_arctic_rat_def)
lemma assoc_arctic_rat_elt: "\<forall>(a::arctic_rat)\<in>set as. shows_prec d a r @ s = shows_prec d a (r @ s)"
  using assoc_arctic_rat by auto
instance 
  by (intro_classes,rule assoc_arctic_rat,unfold shows_list_arctic_rat_def)
     (rule shows_list_aux_assoc[OF assoc_arctic_rat_elt])
end

definition int_name :: "string" where "int_name = ''naturals with negative constants''"
definition rat_name :: "string" where "rat_name = ''rationals (with negative constants)''"
definition arctic_name :: "string" where "arctic_name = ''arctic semiring''"
definition arctic_rat_name :: "string" where "arctic_rat_name = ''arctic rationals''"


definition mat_mono :: "nat \<Rightarrow> ('a \<Rightarrow> bool) \<Rightarrow> ('a mat \<Rightarrow> bool)"
where "mat_mono sd m \<equiv> mat_monoI m sd"

definition int_components :: "int poly_c"
where "int_components \<equiv> \<lparr>
  poly_c.wf = (\<lambda>_. True),
  zero = 0,
  one = 1,
  default = 1,
  plus = op +,
  times = op *,
  max0 = max0,
  ge = op \<ge>,
  gt = op >,
  mono = True,
  description = int_name,
  arc_pos = (\<lambda> _ . True),
  check_mono = int_mono
  \<rparr>"

definition rat_components :: "rat poly_c"
where "rat_components \<equiv> \<lparr>
  poly_c.wf = (\<lambda>_. True),
  zero = 0,
  one = 1,
  default = 1,
  plus = op +,
  times = op *,
  max0 = max0,
  ge = op \<ge>,
  gt = op >,
  mono = True,
  description = rat_name,
  arc_pos = (\<lambda> _ . True),
  check_mono = rat_mono
  \<rparr>"

definition arctic_components :: "arctic poly_c"
where "arctic_components \<equiv> \<lparr>
  poly_c.wf = (\<lambda>_. True),
  zero = 0,
  one = 1,
  default = 1,
  plus = op +,
  times = op *,
  max0 = id,
  ge = ge,
  gt = gt_arctic,
  mono = False,
  description = arctic_name,
  arc_pos = pos_arctic,
  check_mono = (\<lambda> _. False)
  \<rparr>"

definition arctic_rat_components :: "arctic_rat poly_c"
where "arctic_rat_components \<equiv> \<lparr>
  poly_c.wf = (\<lambda>_. True),
  zero = 0,
  one = 1,
  default = 1,
  plus = op +,
  times = op *,
  max0 = id,
  ge = ge,
  gt = weak_gt_arctic_rat,
  mono = False,
  description = arctic_rat_name,
  arc_pos = pos_arctic_rat,
  check_mono = (\<lambda> _. False)
  \<rparr>"


lemma create_mono_poly_redtriple:
  fixes cI :: "('f :: show, 'a :: show)lpoly_interL \<Rightarrow> shows check"
  assumes carrier: "\<And> I as. isOK(cI I) \<Longrightarrow> \<exists> gt. 
    mono_poly_order_carrier (poly_c.wf c) (poly_c.zero c) (poly_c.plus c) (poly_c.times c) (poly_c.ge c) gt (poly_c.mono c) (poly_c.one c) (poly_c.default c) (poly_c.max0 c) (poly_c.arc_pos c) (poly_c.check_mono c) \<and> (\<forall>(a, b)\<in>set as. poly_c.wf c a \<longrightarrow> poly_c.wf c b \<longrightarrow> poly_c.gt c a b \<longrightarrow> gt a b)"
  shows "generic_redtriple_impl (create_poly_redtriple cI c)"
proof(rule create_poly_redtriple)
  fix I as
  assume "isOK(cI I)"
  from carrier[OF this, of as]
  obtain gt where orient: "\<forall>(a, b)\<in>set as. poly_c.wf c a \<longrightarrow> poly_c.wf c b \<longrightarrow> poly_c.gt c a b \<longrightarrow> gt a b"
    and mono: "
    mono_poly_order_carrier (poly_c.wf c) (poly_c.zero c) (poly_c.plus c) (poly_c.times c) (poly_c.ge c) gt (poly_c.mono c) (poly_c.one c) (poly_c.default c) (poly_c.max0 c) (poly_c.arc_pos c) (poly_c.check_mono c)" by auto
  show "\<exists> gt. poly_order_carrier 
  (poly_c.wf c)
  (poly_c.zero c)
  (poly_c.plus c)
  (poly_c.times c)
  (poly_c.ge c)
  gt
  (poly_c.mono c)
  (poly_c.one c)
  (poly_c.default c)
  (poly_c.max0 c)
  (poly_c.arc_pos c) \<and> 
  (\<forall>(a, b)\<in>set as. poly_c.wf c a \<longrightarrow> poly_c.wf c b \<longrightarrow> poly_c.gt c a b \<longrightarrow> gt a b) \<and>
  (poly_c.mono c \<longrightarrow>
    mono_poly_order_carrier (poly_c.wf c) (poly_c.zero c) (poly_c.plus c) (poly_c.times c) (poly_c.ge c) gt (poly_c.mono c) (poly_c.one c) (poly_c.default c) (poly_c.max0 c) (poly_c.arc_pos c) (poly_c.check_mono c))"
    by (intro exI, rule conjI[OF _ conjI[OF orient]], insert mono mono_poly_order_carrier.axioms(1)[OF mono], auto)
qed


abbreviation int_redtriple :: "('f :: show,int)lpoly_interL \<Rightarrow> ('f,'v :: show)redtriple" 
where "int_redtriple \<equiv> create_poly_redtriple (\<lambda>i. OK) int_components"

abbreviation arctic_redtriple :: "('f :: show,arctic)lpoly_interL \<Rightarrow> ('f,'v :: show)redtriple" 
where "arctic_redtriple \<equiv> create_poly_redtriple (\<lambda>i. OK) arctic_components"

abbreviation rat_redtriple :: "('f :: show,rat)lpoly_interL \<Rightarrow> ('f,'v :: show)redtriple" 
where "rat_redtriple \<equiv> create_poly_redtriple (\<lambda>i. OK) rat_components"

abbreviation arctic_rat_redtriple :: "('f :: show,arctic_rat)lpoly_interL \<Rightarrow> ('f,'v :: show)redtriple" 
where "arctic_rat_redtriple \<equiv> create_poly_redtriple (\<lambda>i. OK) arctic_rat_components"

locale mono_poly_interpretation = 
  fixes c :: "('a :: {show, max_ordered_semiring_1})poly_c"
 assumes c: "c = \<lparr>poly_c.wf = (\<lambda>x. True), 
  zero = 0,
  one = 1,
  default = poly_c.default c,
  plus = op +,
  times = op *,
  max0 = poly_c.max0 c,
  ge = ge,
  gt = poly_c.gt c,
  mono = True,
  description = poly_c.description c,
  arc_pos = (\<lambda> _. True),
  check_mono = poly_c.check_mono c\<rparr>"
  and carrier: "\<And> I as. \<exists> gt. mono_poly_order_carrier
  (poly_c.wf c)
  (poly_c.zero c)
  (poly_c.plus c)
  (poly_c.times c)
  (poly_c.ge c)
  gt
  (poly_c.mono c)
  (poly_c.one c)
  (poly_c.default c)
  (poly_c.max0 c)
  (poly_c.arc_pos c)
  (poly_c.check_mono c) \<and> (\<forall> (a,b) \<in> set as. poly_c.wf c a \<longrightarrow> poly_c.wf c b \<longrightarrow> poly_c.gt c a b \<longrightarrow> gt a b)"
 
sublocale mono_poly_interpretation \<subseteq> generic_redtriple_impl "create_poly_redtriple something c"
  by (rule create_mono_poly_redtriple, rule carrier)

lemma int_mono_poly_carrier: defines "c \<equiv> int_components" 
  shows "mono_poly_order_carrier (poly_c.wf c) (poly_c.zero c) (poly_c.plus c) (poly_c.times c) (poly_c.ge c) (poly_c.gt c) (poly_c.mono c) (poly_c.one c) (poly_c.default c) (poly_c.max0 c) (poly_c.arc_pos c) (poly_c.check_mono c)" 
proof -
  interpret mono_matrix_carrier int_mono "op >" 1 ..
  have id: "(op \<succeq> :: int \<Rightarrow> int \<Rightarrow> bool) = (\<lambda> x y. x \<ge> y)" using ge_int.simps by (intro ext, simp)
  from mono_poly_order_carrier
  have  mono: "mono_poly_order_carrier (\<lambda>_. True) 0 op + op * (\<lambda>x y. y \<le> x)
     (\<lambda>gt y. y < gt) True 1 1 max0 (\<lambda>_. True) int_mono" unfolding id .
  interpret mono_poly_order_carrier "(\<lambda>_. True)" 0 "op +" "op *" "(\<lambda>x y. y \<le> x)"
     "(\<lambda>gt y. y < gt)" True 1 1 max0 "(\<lambda>_. True)" int_mono by fact
  show ?thesis unfolding c_def int_components_def poly_c.simps ..
qed

interpretation int_mono_poly_carrier: mono_poly_interpretation int_components
  by (unfold_locales, simp add: int_components_def, intro ext, simp add: ge_int.simps, rule exI[of _ "poly_c.gt int_components"], rule conjI[OF int_mono_poly_carrier], auto) 
  
lemma int_redtriple: "generic_redtriple_impl int_redtriple" ..

lemma rat_mono_poly_carrier: defines "c \<equiv> rat_components" 
  shows "\<exists> gt. mono_poly_order_carrier (poly_c.wf c) (poly_c.zero c) (poly_c.plus c) (poly_c.times c) (poly_c.ge c) gt (poly_c.mono c) (poly_c.one c) (poly_c.default c) (poly_c.max0 c) (poly_c.arc_pos c) (poly_c.check_mono c) \<and> (\<forall> (a,b) \<in> set ab. poly_c.wf c a \<longrightarrow> poly_c.wf c b \<longrightarrow> poly_c.gt c a b \<longrightarrow> gt a b)" 
proof -
  let ?ab = "filter (\<lambda> (a,b). a > b) ab"
  have "\<forall> a b. (a,b) \<in> set ?ab \<longrightarrow> a > b" by auto
  from weak_rat_SN.weak_gt_mono[OF this]
  obtain gt where mono: "SN_strict_mono_ordered_semiring_1 1 gt rat_mono" 
    and orient: "\<forall> a b. (a,b) \<in> set ?ab \<longrightarrow> gt a b" by auto
  from orient have orient: "\<forall> (a,b) \<in> set ab. poly_c.wf c a \<longrightarrow> poly_c.wf c b \<longrightarrow> a > b \<longrightarrow> gt a b" by auto
  from mono interpret SN_strict_mono_ordered_semiring_1 1 gt rat_mono .
  interpret mono_matrix_carrier rat_mono gt 1 ..
  have id: "(op \<succeq> :: rat \<Rightarrow> rat \<Rightarrow> bool) = (\<lambda> x y. x \<ge> y)" using ge_rat.simps by (intro ext, simp)
  from mono_poly_order_carrier
  have  mono: "mono_poly_order_carrier (\<lambda>_. True) 0 op + op * (\<lambda>x y. y \<le> x)
    gt True 1 1 max0 (\<lambda>_. True) rat_mono" unfolding id .
  interpret mono_poly_order_carrier "(\<lambda>_. True)" 0 "op +" "op *" "(\<lambda>x y. y \<le> x)"
     gt True 1 1 max0 "(\<lambda>_. True)" rat_mono by fact
  show ?thesis unfolding c_def rat_components_def poly_c.simps 
    by (intro exI, rule conjI[OF _ orient[unfolded c_def rat_components_def poly_c.simps]], unfold_locales)
qed

interpretation rat_mono_poly_carrier: mono_poly_interpretation rat_components
  by (unfold_locales, simp add: rat_components_def, intro ext, simp add: ge_rat.simps, 
    rule rat_mono_poly_carrier)

lemma rat_redtriple: "generic_redtriple_impl rat_redtriple" ..

locale poly_interpretation = 
  fixes c :: "('a :: {show, max_ordered_semiring_1}) poly_c"
 assumes c: "c = \<lparr>poly_c.wf = (\<lambda>x. True), 
  zero = 0,
  one = 1,
  default = poly_c.default c,
  plus = op +,
  times = op *,
  max0 = id,
  ge = ge,
  gt = poly_c.gt c,
  mono = False,
  description = poly_c.description c,
  arc_pos = poly_c.arc_pos c, 
  check_mono = (\<lambda> _. False)\<rparr>"
  and zero_leastIII: "\<And> x :: 'a. ge x 0"
  and carrier: "\<And> I as. \<exists> gt. poly_order_carrier
  (poly_c.wf c)
  (poly_c.zero c)
  (poly_c.plus c)
  (poly_c.times c)
  (poly_c.ge c)
  gt
  (poly_c.mono c)
  (poly_c.one c)
  (poly_c.default c)
  (poly_c.max0 c)
  (poly_c.arc_pos c) \<and> (\<forall> (a,b) \<in> set as. poly_c.wf c a \<longrightarrow> poly_c.wf c b \<longrightarrow> poly_c.gt c a b \<longrightarrow> gt a b)"

lemma (in poly_interpretation) c_mono: "(a \<and> (poly_c.mono c \<longrightarrow> b)) = a"
proof -
  have "poly_c.mono c = False" using arg_cong[OF c, of "poly_c.mono"] by simp
  thus ?thesis by simp
qed

sublocale poly_interpretation \<subseteq> generic_redtriple_impl "create_poly_redtriple something c"
  by (rule create_poly_redtriple, unfold c_mono, rule carrier)

lemma arctic_poly_carrier: defines "c \<equiv> arctic_components" 
  shows "poly_order_carrier (poly_c.wf c) (poly_c.zero c) (poly_c.plus c) (poly_c.times c) (poly_c.ge c) (poly_c.gt c) (poly_c.mono c) (poly_c.one c) (poly_c.default c) (poly_c.max0 c) (poly_c.arc_pos c)"
proof -
  interpret both_mono_matrix_carrier pos_arctic gt_arctic 1 ..
  from linear_poly_order_carrier
  have  carrier: "poly_order_carrier (\<lambda>_. True) 0 op + op * op \<succeq>
     gt_arctic False 1 1 id pos_arctic" .
  thus ?thesis unfolding c_def arctic_components_def poly_c.simps .
qed

interpretation arctic_poly_carrier: poly_interpretation arctic_components
  by (unfold_locales, simp add: arctic_components_def, simp add: zero_arctic_def,
  rule exI[of _ "poly_c.gt arctic_components"], rule conjI[OF arctic_poly_carrier], auto) 

lemma arctic_redtriple: "generic_redtriple_impl arctic_redtriple" ..

lemma arctic_rat_poly_carrier: defines "c \<equiv> arctic_rat_components" 
  shows "\<exists> gt. poly_order_carrier (poly_c.wf c) (poly_c.zero c) (poly_c.plus c) (poly_c.times c) (poly_c.ge c) gt (poly_c.mono c) (poly_c.one c) (poly_c.default c) (poly_c.max0 c) (poly_c.arc_pos c) \<and> (\<forall> (a,b) \<in> set ab. poly_c.wf c a \<longrightarrow> poly_c.wf c b \<longrightarrow> poly_c.gt c a b \<longrightarrow> gt a b)" 
proof -
  let ?ab = "filter (\<lambda> (a,b). weak_gt_arctic_rat a b) ab"
  have "\<forall> a b. (a,b) \<in> set ?ab \<longrightarrow> weak_gt_arctic_rat a b" by auto
  from weak_arctic_rat_SN.weak_gt_both_mono[OF this]
  obtain gt where mono: "SN_both_mono_ordered_semiring_1 1 gt pos_arctic_rat"
    and orient: "\<forall> a b. (a,b) \<in> set ?ab \<longrightarrow> gt a b" by auto
  from orient have orient: "\<forall> (a,b) \<in> set ab. poly_c.wf c a \<longrightarrow> poly_c.wf c b \<longrightarrow> weak_gt_arctic_rat a b \<longrightarrow> gt a b" by auto
  from mono interpret SN_both_mono_ordered_semiring_1 1 gt pos_arctic_rat .
  interpret both_mono_matrix_carrier pos_arctic_rat gt 1 ..
  from linear_poly_order_carrier
  have  carrier: "poly_order_carrier (\<lambda>_. True) 0 op + op * op \<succeq>
     gt False 1 1 id pos_arctic_rat" .
  thus ?thesis using orient unfolding c_def arctic_rat_components_def poly_c.simps by auto
qed

interpretation arctic_rat_poly_carrier: poly_interpretation arctic_rat_components
  by (unfold_locales, simp add: arctic_rat_components_def, simp add: zero_arctic_rat_def,
  rule arctic_rat_poly_carrier)

lemma arctic_rat_redtriple: "generic_redtriple_impl arctic_rat_redtriple" ..

abbreviation mat_components :: "nat \<Rightarrow> nat \<Rightarrow> ('a :: max_ordered_semiring_1) poly_c \<Rightarrow> 'a mat poly_c"
where "mat_components n sd c \<equiv> 
  \<lparr> poly_c.wf = mat n n, 
    zero = mat0I 0 n n, 
    one = mat1I 0 1 n, 
    default = mat1I 0 (poly_c.default c) n, 
    plus = mat_plusI (op +), 
    times = mat_multI 0 (op +) (op *) n,
    max0 = mat_max0, 
    ge = mat_ge, 
    gt = mat_gtI (poly_c.gt c) sd, 
    mono = poly_c.mono c, 
    description = concat [to_string n, ''x'',to_string n,'' matrices with strict dimension '',to_string sd,'' over '',poly_c.description c], 
    arc_pos = (\<lambda>x. True), 
    check_mono = mat_mono sd (poly_c.check_mono c) \<rparr>"

abbreviation mat_check :: "nat \<Rightarrow> nat \<Rightarrow> shows check"
where "mat_check n sd \<equiv> check (sd > 0 \<and> sd \<le> n) (shows_string ''invalid dimension'')"

lemma (in mono_poly_interpretation) id_c: 
    "poly_c.wf c = (\<lambda>x. True) \<and> poly_c.zero c = 0 \<and> poly_c.one c = 1 \<and> poly_c.arc_pos c = (\<lambda>x. True) \<and> poly_c.plus c = op +
     \<and> poly_c.times c = op * \<and> poly_c.ge c = ge \<and> poly_c.mono c = True" 
  using c[symmetric] by (cases c, auto)


sublocale mono_poly_interpretation \<subseteq> mono_matrix: generic_redtriple_impl "create_poly_redtriple (\<lambda>I. mat_check n sd) (mat_components n sd c)"
proof (rule create_poly_redtriple, simp_all add: id, clarify)
  fix as :: "('a mat \<times> 'a mat) list"
  assume sdpos: "0 < sd" and sd_n: "sd \<le> n"
  let ?goal = "\<exists>gta. Linear_Polynomial.poly_order_carrier (mat n n) (mat0 n n)
                   mat_plus (mat_mult n) mat_ge gta (poly_c.mono c) (mat1 n)
                   (mat1I (0\<Colon>'a) (poly_c.default c) n) mat_max0 (\<lambda>x. True) \<and>
                  (\<forall>(a, b)\<in>set as.
                      mat n n a \<longrightarrow> mat n n b \<longrightarrow> mat_gtI (poly_c.gt c) sd a b \<longrightarrow> gta a b) \<and>
                  (poly_c.mono c \<longrightarrow>
                   mono_poly_order_carrier (mat n n) (mat0 n n) mat_plus
                    (mat_mult n) mat_ge gta True (mat1 n)
                    (mat1I (0\<Colon>'a) (poly_c.default c) n) mat_max0 (\<lambda>x. True)
                    (mat_mono sd (poly_c.check_mono c)))"
  interpret weak_SN_strict_mono_ordered_semiring_1 "poly_c.gt c" "poly_c.default c" "poly_c.check_mono c"
  proof
    fix as :: "('a \<times> 'a)list"
    assume orient1: "\<forall> a b. (a,b) \<in> set as \<longrightarrow> poly_c.gt c a b"
    from carrier[of as, simplified id_c] obtain gt 
      where mono: "mono_poly_order_carrier (\<lambda> _. True) 0 (op +) (op *) (op \<succeq>) gt True 1 (poly_c.default c) (poly_c.max0 c) (\<lambda> _. True) (poly_c.check_mono c)" and orient2: "\<forall> (a,b) \<in> set as. poly_c.gt c a b \<longrightarrow> gt a b" by blast
    from orient1 orient2 have orient: "\<forall> a b. (a,b) \<in> set as \<longrightarrow> gt a b" (is "?orient gt") by auto
    show "\<exists> gt. SN_strict_mono_ordered_semiring_1 (poly_c.default c) gt (poly_c.check_mono c) \<and> ?orient gt"
    proof (intro exI, rule conjI[OF _ orient])
      interpret mono_poly_order_carrier "\<lambda> _. True" 0 "op +" "op *" "op \<succeq>" gt True 1 "poly_c.default c" "poly_c.max0 c" "\<lambda> _. True" "poly_c.check_mono c" by fact
      show "SN_strict_mono_ordered_semiring_1 (poly_c.default c) gt (poly_c.check_mono c)" 
        by (unfold_locales, 
        (rule compat, simp+)[1],
        (rule compat2, simp+)[1],
        (rule gt_trans, simp+)[1],
          rule default_geq_zero,
        (rule plus_gt_left_mono, simp+)[1],
          rule default_gt_zero,
        (insert SN, simp),
          rule check_mono, simp+
      ) 
    qed
  qed
  let ?as = "filter (\<lambda> (a,b). mat n n a \<and> mat n n b \<and> mat_gtI (poly_c.gt c) sd a b) as"
  have as: "\<And> a b. mat n n a \<and> mat n n b \<and> (a,b) \<in> set ?as \<Longrightarrow> mat_gtI (poly_c.gt c) sd a b" by force
  from weak_mat_gt_mono[OF sd_n, of ?as, OF this]
  obtain gt where carrier: "SN_strict_mono_ordered_semiring_1 (poly_c.default c) gt (poly_c.check_mono c)"
    and "\<forall> a b. mat n n a \<and> mat n n b \<and> (a,b) \<in> set ?as \<longrightarrow> mat_gtI gt sd a b" by blast
  hence orient: "\<forall> (a,b) \<in> set as. mat n n a \<longrightarrow> mat n n b \<longrightarrow> mat_gtI (poly_c.gt c) sd a b \<longrightarrow> mat_gtI gt sd a b" by auto
  interpret SN_strict_mono_ordered_semiring_1 "poly_c.default c" gt "poly_c.check_mono c" by fact
  interpret foo: mono_matrix_carrier "poly_c.check_mono c" gt "poly_c.default c" ..
  from foo.mat_mono_poly_order_carrier[OF sd_n sdpos] 
  interpret bar: mono_poly_order_carrier "(mat n n)" "(mat0 n n)" mat_plus "(mat_mult n)" mat_ge "(mat_gtI gt sd)" True "(mat1 n)" "(mat1I 0 (poly_c.default c) n)" mat_max0 "(\<lambda>x. True)" "(mat_monoI (poly_c.check_mono c) sd)" .
  show ?goal
    by (rule exI[of _ "mat_gtI gt sd"], intro conjI, simp add: id_c, unfold_locales, rule orient,
      intro impI, unfold_locales, 
      simp, 
      (rule mat_mono[OF sd_n], auto simp: mat_mono_def)[1],
      rule mat_default_gt_mat0[OF sdpos sd_n])
qed

abbreviation int_mat_redtriple :: "nat \<Rightarrow> nat \<Rightarrow> ('f :: show,int mat)lpoly_interL \<Rightarrow> ('f,'v :: show)redtriple" 
where "int_mat_redtriple n sd \<equiv> create_poly_redtriple (\<lambda>i. mat_check n sd) (mat_components n sd int_components)"

abbreviation rat_mat_redtriple :: "nat \<Rightarrow> nat \<Rightarrow> ('f :: show,rat mat)lpoly_interL \<Rightarrow> ('f,'v :: show)redtriple" 
where "rat_mat_redtriple n sd \<equiv> create_poly_redtriple (\<lambda>i. mat_check n sd) (mat_components n sd rat_components)"

lemma int_mat_redtriple: "generic_redtriple_impl (int_mat_redtriple n sd)" ..
lemma rat_mat_redtriple: "generic_redtriple_impl (rat_mat_redtriple n sd)" ..


abbreviation mat_arctic_components :: "nat \<Rightarrow> ('a :: max_ordered_semiring_1) poly_c \<Rightarrow> 'a mat poly_c"
where "mat_arctic_components n c \<equiv> 
  \<lparr> poly_c.wf = mat n n, 
    zero = mat0I 0 n n, 
    one = mat1I 0 1 n, 
    default = mat1I 0 (poly_c.default c) n, 
    plus = mat_plusI (op +), 
    times = mat_multI 0 (op +) (op *) n,
    max0 = id, 
    ge = mat_ge, 
    gt = mat_comp_all (poly_c.gt c), 
    mono = False, 
    description = concat [to_string n, ''x'',to_string n,'' matrices over '',poly_c.description c], 
    arc_pos = mat_arc_posI (poly_c.arc_pos c),
    check_mono = (\<lambda> _. False) \<rparr>"

abbreviation mat_arctic_check :: "nat \<Rightarrow> shows check"
where "mat_arctic_check n \<equiv> check (n > 0) (shows_string ''invalid dimension'')"



sublocale poly_interpretation \<subseteq> matrix: generic_redtriple_impl "create_poly_redtriple (\<lambda>I. mat_arctic_check n) (mat_arctic_components n c)"
proof (rule create_poly_redtriple, simp_all)
  fix as
  assume n_pos: "0 < n"
  let ?goal = "\<exists>gta. Linear_Polynomial.poly_order_carrier (mat n n) (mat0 n n)
                mat_plus (mat_mult n) mat_ge gta False (mat1 n)
                (mat1I (0\<Colon>'a) (poly_c.default c) n) id
                (mat_arc_posI (poly_c.arc_pos c)) \<and>
               (\<forall>(a, b)\<in>set as.
                   mat n n a \<longrightarrow>
                   mat n n b \<longrightarrow> mat_comp_all (poly_c.gt c) a b \<longrightarrow> gta a b)"
  from c[symmetric] have id: 
    "poly_c.wf c = (\<lambda>x. True) \<and> poly_c.zero c = 0 \<and> poly_c.one c = 1 \<and> poly_c.max0 c = id \<and> poly_c.plus c = op +
     \<and> poly_c.times c = op * \<and> poly_c.ge c = ge \<and> poly_c.mono c = False" by (cases c, auto)
  interpret weak_SN_both_mono_ordered_semiring_1 "poly_c.gt c" "poly_c.default c" "poly_c.arc_pos c"
  proof
    fix as :: "('a \<times> 'a)list"
    assume orient1: "\<forall> a b. (a,b) \<in> set as \<longrightarrow> poly_c.gt c a b"
    from carrier[of as, simplified id] obtain gt 
      where mono: "poly_order_carrier (\<lambda> _. True) 0 (op +) (op *) (op \<succeq>) gt False 1 (poly_c.default c) id (poly_c.arc_pos c)" and orient2: "\<forall> (a,b) \<in> set as. poly_c.gt c a b \<longrightarrow> gt a b" by blast
    from orient1 orient2 have orient: "\<forall> a b. (a,b) \<in> set as \<longrightarrow> gt a b" (is "?orient gt") by auto
    show "\<exists> gt. SN_both_mono_ordered_semiring_1 (poly_c.default c) gt (poly_c.arc_pos c) \<and> ?orient gt"
    proof (intro exI, rule conjI[OF _ orient])
      interpret poly_order_carrier "\<lambda> _. True" 0 "op +" "op *" "op \<succeq>" gt False 1 "poly_c.default c" id "poly_c.arc_pos c" by fact
      show "SN_both_mono_ordered_semiring_1 (poly_c.default c) gt (poly_c.arc_pos c)" 
        by (unfold_locales,
        (rule compat, simp+)[1],
        (rule compat2, simp+)[1],
        (rule gt_trans, simp+)[1],
          rule default_geq_zero,
        (rule plus_gt_both_mono, simp+)[1],
        (rule times_gt_left_mono, simp+)[1],
        (rule zero_leastI, simp+)[1],
        (rule zero_leastII, simp+)[1],
          rule zero_leastIII,
          rule SN[simplified zero_leastIII, simplified],
          rule arc_pos_plus[simplified], simp+,
          rule arc_pos_mult[simplified], simp+,
          rule arc_pos_default)
    qed
  qed
  let ?as = "filter (\<lambda> (a,b). mat n n a \<and> mat n n b \<and> mat_comp_all (poly_c.gt c) a b) as"
  have as: "\<And> a b. mat n n a \<and> mat n n b \<and> (a,b) \<in> set ?as \<Longrightarrow> mat_comp_all (poly_c.gt c) a b" by force
  from weak_mat_gt_both_mono[OF as]
  obtain gt where carrier: "SN_both_mono_ordered_semiring_1 (poly_c.default c) gt (poly_c.arc_pos c)"
    and "\<forall> a b. mat n n a \<and> mat n n b \<and> (a,b) \<in> set ?as \<longrightarrow> mat_comp_all gt a b" by blast
  hence orient: "\<forall> (a,b) \<in> set as. mat n n a \<longrightarrow> mat n n b \<longrightarrow> mat_comp_all (poly_c.gt c) a b \<longrightarrow> mat_comp_all gt a b" by auto
  interpret SN_both_mono_ordered_semiring_1 "poly_c.default c" gt "poly_c.arc_pos c" by fact
  interpret foo: both_mono_matrix_carrier "poly_c.arc_pos c" gt "poly_c.default c" ..
  from foo.mat_poly_order_carrier[OF n_pos] 
  interpret bar: poly_order_carrier "(mat n n)" "(mat0 n n)" mat_plus "(mat_mult n)" mat_ge "(mat_comp_all gt)" False "(mat1 n)" "(mat1I 0 (poly_c.default c) n)" id "(mat_arc_posI (poly_c.arc_pos c))" .
  show ?goal
    by (rule exI[of _ "mat_comp_all gt"], intro conjI[OF _ orient], unfold_locales) 
qed
  
abbreviation arctic_mat_redtriple :: "nat \<Rightarrow> ('f :: show,arctic mat)lpoly_interL \<Rightarrow> ('f,'v :: show)redtriple" 
where "arctic_mat_redtriple n \<equiv> create_poly_redtriple (\<lambda>i. mat_arctic_check n) (mat_arctic_components n arctic_components)"

abbreviation arctic_rat_mat_redtriple :: "nat \<Rightarrow> ('f :: show,arctic_rat mat)lpoly_interL \<Rightarrow> ('f,'v :: show)redtriple" 
where "arctic_rat_mat_redtriple n \<equiv> create_poly_redtriple (\<lambda>i. mat_arctic_check n) (mat_arctic_components n arctic_rat_components)"


lemma arctic_mat_redtriple: "generic_redtriple_impl (arctic_mat_redtriple n)" ..
lemma arctic_rat_mat_redtriple: "generic_redtriple_impl (arctic_rat_mat_redtriple n)" ..



end
