(*
Copyright 2011 Christian Sternagel, René Thiemann

This file is part of IsaFoR/CeTA.

IsaFoR/CeTA is free software: you can redistribute it and/or modify it under the
terms of the GNU Lesser General Public License as published by the Free Software
Foundation, either version 3 of the License, or (at your option) any later
version.

IsaFoR/CeTA is distributed in the hope that it will be useful, but WITHOUT ANY
WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR A
PARTICULAR PURPOSE.  See the GNU Lesser General Public License for more details.

You should have received a copy of the GNU Lesser General Public License along
with IsaFoR/CeTA. If not, see <http://www.gnu.org/licenses/>.
*)
theory Linear_Poly_Order
imports
  Term_Order_Impl
  Term_Impl
  Linear_Polynomial
  Name
begin

text {* we first head for polynomial interpretations *}

type_synonym ('f,'a)lpoly_inter = "'f \<Rightarrow> nat \<Rightarrow> ('a \<times> 'a list)"

declare size_simps[simp]
fun eval_termI :: "'a \<Rightarrow> ('a \<Rightarrow> 'a \<Rightarrow> 'a) \<Rightarrow> ('a \<Rightarrow> 'a \<Rightarrow> 'a) \<Rightarrow> ('a \<Rightarrow> 'a) \<Rightarrow> ('f,'a)lpoly_inter \<Rightarrow> ('v,'a)p_ass \<Rightarrow> ('f,'v)term \<Rightarrow> 'a" 
where "eval_termI _ _ _ _ pi \<alpha> (Var x) = \<alpha> x" 
   |  "eval_termI ze pl ti ma pi \<alpha> (Fun f ts) = (let (a,as) = pi f (length ts) in ma (pl a  (list_sumI ze pl (map (\<lambda> at. ti (fst at) (eval_termI ze pl ti ma pi \<alpha> (snd at))) (zip as ts)))))"
declare size_simps[simp del]


context poly_order_carrier
begin
declare poly_simps[simp] poly_order_simps[simp]

abbreviation eval_term :: "('f,'a)lpoly_inter \<Rightarrow> ('v,'a)p_ass \<Rightarrow> ('f,'v)term \<Rightarrow> 'a" ("_\<guillemotleft>_,_\<guillemotright>" 50)
where "eval_term \<equiv> eval_termI zero plus times max0"

abbreviation wf_lpoly_inter :: "('f,'a)lpoly_inter \<Rightarrow> bool"
where "wf_lpoly_inter pi \<equiv> (\<forall> f n. wf_elem (fst (pi f n)) \<and> (\<forall> a. a \<in> set (snd (pi f n)) \<longrightarrow> wf_elem a \<and> geq a zero))
                         \<and> (\<forall> f n. arc_pos (fst (pi f n)) \<or> (\<exists> a \<in> set (take n (snd (pi f n))) . arc_pos a))"
end

locale poly_order = poly_order_carrier +
  fixes pi :: "('f::show,'a)lpoly_inter"
  assumes wf_pi: "wf_lpoly_inter pi"
begin

lemma wf_terms[simp]: assumes wf_ass: "wf_ass \<alpha>"
  shows "wf_elem (pi\<guillemotleft>\<alpha>,t\<guillemotright>)"
proof (induct t)
  case (Var x)
  from wf_ass have "wf_elem (\<alpha> x)" unfolding wf_ass_def by auto
  thus ?case by auto
next
  case (Fun f ts)
  show ?case 
  proof (cases "pi f (length ts)")
    case (Pair a as)
    have "wf_elem (list_sum (map (\<lambda>at. times (fst at) (pi\<guillemotleft>\<alpha>,snd at\<guillemotright>)) (zip as ts)))" (is "wf_elem ?ls")
    proof (rule wf_list_sum, rule ballI)
      fix p
      assume "p \<in> set (map (\<lambda>at. fst at * (pi\<guillemotleft>\<alpha>,snd at\<guillemotright>)) (zip as ts))" (is "p \<in> set (map _ ?zip)")
      from this obtain a' t where p: "p = a' * (pi\<guillemotleft>\<alpha>,t\<guillemotright>)" and ab: "(a',t) \<in> set ?zip" by auto
      from ab have a': "a' \<in> set as" and t: "t \<in> set ts" using set_zip_leftD[where x = a' and y = t] set_zip_rightD[where y = t and x = a'] by auto
      from a' Pair have "a' \<in> set (snd (pi f (length ts)))" by auto
      with Fun wf_pi have "wf_elem a'" by auto
      with Fun t p show "wf_elem p" by auto
    qed
    from wf_pi have "wf_elem (fst (pi f (length ts)))" by auto
    with Pair `wf_elem ?ls` have "wf_elem (a + ?ls)" by auto
    thus ?thesis using Pair by (auto simp: wf_max0 max0_pos)
  qed
qed

lemma pos_term[simp]: assumes wf_ass: "wf_ass \<alpha>"
  and pos_ass: "pos_ass \<alpha>"
  shows "geq (pi\<guillemotleft>\<alpha>,t\<guillemotright>) zero"
proof (induct t)
  case (Var x)
  from wf_ass pos_ass have "wf_elem (\<alpha> x) \<and> geq (\<alpha> x) zero" unfolding wf_ass_def pos_ass_def by auto
  thus ?case by auto
next
  case (Fun f ts)
  show ?case 
  proof (cases "pi f (length ts)")
    case (Pair a as)
    have "wf_elem (list_sum (map (\<lambda>at. times (fst at) (pi\<guillemotleft>\<alpha>,snd at\<guillemotright>)) (zip as ts)))" (is "wf_elem ?ls")
    proof (rule wf_list_sum, rule ballI)
      fix p
      assume "p \<in> set (map (\<lambda>at. fst at * (pi\<guillemotleft>\<alpha>,snd at\<guillemotright>)) (zip as ts))" (is "p \<in> set (map _ ?zip)")
      from this obtain a' t where p: "p = a' * (pi\<guillemotleft>\<alpha>,t\<guillemotright>)" and ab: "(a',t) \<in> set ?zip" by auto
      from ab have a': "a' \<in> set as" and t: "t \<in> set ts" using set_zip_leftD[where x = a' and y = t] set_zip_rightD[where y = t and x = a'] by auto
      from a' Pair have "a' \<in> set (snd (pi f (length ts)))" by auto
      with Fun wf_pi have "wf_elem a'" by auto
      with Fun t p wf_ass wf_pi show "wf_elem p" by auto
    qed
    from wf_pi have "wf_elem (fst (pi f (length ts)))" by auto
    with Pair `wf_elem ?ls` have "wf_elem (a + ?ls)" by auto
    thus ?thesis using Pair by (auto simp: wf_max0 max0_pos)
  qed
qed

definition apos_ass :: "('v,'a)p_ass \<Rightarrow> bool" where "apos_ass \<alpha> \<equiv> \<forall> x. arc_pos (\<alpha> x)"

lemma apos_term[simp]: assumes wf_ass: "wf_ass \<alpha>"
  and apos_ass: "apos_ass \<alpha>"
  shows "arc_pos (pi\<guillemotleft>\<alpha>,t\<guillemotright>)"
proof (induct t)
  case (Var x)
  from apos_ass show ?case unfolding apos_ass_def by auto
next
  case (Fun f ts)
  show ?case 
  proof (cases "pi f (length ts)")
    case (Pair a as)
    have wf_args: "\<forall>a\<in>set (map (\<lambda>at. fst at * (pi\<guillemotleft>\<alpha>,snd at\<guillemotright>)) (zip as ts)). wf_elem a"
    proof (rule ballI)
      fix p
      assume "p \<in> set (map (\<lambda>at. fst at * (pi\<guillemotleft>\<alpha>,snd at\<guillemotright>)) (zip as ts))" (is "p \<in> set (map _ ?zip)")
      from this obtain a' t where p: "p = a' * (pi\<guillemotleft>\<alpha>,t\<guillemotright>)" and ab: "(a',t) \<in> set ?zip" by auto
      from ab have a': "a' \<in> set as" and t: "t \<in> set ts" using set_zip_leftD[where x = a' and y = t] set_zip_rightD[where y = t and x = a'] by auto
      from a' Pair have "a' \<in> set (snd (pi f (length ts)))" by auto
      with Fun wf_pi have "wf_elem a'" by auto
      with Fun t p wf_ass wf_pi show "wf_elem p" by auto
    qed
    have "wf_elem (list_sum (map (\<lambda>at. times (fst at) (pi\<guillemotleft>\<alpha>,snd at\<guillemotright>)) (zip as ts)))" (is "wf_elem ?ls")
      by (rule wf_list_sum, rule wf_args)
    from wf_pi have "wf_elem (fst (pi f (length ts)))" by auto
    with Pair have wf_a: "wf_elem a" by auto
    with wf_pi have "arc_pos (fst (pi f (length ts))) \<or> (\<exists> a \<in> set (take (length ts) (snd (pi f (length ts)))). arc_pos a)" (is "?ac \<or> ?aa") by auto
    thus ?thesis 
    proof
      assume "?ac"
      hence "arc_pos (a + ?ls)" and "wf_elem (a + ?ls)" using Pair `wf_elem ?ls` wf_a arc_pos_plus by auto
      thus ?thesis using Pair by auto
    next
      assume ?aa
      from this obtain c where c: "c \<in> set (take (length ts) as) \<and> arc_pos c" using Pair by auto
      from wf_pi have "(\<forall> b \<in> set (snd (pi f (length ts))). wf_elem b)" by auto
      with Pair have wf_as: "\<forall> b \<in> set as. wf_elem b" by auto
      from Pair Fun have rec: "\<forall> t \<in> set ts. arc_pos (pi\<guillemotleft>\<alpha>,t\<guillemotright>)" by auto
      from c wf_as rec wf_args have aa: "arc_pos ?ls"
      proof (induct ts arbitrary: as)
	case (Cons t ts) note oCons = this
	show ?case
	proof (cases as)
	  case Nil
	  with Cons show ?thesis by auto
	next
	  case (Cons b bs)
	  let ?prod = "b * (pi\<guillemotleft>\<alpha>,t\<guillemotright>)"
	  have wf: "wf_elem (pi\<guillemotleft>\<alpha>,t\<guillemotright>) \<and> wf_elem b \<and> wf_elem ?prod" using wf_ass Cons oCons by auto
	  hence "wf_elem ?prod" by auto
	  show ?thesis 
	  proof (cases "b = c")
	    case True
	    show ?thesis  by (simp add: Cons, rule arc_pos_plus, rule arc_pos_mult, simp add: True c, auto simp: Cons oCons wf)
	  next
	    case False
	    with Cons oCons have "c \<in> set (take (length ts) bs)" by auto
	    with Cons oCons have ap: "arc_pos (list_sum (map (\<lambda> at. fst at * (pi\<guillemotleft>\<alpha>,snd at\<guillemotright>)) (zip bs ts)))" (is "arc_pos ?ls") by auto
	    with Cons oCons have "wf_elem ?ls" by auto
	    show ?thesis by (simp add: Cons, simp only: plus_comm[OF `wf_elem ?prod` `wf_elem ?ls`], rule arc_pos_plus[OF ap], auto simp: Cons oCons wf)
	  qed
	qed 
      qed simp
      have "a + ?ls = ?ls + a" by (rule plus_comm[OF wf_a `wf_elem ?ls`])
      hence "arc_pos (a + ?ls)" and "wf_elem (a + ?ls)" using aa `wf_elem ?ls` wf_a arc_pos_plus  by auto
      thus ?thesis using Pair by auto
    qed
  qed
qed

definition inter_s :: "('f,'v)trs"
  where "inter_s \<equiv> {(s,t). (\<forall> \<alpha> . wf_ass \<alpha> \<and> pos_ass \<alpha> \<and> apos_ass \<alpha> \<longrightarrow> gt (pi\<guillemotleft>\<alpha>,s\<guillemotright>) (pi\<guillemotleft>\<alpha>,t\<guillemotright>))}"

definition inter_ns :: "('f,'v)trs"
  where "inter_ns \<equiv> {(s,t). (\<forall> \<alpha> . wf_ass \<alpha> \<and> pos_ass \<alpha> \<and> apos_ass \<alpha> \<longrightarrow> geq (pi\<guillemotleft>\<alpha>,s\<guillemotright>) (pi\<guillemotleft>\<alpha>,t\<guillemotright>))}"

definition default_ass :: "('v,'a)p_ass" where "default_ass x \<equiv> default"

lemma wf_pos_apos_default_ass[simp]: "wf_ass default_ass \<and> pos_ass default_ass \<and> apos_ass default_ass"
  unfolding wf_ass_def pos_ass_def apos_ass_def default_ass_def
  by auto

lemma interpr_subst: assumes wf_ass: "wf_ass \<alpha>"
  shows "(pi\<guillemotleft>\<alpha>, t \<cdot> \<sigma>\<guillemotright>) = (pi\<guillemotleft>(\<lambda> x. pi\<guillemotleft>\<alpha>, \<sigma> x\<guillemotright>),t\<guillemotright>)" 
proof -
  have "(pi\<guillemotleft>\<alpha>, t \<cdot> \<sigma>\<guillemotright>) = (pi\<guillemotleft>(\<lambda> x. pi\<guillemotleft>\<alpha>, \<sigma> x\<guillemotright>),t\<guillemotright>)" (is "_ = (pi\<guillemotleft>?B, t\<guillemotright>)")
  proof (induct t)
    case (Var x)
    thus ?case by simp
  next
    case (Fun f ss)
    let ?ts = "map (\<lambda> s. s \<cdot> \<sigma>) ss"
    have wf_assB: "wf_ass ?B" unfolding wf_ass_def using wf_ass wf_pi by auto
    have map: "(map (eval_term pi \<alpha>) (?ts))
      = (map (eval_term pi ?B) ss)" using Fun by (induct ss, auto)
    let ?i = "pi f (length ss)"
    show ?case 
    proof (cases "?i")
      case (Pair c cs)
      from wf_pi have "\<forall> d \<in> set (snd ?i). wf_elem d" by auto
      with Pair have wf_i: "\<forall> d \<in> set cs. wf_elem d" by auto
      from wf_i map have "map (\<lambda>at. fst at * (pi\<guillemotleft>\<alpha>,snd at\<guillemotright>)) (zip cs (map (\<lambda>t. t \<cdot> \<sigma>) ss)) = 
	map (\<lambda>at. fst at * (pi\<guillemotleft>\<lambda>x. pi\<guillemotleft>\<alpha>, \<sigma> x\<guillemotright>,snd at\<guillemotright>)) (zip cs ss)"
      proof (induct cs arbitrary: ss)
	case (Cons c cs) note oCons = this
	thus ?case 
	proof (cases ss, simp)
	  case (Cons s sss)
	  with oCons show ?thesis by auto
	qed
      qed simp
      with Pair show ?thesis by auto
    qed
  qed
  thus ?thesis by simp
qed

end

declare size_simps[simp]
fun PleftI :: "'a \<Rightarrow> 'a \<Rightarrow> ('a \<Rightarrow> 'a \<Rightarrow> 'a) \<Rightarrow> ('a \<Rightarrow> 'a \<Rightarrow> 'a) \<Rightarrow> ('a \<Rightarrow> 'a) \<Rightarrow> ('f,'a)lpoly_inter \<Rightarrow> ('f,'v)term \<Rightarrow> ('v,'a)l_poly" 
where "PleftI ze on pl ti mz pi (Var x) = (LPoly ze [(x,on)] )" 
   |  "PleftI ze on pl ti mz pi (Fun f ts) = (let (c, as) = pi f (length ts) in 
              (case (sum_lpolyI ze pl (LPoly c []) (list_sumI (LPoly ze []) (sum_lpolyI ze pl) (map (\<lambda> at . (mul_lpolyI ze ti (fst at)  (PleftI ze on pl ti mz pi (snd at)))) (zip as ts)))) :: ('v,'a)l_poly of
                  LPoly d [] \<Rightarrow> LPoly (mz d) [] | p \<Rightarrow> p))"

fun PrightI :: "'a \<Rightarrow> 'a \<Rightarrow> ('a \<Rightarrow> 'a \<Rightarrow> 'a) \<Rightarrow> ('a \<Rightarrow> 'a \<Rightarrow> 'a) \<Rightarrow> ('a \<Rightarrow> 'a) \<Rightarrow> ('f,'a)lpoly_inter \<Rightarrow> ('f,'v)term \<Rightarrow> ('v,'a)l_poly" 
where "PrightI ze on pl ti mz pi (Var x) = (LPoly ze [(x,on)] )" 
   |  "PrightI ze on pl ti mz pi (Fun f ts) = (let (c, as) = pi f (length ts) in 
              (case (sum_lpolyI ze pl (LPoly c []) (list_sumI (LPoly ze []) (sum_lpolyI ze pl) (map (\<lambda> at . (mul_lpolyI ze ti (fst at)  (PrightI ze on pl ti mz pi (snd at)))) (zip as ts)))) :: ('v,'a)l_poly of
                  LPoly d vp \<Rightarrow> LPoly (mz d) vp))"
declare size_simps[simp del]

fun
  check_polo_nsI
where
  "check_polo_nsI ze on pl ti mz ge pi (s,t) = do {
     let left = PleftI ze on pl ti mz pi s;
     let right = PrightI ze on pl ti mz pi t;
     check_lpoly_nsI ze on ge left right 
      <+? (\<lambda>e. ''could not ensure '' +#+ shows s +@+ '' >= '' +#+ shows t +@+ shows_nl +@+ e)
   }"


context poly_order
begin

abbreviation zero_lpoly :: "('v,'a)l_poly"
where "zero_lpoly \<equiv> LPoly 0 []"

abbreviation c_lpoly :: "'a \<Rightarrow> ('v,'a)l_poly"
where "c_lpoly c \<equiv> LPoly c []"

abbreviation list_sum_lpoly :: "('v,'a)l_poly list \<Rightarrow> ('v,'a)l_poly"
where "list_sum_lpoly \<equiv> list_sumI zero_lpoly sum_lpoly"

lemma wf_list_sum_lpoly: assumes "\<forall> p \<in> set ps. wf_lpoly p"
  shows "wf_lpoly (list_sum_lpoly ps)"
using assms
proof (induct ps)
  case Nil thus ?case by (simp add: wf_pvars_def)
next
  case (Cons p ps)
  show ?case by (simp only: list_sumI.simps, rule wf_sum_lpoly, auto simp: Cons)
qed

lemma pos_list_sum_lpoly: 
  assumes "\<forall> p \<in> set ps. wf_lpoly p"
    and  "\<forall> p \<in> set ps. pos_coeffs p"
  shows "pos_coeffs (list_sum_lpoly ps)"
using assms
proof (induct ps)
  case Nil thus ?case by (simp add: pos_coeffs_def pos_pvars_def)
next
  case (Cons p ps)
  show ?case by (simp only: list_sumI.simps, rule pos_sum_lpoly, auto simp: Cons, rule wf_list_sum_lpoly, auto simp: Cons)
qed

lemma list_sum_lpoly_sound:
  assumes wf_ps: "\<forall> p \<in> set ps. wf_lpoly p"
    and wf_ass: "wf_ass \<alpha>"
  shows "eval_lpoly \<alpha> (list_sum_lpoly ps) = list_sum (map (eval_lpoly \<alpha>) ps)"
using wf_ps
proof (induct ps)
  case (Cons p ps)
  have "eval_lpoly \<alpha> (list_sum_lpoly (p # ps)) = eval_lpoly \<alpha> (sum_lpoly p (list_sum_lpoly ps))" by auto
  also have "\<dots> = eval_lpoly \<alpha> p + eval_lpoly \<alpha> (list_sum_lpoly ps)" 
    by (rule sum_poly_sound, rule wf_ass, simp add: Cons, rule wf_list_sum_lpoly, auto simp: Cons)
  also have "\<dots> = eval_lpoly \<alpha> p + list_sum (map (eval_lpoly \<alpha>) ps)" using Cons by auto
  finally show ?case by (simp only: list_sumI.simps, auto simp: Cons wf_ass)
qed simp

abbreviation Pleft :: "('f,'v)term \<Rightarrow> ('v,'a)l_poly"
where "Pleft \<equiv> PleftI zero one plus times max0 pi"

abbreviation Pright :: "('f,'v)term \<Rightarrow> ('v,'a)l_poly"
where "Pright \<equiv> PrightI zero one plus times max0 pi"

lemma Pleft_both:  assumes wf_ass: "wf_ass \<alpha>"
  and pos_ass: "pos_ass \<alpha>"
  shows "((pi\<guillemotleft>\<alpha>,t\<guillemotright>) \<ge> eval_lpoly \<alpha> (Pleft t)) \<and> wf_lpoly (Pleft t)" 
proof (induct t)
  case (Var x)
  show ?case using pos_ass wf_ass unfolding pos_ass_def wf_ass_def by (auto simp: wf_pvars_def)
  case (Fun f ts)
  let ?I = "\<lambda> s. pi\<guillemotleft>\<alpha>,s\<guillemotright>"
  let ?E = "\<lambda> p. eval_lpoly \<alpha> p"
  show ?case 
  proof (cases "pi f (length ts)")
    case (Pair c cs)
    from wf_pi have "wf_elem (fst (pi f (length ts))) \<and> (\<forall> c \<in> set (snd (pi f (length ts))). wf_elem c \<and> geq c zero)" by auto
    with Pair have wf_c: "wf_elem c" and wf_cs: "\<forall> c \<in> set cs. wf_elem c" and pos_cs: "\<forall> c \<in> set cs. c \<ge> 0" by auto
    let ?plist = "map (\<lambda> at. mul_lpoly (fst at) (Pleft (snd at))) (zip cs ts)"
    have wf_ps: "\<forall> p \<in> set ?plist. wf_lpoly p"
    proof
      fix p
      assume "p \<in> set ?plist"
      from this obtain a t where p: "p = mul_lpoly a (Pleft t)" and at: "(a,t) \<in> set (zip cs ts)" by auto 
      from at have a: "a \<in> set cs" and t: "t \<in> set ts" using set_zip_leftD[where x = a and y = t] set_zip_rightD[where x = a and y = t] by auto
      show "wf_lpoly p" by (simp only: p, rule wf_mul_lpoly, auto simp: wf_cs a t Fun)
    qed
    have wf_sum: "wf_lpoly (list_sum_lpoly ?plist)"
      by (rule wf_list_sum_lpoly, rule wf_ps)
    have "(list_sum (map (\<lambda> at. fst at * (?I (snd at))) (zip cs ts))) \<ge> list_sum (map ?E ?plist)" (is "geq (list_sum ?vlist) _")
    proof (simp only: o_def map_map, rule list_sum_mono, rule ballI)
      fix ct
      assume mem: "ct \<in> set (zip cs ts)"
      from this obtain c t where ct: "ct = (c,t)" by force
      from ct mem have c: "c \<in> set cs" and t: "t \<in> set ts" using set_zip_leftD[where x = c and y = t] set_zip_rightD[where x = c and y = t] by auto
      have "((c * (?I t)) \<ge> ?E (mul_lpoly c (Pleft t))) \<and> wf_elem (c * ?I t) \<and> wf_elem (?E (mul_lpoly c (Pleft t)))" 
      proof (rule conjI)
	have id: "?E (mul_lpoly c (Pleft t)) = c * (?E (Pleft t))" by (rule mul_poly_sound, auto simp: c wf_cs Fun t wf_ass)
	show "c * (?I t) \<ge> ?E (mul_lpoly c (Pleft t))" 
	  by (simp only: id, rule times_right_mono, auto simp: c wf_cs Fun  pos_cs t Fun wf_ass, rule wf_eval_lpoly, auto simp: wf_ass t Fun)
      next
	show "wf_elem (c * (?I t)) \<and> wf_elem (?E (mul_lpoly c (Pleft t)))"
	by (rule conjI, rule wf_times, auto simp: c wf_cs wf_ass, rule wf_eval_lpoly, auto simp: wf_ass t Fun, rule wf_mul_lpoly, auto simp: c wf_cs t Fun)
      qed
      with ct show " (fst ct * (pi\<guillemotleft>\<alpha>,snd ct\<guillemotright>) \<ge> eval_lpoly \<alpha> (mul_lpoly (fst ct) (Pleft (snd ct)))) \<and> wf_elem (fst ct * (pi\<guillemotleft>\<alpha>,snd ct\<guillemotright>)) \<and> wf_elem (eval_lpoly \<alpha> (mul_lpoly (fst ct) (Pleft (snd ct))))" by simp
    qed
    also have  "\<dots> = ?E (list_sum_lpoly ?plist)" by (rule list_sum_lpoly_sound[symmetric], rule wf_ps, rule wf_ass)
    finally have part1: "list_sum ?vlist \<ge> ?E (list_sum_lpoly ?plist)" . 
    have wf_csump: "wf_lpoly (sum_lpoly (c_lpoly c) (list_sum_lpoly ?plist))" (is "wf_lpoly ?csump")
      by (rule wf_sum_lpoly, auto simp: wf_c wf_pvars_def, rule wf_sum)
    from Pair Fun have id: "Pleft (Fun f ts) = (case ?csump of LPoly d [] \<Rightarrow> LPoly (max0 d) [] | p \<Rightarrow> p)" by auto    
    from wf_csump have wf_final: "wf_lpoly (Pleft (Fun f ts))" by (simp only: id, cases ?csump, cases "get_nc_lpoly ?csump", auto simp: wf_pvars_def wf_max0)
    have "?E ?csump = ?E (c_lpoly c) + ?E (list_sum_lpoly ?plist)" by (rule sum_poly_sound, auto simp: wf_ass wf_c wf_pvars_def wf_sum)
    also have "\<dots> = c + ?E (list_sum_lpoly ?plist)" by (simp add: wf_c)
    finally have idr: "?E ?csump = c + ?E (list_sum_lpoly ?plist)" .    
    have wf_suml: "wf_elem (list_sum ?vlist)"
    proof (rule wf_list_sum, rule ballI)
      fix v
      assume "v \<in> set ?vlist"
      from this obtain c t where v: "v = c * (?I t)" and ct: "(c,t) \<in> set (zip cs ts)" by auto 
      from ct have c: "c \<in> set cs" and t: "t \<in> set ts" using set_zip_leftD[where x = c and y = t] set_zip_rightD[where x = c and y = t] by auto
      show "wf_elem v" by (simp only: v, rule wf_times, auto simp: wf_cs c wf_ass)
    qed
    have ge_csum: "c + list_sum ?vlist \<ge> ?E ?csump" 
      by (simp only: idr, rule plus_right_mono, rule part1, rule wf_c, rule wf_suml,
      rule wf_eval_lpoly, rule wf_ass, rule wf_sum)
    have wf_csuml: "wf_elem (c + list_sum ?vlist)" using wf_suml wf_c by auto
    from Pair have left: "?I (Fun f ts) = max0 ( c + list_sum ?vlist)" by auto 
    have "\<dots> \<ge> max0 ( ?E ?csump)"
      by (rule max0_mono, rule ge_csum, rule wf_csuml, rule wf_eval_lpoly, rule wf_ass, rule wf_csump)
    hence part2: "?I (Fun f ts) \<ge> max0 ( ?E ?csump)" by (simp only: left)    
    have part3: "max0 ( ?E ?csump) \<ge> ?E (Pleft (Fun f ts))"
      using wf_csump by (simp only: id, cases ?csump, cases "get_nc_lpoly ?csump", auto simp: max0_x wf_ass wf_max0)
    have ge_final: "?I (Fun f ts) \<ge> ?E (Pleft (Fun f ts))"
      by (rule geq_trans, rule part2, rule part3, rule wf_terms, rule wf_ass, rule wf_max0, 
	rule wf_eval_lpoly, rule wf_ass, rule wf_csump, rule wf_eval_lpoly, rule wf_ass, rule wf_final)    
    from wf_final ge_final show ?thesis by auto
  qed
qed
  

lemma Pleft_sound:   assumes wf_ass: "wf_ass \<alpha>"
  and pos_ass: "pos_ass \<alpha>"
  shows "(pi\<guillemotleft>\<alpha>,t\<guillemotright>) \<ge> eval_lpoly \<alpha> (Pleft t)"
  using Pleft_both assms by auto

lemma wf_Pleft[simp]: "wf_lpoly (Pleft t)"
  using Pleft_both[where \<alpha> = zero_ass] by auto

lemma pos_eval_pvars: assumes wf_ass: "wf_ass \<alpha>"
  and pos_ass: "pos_ass \<alpha>"
  and wf_vas: "wf_pvars vas"
  and pos_vas: "pos_pvars vas"
  shows "eval_pvars \<alpha> vas \<ge> 0"
using wf_vas pos_vas 
proof (induct vas)
  case (Cons xa vas)
  show ?case 
  proof (cases xa)
    case (Pair x a)
    from pos_ass wf_ass have pos_x: "\<alpha> x \<ge> 0" and wf_x: "wf_elem (\<alpha> x)" unfolding pos_ass_def wf_ass_def by auto
    from Pair Cons have pos_a: "a \<ge> 0" and wf_a: "wf_elem a" unfolding wf_pvars_def pos_pvars_def by auto
    have "a * \<alpha> x \<ge> 0 * 0" by (rule geq_trans[where y = "a * 0"], rule times_right_mono, auto simp: wf_a pos_a wf_x pos_x)
    hence pos_ax: "a * \<alpha> x \<ge> 0" by auto
    have wf_ax: "wf_elem (a * \<alpha> x)" by (auto simp: wf_x wf_a)
    from Cons have pos_rec: "eval_pvars \<alpha> vas \<ge> 0" unfolding wf_pvars_def pos_pvars_def by auto
    from Cons have "wf_pvars vas" unfolding wf_pvars_def by auto
    with Cons have wf_rec: "wf_elem (eval_pvars \<alpha> vas)" using wf_ass by auto
    have "a * \<alpha> x + eval_pvars \<alpha> vas \<ge> 0 + 0" by (rule plus_left_right_mono, auto simp: pos_ax wf_ax pos_rec wf_rec)
    thus ?thesis by (simp add: Pair)
  qed
qed simp
    

lemma Pright_both:  assumes wf_ass: "wf_ass \<alpha>"
  and pos_ass: "pos_ass \<alpha>"
  shows "(eval_lpoly \<alpha> (Pright t) \<ge> (pi\<guillemotleft>\<alpha>,t\<guillemotright>)) \<and> wf_lpoly (Pright t) \<and> pos_coeffs (Pright t)" 
proof (induct t)
  case (Var x)
  show ?case using pos_ass wf_ass unfolding pos_ass_def wf_ass_def pos_coeffs_def pos_pvars_def by (auto simp: wf_pvars_def)
  case (Fun f ts)
  let ?I = "\<lambda> s. pi\<guillemotleft>\<alpha>,s\<guillemotright>"
  let ?E = "\<lambda> p. eval_lpoly \<alpha> p"
  show ?case 
  proof (cases "pi f (length ts)")
    case (Pair c cs)
    from wf_pi have "wf_elem (fst (pi f (length ts))) \<and> (\<forall> c \<in> set (snd (pi f (length ts))). wf_elem c \<and> geq c zero)" by auto
    with Pair have wf_c: "wf_elem c" and wf_cs: "\<forall> c \<in> set cs. wf_elem c" and pos_cs: "\<forall> c \<in> set cs. c \<ge> 0" by auto
    let ?plist = "map (\<lambda> at. mul_lpoly (fst at) (Pright (snd at))) (zip cs ts)"
    have wf_ps: "\<forall> p \<in> set ?plist. wf_lpoly p"
    proof
      fix p
      assume "p \<in> set ?plist"
      from this obtain a t where p: "p = mul_lpoly a (Pright t)" and at: "(a,t) \<in> set (zip cs ts)" by auto 
      from at have a: "a \<in> set cs" and t: "t \<in> set ts" using set_zip_leftD[where x = a and y = t] set_zip_rightD[where x = a and y = t] by auto
      show "wf_lpoly p" by (simp only: p, rule wf_mul_lpoly, auto simp: wf_cs a t Fun)
    qed
    have wf_sum: "wf_lpoly (list_sum_lpoly ?plist)"
      by (rule wf_list_sum_lpoly, rule wf_ps)
    have pos_ps: "\<forall> p \<in> set ?plist. pos_coeffs p"
    proof
      fix p
      assume "p \<in> set ?plist"
      from this obtain a t where p: "p = mul_lpoly a (Pright t)" and at: "(a,t) \<in> set (zip cs ts)" by auto 
      from at have a: "a \<in> set cs" and t: "t \<in> set ts" using set_zip_leftD[where x = a and y = t] set_zip_rightD[where x = a and y = t] by auto
      show "pos_coeffs p" by (simp only: p, rule pos_mul_lpoly, auto simp: a pos_cs wf_cs t Fun)
    qed
    have pos_sum: "pos_coeffs (list_sum_lpoly ?plist)"
      by (rule pos_list_sum_lpoly, rule wf_ps, rule pos_ps)
    have id1: "?E (list_sum_lpoly ?plist) = list_sum (map ?E ?plist)" by (rule list_sum_lpoly_sound, rule wf_ps, rule wf_ass)
    have "\<dots> \<ge> (list_sum (map (\<lambda> at. fst at * (?I (snd at))) (zip cs ts)))" (is "geq _ (list_sum ?vlist)")
    proof (simp only: o_def map_map, rule list_sum_mono, rule ballI)
      fix ct
      assume mem: "ct \<in> set (zip cs ts)"
      from this obtain c t where ct: "ct = (c,t)" by force
      from ct mem have c: "c \<in> set cs" and t: "t \<in> set ts" using set_zip_leftD[where x = c and y = t] set_zip_rightD[where x = c and y = t] by auto
      have "(?E (mul_lpoly c (Pright t)) \<ge> (c * (?I t))) \<and> wf_elem (c * ?I t) \<and> wf_elem (?E (mul_lpoly c (Pright t)))" 
      proof (rule conjI)
	have id: "?E (mul_lpoly c (Pright t)) = c * (?E (Pright t))" by (rule mul_poly_sound, auto simp: c wf_cs Fun t wf_ass)
	show "?E (mul_lpoly c (Pright t)) \<ge> c * (?I t)" 
	  by (simp only: id, rule times_right_mono[OF _ _ _ wf_eval_lpoly], auto simp: c wf_cs Fun[OF t]  pos_cs  wf_ass)
      next
	show "wf_elem (c * (?I t)) \<and> wf_elem (?E (mul_lpoly c (Pright t)))"
	  by (rule conjI[OF wf_times wf_eval_lpoly[OF _ wf_mul_lpoly]], auto simp: c wf_cs wf_ass Fun[OF t])
      qed	
      with ct show " (eval_lpoly \<alpha> (mul_lpoly (fst ct) (Pright (snd ct))) \<ge> fst ct * (pi\<guillemotleft>\<alpha>,snd ct\<guillemotright>)) \<and> wf_elem (eval_lpoly \<alpha> (mul_lpoly (fst ct) (Pright (snd ct)))) \<and> wf_elem (fst ct * (pi\<guillemotleft>\<alpha>,snd ct\<guillemotright>))" by simp
    qed
    hence part1: "?E (list_sum_lpoly ?plist) \<ge> list_sum ?vlist" by (simp only: id1)  
    have wf_csump: "wf_lpoly (sum_lpoly (c_lpoly c) (list_sum_lpoly ?plist))" (is "wf_lpoly ?csump")
      by (rule wf_sum_lpoly[OF _ wf_sum], auto simp: wf_c wf_pvars_def)
    have pos_csump: "pos_coeffs ?csump"
      by (rule pos_sum_lpoly, auto simp: wf_sum wf_c pos_sum, auto simp: wf_c wf_pvars_def pos_coeffs_def pos_pvars_def)
    from Pair Fun have id: "Pright (Fun f ts) = (case ?csump of LPoly d nc \<Rightarrow> LPoly (max0 d) nc)" by auto    
    from wf_csump have wf_final: "wf_lpoly (Pright (Fun f ts))" by (simp only: id, cases ?csump, auto simp: wf_pvars_def wf_max0)
    from pos_csump have pos_final: "pos_coeffs (Pright (Fun f ts))" unfolding pos_coeffs_def by (simp only: id, cases ?csump, auto)
    have "?E ?csump = ?E (c_lpoly c) + ?E (list_sum_lpoly ?plist)" by (rule sum_poly_sound, auto simp: wf_ass wf_c wf_pvars_def wf_sum)
    also have "\<dots> = c + ?E (list_sum_lpoly ?plist)" by (simp add: wf_c)
    finally have idr: "?E ?csump = c + ?E (list_sum_lpoly ?plist)" .    
    have wf_sumr: "wf_elem (list_sum ?vlist)"
    proof (rule wf_list_sum, rule ballI)
      fix v
      assume "v \<in> set ?vlist"
      from this obtain c t where v: "v = c * (?I t)" and ct: "(c,t) \<in> set (zip cs ts)" by auto 
      from ct have c: "c \<in> set cs" and t: "t \<in> set ts" using set_zip_leftD[where x = c and y = t] set_zip_rightD[where x = c and y = t] by auto
      show "wf_elem v" by (simp only: v, rule wf_times, auto simp: wf_cs c wf_ass)
    qed
    have ge_csum: "?E ?csump \<ge> c + list_sum ?vlist" 
      by (simp only: idr, rule plus_right_mono, rule part1, rule wf_c,
      rule wf_eval_lpoly, rule wf_ass, rule wf_sum, rule wf_sumr)
    have wf_csuml: "wf_elem (c + list_sum ?vlist)" using wf_sumr wf_c by auto
    from Pair have right: "?I (Fun f ts) = max0 ( c + list_sum ?vlist)" by auto 
    have "max0 ( ?E ?csump) \<ge> \<dots>"
      by (rule max0_mono, rule ge_csum, rule wf_eval_lpoly, rule wf_ass, rule wf_csump, rule wf_csuml)
    hence part2: "max0 ( ?E ?csump) \<ge> ?I (Fun f ts)" by (simp only: right)    
    have part3: "?E (Pright (Fun f ts)) \<ge> max0 (?E ?csump)"
    proof (cases ?csump)
      case (LPoly d nc)
      have wf_nc: "wf_pvars nc" using wf_csump LPoly by (simp only: id, auto)
      have wf_enc: "wf_elem (eval_pvars \<alpha> nc)" using wf_csump LPoly by (simp only: id, auto simp: wf_ass)
      have wf_d: "wf_elem d" using wf_csump LPoly by (simp only: id, auto)
      have pos_nc: "pos_pvars nc" using pos_csump LPoly unfolding pos_coeffs_def by (simp only: id, auto)
      have nc_pos: "eval_pvars \<alpha> nc \<ge> 0" by (rule pos_eval_pvars, auto simp: wf_ass pos_ass pos_nc wf_nc)
      have md_pos: "max0 d \<ge> 0" by (rule max0_pos, rule wf_d)
      have "max0 d + eval_pvars \<alpha> nc \<ge> 0 + 0" (is "geq ?pr _") by (rule plus_left_right_mono, rule md_pos, rule nc_pos, auto simp: wf_enc wf_d wf_max0)
      hence "?pr \<ge> 0" by simp
      hence max_id: "max0 (?pr) = ?pr" by (rule max0_id_pos, auto simp: wf_enc wf_d wf_max0)
      have "?pr \<ge> max0 (d + eval_pvars \<alpha> nc)"
      proof (rule geq_trans[where ?y = "max0 (?pr)"], simp only: max_id, rule geq_refl) 
        show "max0 ?pr \<ge> max0 (d + eval_pvars \<alpha> nc)"
          by (rule max0_mono[OF plus_left_mono[OF max0_x]], auto simp: wf_d wf_max0 wf_enc)
      qed (auto simp: wf_d wf_max0 wf_enc)
      thus ?thesis by (simp only: id, simp only: LPoly, simp)
    qed
    have ge_final: "?E (Pright (Fun f ts)) \<ge> ?I (Fun f ts)"
      by (rule geq_trans, rule part3, rule part2, rule wf_eval_lpoly, rule wf_ass, rule wf_final, rule wf_max0,
	rule wf_eval_lpoly, rule wf_ass, rule wf_csump, rule wf_terms, rule wf_ass)
    from wf_final ge_final pos_final show ?thesis by auto
  qed
qed


lemma Pright_sound:   assumes wf_ass: "wf_ass \<alpha>"
  and pos_ass: "pos_ass \<alpha>"
  shows "eval_lpoly \<alpha> (Pright t) \<ge> (pi\<guillemotleft>\<alpha>,t\<guillemotright>)"
  using Pright_both assms by auto

lemma wf_Pright[simp]: "wf_lpoly (Pright t)"
  using Pright_both[where \<alpha> = zero_ass] by auto

abbreviation check_polo_ns :: "('f,'v::show)rule \<Rightarrow> shows check"
where "check_polo_ns \<equiv> check_polo_nsI zero one plus times max0 geq pi"
end

type_synonym ('f,'a)lpoly_interL = "(('f \<times> nat) \<times> ('a \<times> 'a list)) list"

(* default interpretation f/n \<rightarrow> d + x_1 + ... + x_n *)
fun to_lpoly_inter :: "'a \<Rightarrow> 'a \<Rightarrow> ('f,'a)lpoly_interL \<Rightarrow> ('f,'a)lpoly_inter"
where "to_lpoly_inter d on I f n = (case lookup (f,n) I of None \<Rightarrow> (d,replicate n on) | Some cas \<Rightarrow> cas)"


(* create an arity that is not used *)
fun create_arity :: "('f,'a)lpoly_interL \<Rightarrow> nat"
  where "create_arity fcns = Suc (
            foldr (\<lambda> (fn,ccs) m. max m (snd fn)) fcns 0
            )"

lemma create_arity_sound: assumes
  m: "m \<ge> create_arity I"
  shows "to_lpoly_inter d on I f m = (d, replicate m on)" (is "?I f m = _")
proof (cases "lookup (f,m) I")
  case None thus ?thesis by auto
next
  case (Some val)
  hence one: "((f,m),val) \<in> set I" by (rule lookup_sound)
  hence "m \<le> foldr (\<lambda> (fn, ccs) n. max n (snd fn)) I 0" 
  proof (induct I)
    case (Cons fnccs I)
    thus ?case by  (cases fnccs, cases "snd fnccs", auto)
  qed simp
  with m show ?thesis by auto
qed

declare create_arity.simps [simp del]

context poly_order
begin

lemma inter_ns_ce_compat: assumes pi: "pi = to_lpoly_inter default one I" (is "_ = ?pi")
  shows "\<exists> n. \<forall> m. m \<ge> n \<longrightarrow> (\<forall> c c'. ce_trs (c, c', m) \<subseteq> inter_ns)"
unfolding inter_ns_def
proof (rule exI[of _ "create_arity I"], intro allI impI, rule subsetI, simp only: ce_trs.simps)
  fix m c c' x
  let ?n = "create_arity I"
  assume m: "?n \<le> m"
  let ?m = "Suc (Suc m)"
  from m have "?m \<ge> ?n" by auto
  hence inter: "pi c ?m = (default, replicate ?m 1)" by (simp only: pi, rule create_arity_sound)
  let ?sum = "list_sum (replicate m (1 * (\<lambda>(a, as). max0 (a + 0)) (pi c' 0)))"
  obtain a as where pi: "pi c' 0 = (a,as)" by force
  have "wf_elem (fst (pi c' 0))" using wf_pi by auto
  with pi have wf_a: "wf_elem a" by auto
  obtain b where b: "b = max0 (a + 0)" by auto
  have idb: "1 * (\<lambda>(a, as). max0 (a + 0)) (pi c' 0) = b"
    by (simp add: pi b wf_a wf_max0)
  have wf_pos_b: "wf_elem b \<and> geq b 0" by (simp add: b, auto simp: wf_a wf_max0 max0_pos)
  have wf_sum: "wf_elem ?sum"
    by (simp only: idb, rule wf_list_sum, rule replicate_prop, auto simp: wf_pos_b)
  have pos_sum: "?sum \<ge> 0"
    by (simp only: idb, rule pos_list_sum, rule replicate_prop, auto simp: wf_pos_b)
  from wf_sum pos_sum obtain d where d: "?sum = d" and wf_pos_d: "wf_elem d \<and> geq d 0" by force
  assume "x \<in> {(Fun c (t # s # replicate m (Fun c' [])), t)| t s. True} \<union> 
              {(Fun c (t # s # replicate m (Fun c' [])), s)| t s. True}" (is "_ \<in> ?tc \<union> ?sc")
  thus "x \<in> {(s,t). \<forall> \<alpha>. wf_ass \<alpha> \<and> pos_ass \<alpha> \<and> apos_ass \<alpha> \<longrightarrow> (pi\<guillemotleft>\<alpha>,s\<guillemotright> \<ge> pi\<guillemotleft>\<alpha>,t\<guillemotright>)}"
  proof
    assume "x \<in> ?tc"
    from this obtain t s where id: "x = (Fun c (t # s # replicate m (Fun c' [])), t)" (is "x = (?l, _)") by auto    
    have "\<forall> \<alpha>. wf_ass \<alpha> \<and> pos_ass \<alpha> \<and> apos_ass \<alpha> \<longrightarrow> (pi\<guillemotleft>\<alpha>,?l\<guillemotright> \<ge> pi\<guillemotleft>\<alpha>,t\<guillemotright>)" 
    proof (intro allI)
      fix \<alpha>
      show "wf_ass \<alpha> \<and> pos_ass \<alpha> \<and> apos_ass \<alpha> \<longrightarrow> (pi\<guillemotleft>\<alpha>,?l\<guillemotright> \<ge> pi\<guillemotleft>\<alpha>,t\<guillemotright>)" (is "?wf_pos \<longrightarrow> ?goal")
      proof
	assume wf_pos_ass: ?wf_pos
	let ?term = "((default + (pi\<guillemotleft>\<alpha>,t\<guillemotright>)) + (pi\<guillemotleft>\<alpha>,s\<guillemotright>)) + d"
	have max: "max0 ?term = ?term" 
          by (rule max0_id_pos, auto simp: wf_pos_ass wf_pos_d, (rule sum_pos)+, auto simp: wf_pos_ass wf_pos_d)
	have "max0 ?term \<ge> ((0 + (pi\<guillemotleft>\<alpha>,t\<guillemotright>)) + 0) + 0"
	  by (simp only: max, (rule plus_left_right_mono)+, auto simp: wf_pos_ass wf_pos_d)
	hence "max0 ?term \<ge> (pi\<guillemotleft>\<alpha>,t\<guillemotright>)" by (auto simp: wf_pos_ass)
	with wf_pos_ass show ?goal using inter 
	  by (auto simp: d wf_pos_d wf_max0)
      qed      
    qed
    thus ?thesis using id by simp
  next
    assume "x \<in> ?sc"
    from this obtain t s where id: "x = (Fun c (t # s # replicate m (Fun c' [])), s)" (is "x = (?l, _)") by auto    
    have "\<forall> \<alpha>. wf_ass \<alpha> \<and> pos_ass \<alpha> \<and> apos_ass \<alpha> \<longrightarrow> (pi\<guillemotleft>\<alpha>,?l\<guillemotright> \<ge> pi\<guillemotleft>\<alpha>,s\<guillemotright>)" 
    proof (intro allI)
      fix \<alpha>
      show "wf_ass \<alpha> \<and> pos_ass \<alpha> \<and> apos_ass \<alpha> \<longrightarrow> (pi\<guillemotleft>\<alpha>,?l\<guillemotright> \<ge> pi\<guillemotleft>\<alpha>,s\<guillemotright>)" (is "?wf_pos \<longrightarrow> ?goal")
      proof
	assume wf_pos_ass: ?wf_pos
	let ?term = "((default + (pi\<guillemotleft>\<alpha>,t\<guillemotright>)) + (pi\<guillemotleft>\<alpha>,s\<guillemotright>)) + d"
	have max: "max0 ?term = ?term" 
          by (rule max0_id_pos, auto simp: wf_pos_ass wf_pos_d, (rule sum_pos)+, auto simp: wf_pos_ass wf_pos_d)
	have "max0 ?term \<ge> ((0 + 0) + (pi\<guillemotleft>\<alpha>,s\<guillemotright>)) + 0" 
	  by (simp only: max, (rule plus_left_right_mono)+, auto simp: wf_pos_ass wf_pos_d) 
	hence "max0 ?term \<ge> (pi\<guillemotleft>\<alpha>,s\<guillemotright>)" by (auto simp: wf_pos_ass)
	with wf_pos_ass show ?goal using inter 
	  by (auto simp: d wf_pos_d wf_max0)
      qed
    qed
    thus ?thesis using id by simp
  qed
qed
end

fun create_afI :: "'a \<Rightarrow> ('f, 'a) lpoly_interL \<Rightarrow> 'f af" where
  "create_afI ze [] (f, n) = (\<lambda>_. True)"
| "create_afI ze (((g, m), (c, coeffs)) # rest) (f, n) = (if f = g \<and> m = n
    then (let len = length coeffs in (\<lambda> i. i < len \<and> coeffs ! i \<noteq> ze))
    else create_afI ze rest (f, n))"

lemma create_af_sound:
  assumes "\<not> create_afI ze I (f, n) i"
  shows "length (snd (to_lpoly_inter def on I f n)) \<le> i \<or> (snd (to_lpoly_inter def on I f n)) ! i = ze"
using assms
proof (induct I, simp)
  case (Cons gmcas I)
  show ?case 
  proof (cases gmcas)
    case (fields gm c as)
    show ?thesis
    proof (cases gm)
      case (Pair g m)
      with Cons fields show ?thesis by (cases "f = g \<and> m = n", auto simp: Let_def)
    qed
  qed
qed
      
declare create_afI.simps[simp del]

context poly_order_carrier 
begin
abbreviation create_af :: "('f,'a)lpoly_interL \<Rightarrow> 'f af"
  where "create_af \<equiv> create_afI zero"
end

context poly_order
begin

lemma inter_ns_af_compat: assumes pi: "pi = to_lpoly_inter def on I" (is "_ = ?pi")
  shows "af_compatible (create_af I) inter_ns"
proof (unfold af_compatible_def, intro allI)
  fix f :: 'f and bef and s t :: "('f,'v)term" and aft
  obtain v :: 'v where True by simp
  let ?st = "Fun f (bef @ s # aft)"
  let ?tt = "Fun f (bef @ t # aft)"
  let ?pair = "(?st, ?tt)"
  let ?n = "Suc (length bef + length aft)"
  let ?i = "length bef"
  show "create_af I (f, ?n) ?i \<or> ?pair \<in> inter_ns"
  proof (cases "create_af I (f, ?n) ?i")
    case True thus ?thesis ..
  next
    case False
    hence "length (snd (?pi f ?n)) \<le> ?i \<or> snd (?pi f ?n) ! ?i = 0" by (rule create_af_sound)
    with pi have pi3: "length (snd (pi f ?n)) \<le> ?i \<or> snd (pi f ?n) ! ?i = 0" by simp
    obtain c cs where pi2: "pi f ?n = (c,cs)" by force
    with pi3 have short_zero: "length cs \<le> ?i \<or> cs ! ?i = 0" by auto
    have "\<forall> \<alpha>. wf_ass \<alpha> \<and> pos_ass \<alpha> \<longrightarrow> ((pi\<guillemotleft>\<alpha>,?st\<guillemotright>) \<ge> (pi\<guillemotleft>\<alpha>,?tt\<guillemotright>))" (is "\<forall> \<alpha>. ?wfpos \<alpha> \<longrightarrow> ?rel \<alpha>")
    proof
      fix \<alpha>
      show "?wfpos \<alpha> \<longrightarrow> ?rel \<alpha>"
      proof
	assume wfpos: "?wfpos \<alpha>"
	hence wf_ass: "wf_ass \<alpha>" ..
	let ?in = "\<lambda> t :: ('f,'v) term. (pi\<guillemotleft>\<alpha>,t\<guillemotright>)"
	let ?inter = "\<lambda> s bef cs. list_sum (map (\<lambda> at. fst at * (?in (snd at))) (zip cs (bef @ s # aft)))"
	let ?intre = "\<lambda> aft cs. list_sum (map (\<lambda> at. fst at * (?in (snd at))) (zip cs aft))"
	have int_s: "?in ?st = max0 (c + ?inter s bef cs)" by (auto simp: pi2)
	have int_t: "?in ?tt = max0 (c + ?inter t bef cs)" by (auto simp: pi2)	
	from short_zero have "?inter s bef cs = ?inter t bef cs" 
	proof (induct cs arbitrary: bef)
	  case Nil thus ?case by auto
	next
	  case (Cons c cs) note oCons = this
	  show ?case 
	  proof (cases bef)
	    case Nil
	    with Cons show ?thesis by (simp add: wf_ass)
	  next
	    case (Cons b befs)
	    with oCons have "length cs \<le> length befs \<or> cs ! length befs = 0" by auto
	    with Cons oCons show ?thesis by auto
	  qed
	qed
	with int_s int_t have eq: "?in ?st = ?in ?tt" by auto
	show "?rel \<alpha>" by (simp only: eq, rule geq_refl, rule wf_terms, auto simp: wf_ass)
      qed
    qed
    hence "?pair \<in> inter_ns" unfolding inter_ns_def by auto
    thus ?thesis .. 
  qed
qed
end

fun
  check_poly_monoI :: "'a::show \<Rightarrow> ('a \<Rightarrow> 'a \<Rightarrow> bool) \<Rightarrow> ('a \<Rightarrow> bool) \<Rightarrow> ('f::show, 'a) lpoly_interL \<Rightarrow> shows check"
where
  "check_poly_monoI ze ge cm = check_allm (\<lambda>((f, n), (c, cs)). do {
     check (ge c ze) (''constant part '' +#+ shows c +@+ '' must be at least '' +#+ shows ze +@+ shows_nl);
     check (n \<le> length cs) (''the last argument is ignored'' +#+ shows_nl);
     check_allm (\<lambda>d. check (cm d) (''coefficient '' +#+ shows d +@+ '' is not allowed'' +#+ shows_nl)) cs
   } <+? (\<lambda>s. ''problem with monotonicity due to interpretation of '' +#+ shows f +@+ ''/''  +#+ shows n +@+ shows_nl +@+ s)
   )"

context mono_poly_order_carrier 
begin
abbreviation check_poly_mono :: "('f::show,'a::show)lpoly_interL \<Rightarrow> shows check"
where "check_poly_mono \<equiv> check_poly_monoI 0 geq check_mono"

lemma plus_gt_right_mono: "\<lbrakk>y > z; wf_elem x; wf_elem y; wf_elem z\<rbrakk> \<Longrightarrow> x + y > x + z"
  by (simp add: plus_comm[where x = x], rule plus_gt_left_mono, auto simp: plus_single_mono)

abbreviation pi_of where "pi_of \<equiv> to_lpoly_inter default 1"

lemma check_poly_mono_sound: assumes "isOK(check_poly_mono I)"
  shows "\<forall> f n. (fst (pi_of I f n) \<ge> 0) \<and> (length (snd (pi_of I f n)) \<ge> n \<and> (\<forall> c \<in> set (snd (pi_of I f n)). check_mono c \<or> c = 1))" (is "\<forall> f n. ?prop f n")
proof (intro allI)
  fix f n
  show "?prop f n" 
  proof (cases "lookup (f,n) I")
    case None
    hence id: "pi_of I f n = (default, replicate n one)" by auto
    show ?thesis
      by (simp only: id, rule conjI, simp, rule conjI, simp, simp)
  next
    case (Some ccs)
    from this obtain c cs where lookup: "lookup (f,n) I = Some (c,cs)" by force
    hence "((f,n),(c,cs)) \<in> set I" by (rule lookup_sound)
    with assms lookup show "?prop f n" by auto
  qed
qed

declare check_poly_monoI.simps[simp del]
end

locale mono_poly_order = poly_order + mono_poly_order_carrier + 
  assumes pi_mono: "\<forall> f n. (fst (pi f n) \<ge> 0) \<and> (length (snd (pi f n)) \<ge> n \<and> (\<forall> c \<in> set (snd (pi f n)). check_mono c \<or> c = 1))"
begin

lemma inter_s_ce_compat: assumes pi: "pi = to_lpoly_inter default one I" (is "_ = ?pi")
  shows "\<exists> n. \<forall> m. m \<ge> n \<longrightarrow> (\<forall> c c'. ce_trs (c, c', m) \<subseteq> inter_s)"
unfolding inter_s_def
proof (rule exI[of _ "create_arity I"], intro allI impI, rule subsetI, simp only: ce_trs.simps)
  fix m c c' x
  let ?n = "create_arity I"
  assume m: "?n \<le> m"
  let ?m = "Suc (Suc m)"
  from m have "?m \<ge> ?n" by auto
  hence inter: "pi c ?m = (default, replicate ?m 1)" by (simp only: pi, rule create_arity_sound)
  let ?sum = "list_sum (replicate m (1 * (\<lambda>(a, as). max0 (a + 0)) (pi c' 0)))"
  obtain a as where pi: "pi c' 0 = (a,as)" by force
  have "wf_elem (fst (pi c' 0))" using wf_pi by auto
  with pi have wf_a: "wf_elem a" by auto
  obtain b where b: "b = max0 (a + 0)" by auto
  have idb: "1 * (\<lambda>(a, as). max0 (a + 0)) (pi c' 0) = b"
    by (simp add: pi b wf_a wf_max0)
  have wf_pos_b: "wf_elem b \<and> geq b 0" by (simp add: b, auto simp: wf_a wf_max0 max0_pos)
  have wf_sum: "wf_elem ?sum"
    by (simp only: idb, rule wf_list_sum, rule replicate_prop, auto simp: wf_pos_b)
  have pos_sum: "?sum \<ge> 0"
    by (simp only: idb, rule pos_list_sum, rule replicate_prop, auto simp: wf_pos_b)
  from wf_sum pos_sum obtain d where d: "?sum = d" and wf_pos_d: "wf_elem d \<and> geq d 0" by force
  fix x
  assume "x \<in> {(Fun c (t # s # replicate m (Fun c' [])), t) | t s. True} \<union>
              {(Fun c (t # s # replicate m (Fun c' [])), s) | t s. True}" (is "_ \<in> ?tc \<union> ?sc")
  thus "x \<in> {(s,t). \<forall>\<alpha>. wf_ass \<alpha> \<and> pos_ass \<alpha> \<and> apos_ass \<alpha> \<longrightarrow> (pi\<guillemotleft>\<alpha>,s\<guillemotright> > pi\<guillemotleft>\<alpha>,t\<guillemotright>)}"
  proof
    assume "x \<in> ?tc"
    then obtain t s where id: "x = (Fun c (t # s # replicate m (Fun c' [])), t)" (is "x = (?l, _)") by auto    
    have "\<forall> \<alpha>. wf_ass \<alpha> \<and> pos_ass \<alpha> \<and> apos_ass \<alpha> \<longrightarrow> (pi\<guillemotleft>\<alpha>,?l\<guillemotright> > pi\<guillemotleft>\<alpha>,t\<guillemotright>)" 
    proof (intro allI)
      fix \<alpha>
      show "wf_ass \<alpha> \<and> pos_ass \<alpha> \<and> apos_ass \<alpha> \<longrightarrow> (pi\<guillemotleft>\<alpha>,?l\<guillemotright> > pi\<guillemotleft>\<alpha>,t\<guillemotright>)" (is "?wf_pos \<longrightarrow> ?goal")
      proof
	assume wf_pos_ass: ?wf_pos
	let ?term = "((default + (pi\<guillemotleft>\<alpha>,t\<guillemotright>)) + (pi\<guillemotleft>\<alpha>,s\<guillemotright>)) + d"
	have max: "max0 ?term = ?term" 
          by (rule max0_id_pos, auto simp: wf_pos_ass wf_pos_d, (rule sum_pos)+, auto simp: wf_pos_ass wf_pos_d)
	have "max0 ?term > ((0 + (pi\<guillemotleft>\<alpha>,t\<guillemotright>)) + 0) + 0"
	  by (simp only: max, (rule plus_gt_left_mono2)+, auto simp: wf_pos_ass wf_pos_d plus_single_mono default_gt_zero)
	hence "max0 ?term > (pi\<guillemotleft>\<alpha>,t\<guillemotright>)" by (auto simp: wf_pos_ass)
	with wf_pos_ass show ?goal using inter 
	  by (auto simp: d wf_pos_d wf_max0)
      qed
    qed
    thus ?thesis using id by simp
  next
    assume "x \<in> ?sc"
    then obtain t s where id: "x = (Fun c (t # s # replicate m (Fun c' [])), s)" (is "x = (?l, _)") by auto    
    have "\<forall> \<alpha>. wf_ass \<alpha> \<and> pos_ass \<alpha> \<and> apos_ass \<alpha> \<longrightarrow> (pi\<guillemotleft>\<alpha>,?l\<guillemotright> > pi\<guillemotleft>\<alpha>,s\<guillemotright>)" 
    proof (intro allI)
      fix \<alpha>
      show "wf_ass \<alpha> \<and> pos_ass \<alpha> \<and> apos_ass \<alpha> \<longrightarrow> (pi\<guillemotleft>\<alpha>,?l\<guillemotright> > pi\<guillemotleft>\<alpha>,s\<guillemotright>)" (is "?wf_pos \<longrightarrow> ?goal")
      proof
	assume wf_pos_ass: ?wf_pos
	let ?term = "((default + (pi\<guillemotleft>\<alpha>,t\<guillemotright>)) + (pi\<guillemotleft>\<alpha>,s\<guillemotright>)) + d"
	have max: "max0 ?term = ?term" 
          by (rule max0_id_pos, auto simp: wf_pos_ass wf_pos_d, (rule sum_pos)+, auto simp: wf_pos_ass wf_pos_d)
	have "max0 ?term > ((0 + 0) + (pi\<guillemotleft>\<alpha>,s\<guillemotright>)) + 0" 
	  by (simp only: max, (rule plus_gt_left_mono2)+, auto simp: wf_pos_ass wf_pos_d plus_single_mono default_gt_zero) 
	hence "max0 ?term > (pi\<guillemotleft>\<alpha>,s\<guillemotright>)" by (auto simp: wf_pos_ass)
	with wf_pos_ass show ?goal using inter 
	  by (auto simp: d wf_pos_d wf_max0)
      qed
    qed
    thus ?thesis using id by simp
  qed
qed

lemma inter_s_mono: 
  shows "one_step_ctxt_closed inter_s" 
unfolding one_step_ctxt_closed_def
proof (intro allI)
  fix f :: 'b and bef and s t :: "('b,'v)term" and aft
  let ?st = "Fun f (bef @ s # aft)"
  let ?tt = "Fun f (bef @ t # aft)"
  let ?pair = "(?st, ?tt)"
  let ?n = "Suc (length bef + length aft)"
  let ?i = "length bef"
  obtain c cs where ccs: "pi f ?n = (c,cs)" by force
  from pi_mono have "((fst (pi f ?n)) \<ge> 0) \<and> length (snd (pi f ?n)) \<ge> ?n \<and> (\<forall> c \<in> set (snd (pi f ?n)). check_mono c \<or> c = 1)" by auto
  with ccs have mono: "(c \<ge> 0) \<and> length cs \<ge> ?n \<and> (\<forall> d \<in> set cs. check_mono d \<or> d = 1)" by auto
  from wf_pi have "wf_elem (fst (pi f ?n))" and "\<forall> d \<in> set (snd (pi f ?n)). wf_elem d \<and> geq d 0" by auto
  with ccs have wf_c: "wf_elem c" and wf_cs: "\<forall> d \<in> set cs. wf_elem d \<and> geq d 0" by auto
  show "(s,t) \<in> inter_s \<longrightarrow> (Fun f (bef @ s # aft), Fun f (bef @ t # aft)) \<in> inter_s"
  proof
    assume st: "(s,t) \<in> inter_s"
    show "(Fun f (bef @ s # aft), Fun f (bef @ t # aft)) \<in> inter_s"
      unfolding inter_s_def
    proof (clarify)
      fix \<alpha> :: "('v,'a)p_ass"
      assume wf_ass: "wf_ass \<alpha>" and pos_ass: "pos_ass \<alpha>" and apos_ass: "apos_ass \<alpha>"
      let ?in = "\<lambda> t :: ('b,'v) term. (pi\<guillemotleft>\<alpha>,t\<guillemotright>)"
      let ?inter = "\<lambda> s bef cs. list_sum (map (\<lambda> at. fst at * (?in (snd at))) (zip cs (bef @ s # aft)))"
      have int_s: "?in ?st = max0 (c + ?inter s bef cs)" by (auto simp: ccs)
      have int_t: "?in ?tt = max0 (c + ?inter t bef cs)" by (auto simp: ccs)	
      from mono wf_cs have main: "?inter s bef cs > ?inter t bef cs" 
      proof (induct cs arbitrary: bef)
	case Nil thus ?case by auto
      next
	case (Cons c cs) note oCons = this
	hence wf_c: "wf_elem c" and c_zero: "c \<ge> 0" by auto
        {
	  fix aft and a :: 'a and b :: "('b,'v)term"
	  assume "(a,b) \<in> set (zip cs aft)"
	  hence "a \<in> set cs" by (rule set_zip_leftD)
	  with Cons have "wf_elem a" by auto
	  hence "wf_elem (a * ?in b)" using wf_ass by auto
        } hence wf_rec: "\<And> aft. wf_elem (list_sum (map (\<lambda>at. fst at * (pi\<guillemotleft>\<alpha>,snd at\<guillemotright>)) (zip cs aft)))"
	  by (intro wf_list_sum, auto)
	show ?case
	proof (cases bef)
	  case Nil
	  from st have st: "?in s > ?in t" unfolding inter_s_def using wf_ass pos_ass apos_ass by auto
          from Cons have "check_mono c \<or> c = 1" by auto
          hence cst: "c * ?in s > c * ?in t"
          proof
            assume "c = 1"
            with st show ?thesis by (simp add: wf_ass)
          next
            assume "check_mono c"
            thus ?thesis
              by (rule check_mono[OF _ st], auto simp: wf_ass wf_c c_zero)
          qed
	  show ?thesis 
            by (simp add: Nil Cons, rule plus_gt_left_mono2, rule cst, auto simp: Cons wf_ass plus_single_mono, rule geq_refl,(rule wf_rec)+) 
	next
	  case (Cons b befs)
	  with oCons have "Suc (length befs + length aft) \<le> length cs" by auto
	  with Cons oCons have rec: "?inter s befs cs > ?inter t befs cs" by auto
	  show ?thesis by (simp add: Cons oCons, rule plus_gt_right_mono[OF rec, simplified, OF _ wf_rec wf_rec],  auto simp: wf_c wf_ass)
	qed
      qed
      have wf_inter: "\<And> t. wf_elem (?inter t bef cs)"
      proof (rule wf_list_sum, auto)
	fix t a b
	assume "(a,b) \<in> set (zip cs (bef @ t # aft))"
	hence "a \<in> set cs" by (rule set_zip_leftD)
	with wf_cs wf_ass show "wf_elem (a * (?in b))" by auto
      qed	
      have wf_cint: "\<And> t. wf_elem (c + ?inter t bef cs)"
	by (rule wf_plus, rule wf_c, rule wf_inter)
      have pos_inter: "\<And> t. (?inter t bef cs) \<ge> 0"
      proof (rule pos_list_sum, (clarify, simp)+)
	fix t a b 
	assume "(a,b) \<in> set (zip cs (bef @ t # aft))"
	hence "a \<in> set cs" by (rule set_zip_leftD)
	hence pos_a: "a \<ge> 0" and wf_a: "wf_elem a" using wf_cs  by auto
	show "wf_elem (a * (?in b)) \<and> (a * (?in b) \<ge> 0)"
	proof (rule conjI, auto simp: wf_ass wf_a)
	  have int: "(a * (?in b) \<ge> a * 0)"
	    by (rule times_right_mono, rule pos_a, rule pos_term, auto simp: wf_a wf_ass pos_ass)
	  show "(a * (?in b)) \<ge> 0" by (rule geq_trans[where y = "a * 0"], rule int, auto simp: wf_a wf_ass)
	qed
      qed
      have "\<And> t. c + (?inter t bef cs) \<ge> 0 + 0"
	by (rule plus_left_right_mono, auto simp: mono, rule pos_inter, rule wf_c, rule wf_inter)
      hence pos_cint: "\<And> t. c + (?inter t bef cs) \<ge> 0" by auto
      have id_cint: "\<And> t. max0(c + (?inter t bef cs)) = c + (?inter t bef cs)"
	by (rule max0_id_pos, rule pos_cint, rule wf_cint)
      {
	fix t aft and a :: 'a and b :: "('b,'v)term"
	assume "(a,b) \<in> set (zip cs (bef @ t # aft))"
	hence "a \<in> set cs" by (rule set_zip_leftD)
	with wf_cs have "wf_elem a" by auto
	hence "wf_elem (a * ?in b)" using wf_ass by auto
      }      
      hence wf_rec: "\<And> t aft. wf_elem (list_sum (map (\<lambda>at. fst at * (pi\<guillemotleft>\<alpha>,snd at\<guillemotright>)) (zip cs (bef @ t # aft))))"
        by (intro wf_list_sum, auto)
      show "?in ?st > ?in ?tt" by (simp only: int_s int_t id_cint, rule plus_gt_right_mono, rule main, rule wf_c, (rule wf_rec)+)
    qed
  qed
qed
end

fun
  check_lpoly_coeffs :: "('a::show \<Rightarrow> bool) \<Rightarrow> ('a \<Rightarrow> bool) \<Rightarrow> ('a \<Rightarrow> bool) \<Rightarrow> ('f::show, 'a) lpoly_interL \<Rightarrow> shows check"
where
  "check_lpoly_coeffs wfe apos ch I =
     check_allm (\<lambda>((f, n), (c, cs)). do {
       check (wfe c) (''constant part '' +#+ shows c +@+ '' is not well-formed'' +#+ shows_nl);
       check (apos c \<or> list_ex apos (take n cs)) (''could not find positive entry which is required for arctic interpretations'' +#+ shows_nl);
       check_allm (\<lambda>a. check (ch a \<and> wfe a) (''coefficient '' +#+ shows a +@+ '' is not allowed'' +#+ shows_nl)) cs
     } <+?  (\<lambda>s. ''problem with interpretation of '' +#+ shows f +@+ ''/'' +#+ shows n +@+ shows_nl +@+ s)
     ) I"

context poly_order 
begin
lemmas compat_orig = compat
end

sublocale poly_order \<subseteq> redtriple_order inter_s inter_ns inter_ns
proof
  show "SN inter_s"
  proof
    fix f assume steps: "(\<forall> i. (f i, f (Suc i)) \<in> inter_s)"
    let ?b = default_ass
    have "\<forall> i. ((pi\<guillemotleft>?b,f i\<guillemotright>) > (pi\<guillemotleft>?b, f (Suc i)\<guillemotright>)) \<and> (arc_pos (pi\<guillemotleft>?b, f (Suc i)\<guillemotright>))"
    proof 
      fix i
      from steps have "(f i, f (Suc i)) \<in> inter_s" .. 
      hence dec: "(pi\<guillemotleft>?b, f i\<guillemotright>) >  (pi\<guillemotleft>?b, f (Suc i)\<guillemotright>)" (is ?dec) unfolding inter_s_def 
        by auto
      show "?dec \<and> arc_pos (pi\<guillemotleft>?b, f (Suc i)\<guillemotright>)" by (rule conjI, rule dec, rule apos_term, auto)
    qed
    from this obtain g
      where nSN: "\<forall> i. gt ((g i) :: 'a) (g (Suc i)) \<and> wf_elem (g i) \<and> wf_elem (g (Suc i))
      \<and> geq (g (Suc i)) zero \<and> arc_pos (g (Suc i))" by force
    with SN show False by (auto simp: SN_defs)
  qed
next
  show "inter_ns O inter_s \<subseteq> inter_s"
  proof (clarify)
    fix s t u :: "('f,'v)term"
    assume st: "(s,t) \<in> inter_ns" and tu: "(t,u) \<in> inter_s"
    show "(s,u) \<in> inter_s" unfolding inter_s_def
    proof(rule, unfold split, clarify)
      fix \<alpha> :: "'v \<Rightarrow> 'a"
      assume wf: "wf_ass \<alpha>" "pos_ass \<alpha>" "apos_ass \<alpha>"
      show "pi\<guillemotleft>\<alpha>,s\<guillemotright> > pi\<guillemotleft>\<alpha>,u\<guillemotright>"
        by (rule compat, insert st tu wf, unfold inter_s_def inter_ns_def, auto intro: wf_terms)
    qed
  qed
next
  show "inter_s O inter_ns \<subseteq> inter_s"
  proof (clarify)
    fix s t u :: "('f,'v)term"
    assume st: "(s,t) \<in> inter_s" and tu: "(t,u) \<in> inter_ns"
    show "(s,u) \<in> inter_s" unfolding inter_s_def
    proof(rule, unfold split, clarify)
      fix \<alpha> :: "'v \<Rightarrow> 'a"
      assume wf: "wf_ass \<alpha>" "pos_ass \<alpha>" "apos_ass \<alpha>"
      show "pi\<guillemotleft>\<alpha>,s\<guillemotright> > pi\<guillemotleft>\<alpha>,u\<guillemotright>"
        by (rule compat2, insert st tu wf, unfold inter_s_def inter_ns_def, auto intro: wf_terms)
    qed
  qed
next
  show "one_step_ctxt_closed inter_ns"
  proof (unfold one_step_ctxt_closed_def inter_ns_def, intro allI impI)
    fix f :: 'f
      and bef :: "('f,'v) term list" 
      and s :: "('f,'v) term" 
      and t :: "('f,'v) term" 
      and aft :: "('f,'v) term list" 
    let ?P = "{(s,t). \<forall> \<alpha>. wf_ass \<alpha> \<and> pos_ass \<alpha> \<and> apos_ass \<alpha> \<longrightarrow> (pi\<guillemotleft>\<alpha>,s\<guillemotright> \<ge> pi\<guillemotleft>\<alpha>,t\<guillemotright>)}"
    assume "(s, t) \<in> ?P"
    hence st: "\<forall> \<alpha>. wf_ass \<alpha> \<and> pos_ass \<alpha> \<and> apos_ass \<alpha> \<longrightarrow> (pi\<guillemotleft>\<alpha>, s\<guillemotright> \<ge> pi\<guillemotleft>\<alpha>, t\<guillemotright>)" by (simp)
    let ?inter = "pi f (Suc (length bef + length aft))"
    show "(Fun f (bef @ s # aft), Fun f (bef @ t # aft)) \<in> ?P"
    proof (cases "?inter")
      case (Pair c list)
      have "\<forall> \<alpha>. wf_ass \<alpha> \<and> pos_ass \<alpha> \<and> apos_ass \<alpha> \<longrightarrow> (pi\<guillemotleft>\<alpha>, Fun f (bef @ s # aft)\<guillemotright> \<ge> pi\<guillemotleft>\<alpha>,Fun f (bef @ t # aft)\<guillemotright>)" (is "\<forall> \<alpha>. ?ass \<alpha> \<longrightarrow> ?ge \<alpha>")
      proof 
        fix \<alpha>
        show "?ass \<alpha> \<longrightarrow> ?ge \<alpha>"
        proof 
	  assume "?ass \<alpha>"
	  hence wf_ass: "wf_ass \<alpha>" and pos_ass: "pos_ass \<alpha>" and apos_ass: "apos_ass \<alpha>" by auto
	  with st have sta: "pi\<guillemotleft>\<alpha>,s\<guillemotright> \<ge> pi\<guillemotleft>\<alpha>,t\<guillemotright>" by auto
	  from wf_pi have "wf_elem (fst ?inter) \<and> (\<forall> a. a \<in> set (snd ?inter) \<longrightarrow> wf_elem a \<and> geq a zero)"  by auto
	  with Pair have wf_c: "wf_elem c" and wf_list: "\<forall> a \<in> set list. wf_elem a" and pos_list: "\<forall> a \<in> set list. geq a zero" by auto
	  let ?expr = "\<lambda> list rest.  list_sum (map (\<lambda>at. fst at * (pi\<guillemotleft>\<alpha>,snd at\<guillemotright>)) (zip list rest))"
	  from wf_list pos_list have main: "(?expr list (bef @ s # aft) \<ge> ?expr list (bef @ t # aft)) \<and> wf_elem (?expr list (bef @ s # aft)) \<and> wf_elem (?expr list (bef @ t # aft))"
	  proof (induct list arbitrary:bef)
	    case (Cons a as) note oCons = this
	    hence wf_a: "wf_elem a" and pos_a: "a \<ge> 0" and wf_as: "\<forall> a \<in> set as. wf_elem a" and pos_as: "\<forall> a \<in> set as. a \<ge> 0"  by auto
	    thus ?case 
	    proof (cases bef)
	      case Nil
              {
	        fix aas aaft
	        assume "(aas,aaft) \<in> set (zip as aft)"
	        hence "aas \<in> set as" using set_zip_leftD[where x = aas and y = aaft] by auto
	        with wf_as have wfaas: "wf_elem aas" by auto
	        hence "wf_elem (aas * (pi\<guillemotleft>\<alpha>,aaft\<guillemotright>))" by (auto simp: wf_ass wfaas)
              }
	      hence wf_as_aft: "wf_elem (?expr as aft)" 
	        by (intro wf_list_sum, auto)
	      show ?thesis
	        by (rule conjI, simp add: Nil, rule plus_left_mono, rule times_right_mono, 
		  auto simp: wf_a pos_a wf_ass Nil wf_as_aft wf_a wf_ass sta)
	    next
	      case (Cons u us)
	      with oCons wf_as pos_as have wf_rec: "wf_elem (?expr as (us @ s # aft)) \<and> wf_elem (?expr as (us @ t # aft))" by best
	      show ?thesis 
	        by (simp add: Cons, rule conjI, rule plus_right_mono, auto simp: Cons oCons wf_ass wf_as pos_as)
	    qed
	  qed simp
	  show "?ge \<alpha>" by (simp add: Pair, rule max0_mono, rule plus_right_mono, auto simp: main wf_c)
        qed
      qed
      thus ?thesis by auto
    qed
  qed
next
  show "subst_closed inter_s" 
  proof (unfold subst_closed_def inter_s_def, rule subsetI)
    fix st_sig :: "('f,'v)term \<times> ('f,'v)term"
    assume st_sig: "st_sig \<in> CS {(s,t). \<forall>\<alpha>. wf_ass \<alpha> \<and> pos_ass \<alpha> \<and> apos_ass \<alpha> \<longrightarrow> (pi\<guillemotleft>\<alpha>,s\<guillemotright> > pi\<guillemotleft>\<alpha>,t\<guillemotright>) }" (is "_ \<in> CS ?P")
    thus "st_sig \<in> ?P"
    proof (cases st_sig)
      case (Pair s_sig t_sig)
      with st_sig have "(s_sig, t_sig) \<in> CS ?P" by simp
      hence "(s_sig, t_sig) \<in> ?P" 
      proof induct
        case (subst_closureI s t \<sigma>)
        show ?case 
        proof (simp del:interpr_subst, intro allI)
	  fix \<alpha> 
	  show "wf_ass \<alpha> \<and> pos_ass \<alpha> \<and> apos_ass \<alpha> \<longrightarrow> (pi\<guillemotleft>\<alpha>, s \<cdot> \<sigma>\<guillemotright> > pi\<guillemotleft>\<alpha>, t \<cdot> \<sigma>\<guillemotright>)" (is "?wf \<and> ?ps \<longrightarrow> ?go") 
	  proof -
	    {
	      assume ?wf and ?ps
	      have wf_assB: "wf_ass (\<lambda>x. pi\<guillemotleft>\<alpha>,\<sigma> x\<guillemotright>)" (is "wf_ass ?bet") unfolding wf_ass_def using `?wf`  by auto
	      have pos_assB: "pos_ass ?bet" unfolding pos_ass_def using `?ps` `?wf`  by auto
	      have apos_assB: "apos_ass ?bet" unfolding apos_ass_def using `?ps` `?wf` by auto
	      with Pair subst_closureI Cons wf_assB pos_assB have "pi\<guillemotleft>?bet,s\<guillemotright> > pi\<guillemotleft>?bet,t\<guillemotright>" by force
	      with `?wf` have "?go" by (simp add: interpr_subst)
	    }
	    thus ?thesis by blast
	  qed
        qed
      qed
      with Pair show ?thesis by simp
    qed
  qed
next
  show "subst_closed inter_ns" 
  proof (unfold subst_closed_def inter_ns_def, rule subsetI)
    fix st_sig :: "('f,'v)term \<times> ('f,'v)term"
    assume st_sig: "st_sig \<in> CS {(s,t). \<forall>\<alpha>. wf_ass \<alpha> \<and> pos_ass \<alpha> \<and> apos_ass \<alpha> \<longrightarrow> (pi\<guillemotleft>\<alpha>,s\<guillemotright> \<ge> pi\<guillemotleft>\<alpha>,t\<guillemotright>) }" (is "_ \<in> CS ?P")
    thus "st_sig \<in> ?P"
    proof (cases st_sig)
      case (Pair s_sig t_sig)
      with st_sig have "(s_sig, t_sig) \<in> CS ?P" by simp
      hence "(s_sig, t_sig) \<in> ?P" 
      proof induct
        case (subst_closureI s t \<sigma>)
        show ?case 
        proof (simp del:interpr_subst, intro allI)
	  fix \<alpha> 
	  show "wf_ass \<alpha> \<and> pos_ass \<alpha> \<and> apos_ass \<alpha> \<longrightarrow> (pi\<guillemotleft>\<alpha>, s \<cdot> \<sigma>\<guillemotright> \<ge> pi\<guillemotleft>\<alpha>, t \<cdot> \<sigma>\<guillemotright>)" (is "?wf \<and> ?ps \<longrightarrow> ?go") 
	  proof -
	    {
	      assume ?wf and ?ps
	      have wf_assB: "wf_ass (\<lambda>x. pi\<guillemotleft>\<alpha>, \<sigma> x\<guillemotright>)" (is "wf_ass ?bet") unfolding wf_ass_def using `?wf`  by auto
	      have pos_assB: "pos_ass ?bet" unfolding pos_ass_def using `?ps` `?wf`  by auto
	      have apos_assB: "apos_ass ?bet" unfolding apos_ass_def using `?ps` `?wf` by auto
	      with Cons subst_closureI Pair wf_assB pos_assB have "pi\<guillemotleft>?bet,s\<guillemotright> \<ge> pi\<guillemotleft>?bet,t\<guillemotright>" by force
	      with `?wf` have "?go" by (simp add: interpr_subst)
	    }
	    thus ?thesis by blast
	  qed
        qed
      qed
      with Pair show ?thesis by simp
    qed
  qed
next
  show "trans inter_ns" unfolding trans_def
  proof (clarify)
    fix s t u :: "('f,'v)term"
    assume st: "(s,t) \<in> inter_ns" and tu: "(t,u) \<in> inter_ns"
    show "(s,u) \<in> inter_ns" unfolding inter_ns_def
    proof(rule, unfold split, clarify)
      fix \<alpha> :: "'v \<Rightarrow> 'a"
      assume wf: "wf_ass \<alpha>" "pos_ass \<alpha>" "apos_ass \<alpha>"
      show "pi\<guillemotleft>\<alpha>,s\<guillemotright> \<ge> pi\<guillemotleft>\<alpha>,u\<guillemotright>"
        by (rule geq_trans, insert st tu wf, unfold inter_ns_def, auto intro: wf_terms)
    qed
  qed
next
  show "trans inter_s" unfolding trans_def
  proof (clarify)
    fix s t u :: "('f,'v)term"
    assume st: "(s,t) \<in> inter_s" and tu: "(t,u) \<in> inter_s"
    show "(s,u) \<in> inter_s" unfolding inter_s_def
    proof(rule, unfold split, clarify)
      fix \<alpha> :: "'v \<Rightarrow> 'a"
      assume wf: "wf_ass \<alpha>" "pos_ass \<alpha>" "apos_ass \<alpha>"
      show "pi\<guillemotleft>\<alpha>,s\<guillemotright> > pi\<guillemotleft>\<alpha>,u\<guillemotright>"
        by (rule gt_trans, insert st tu wf, unfold inter_s_def, auto intro: wf_terms)
    qed
  qed
next
  show "refl inter_ns"
    unfolding inter_ns_def refl_on_def using geq_refl by (auto intro: wf_terms)
qed

fun
  check_polo_sI
where
  "check_polo_sI mode ze on pl ti mz ge gt pi (s,t) = do {
     let left = PleftI ze on pl ti mz pi s;
     let right = PrightI ze on pl ti mz pi t;
     check_lpoly_sI mode ze on ge gt left right
       <+? (\<lambda>e. ''could not ensure '' +#+ shows s +@+ '' > '' +#+ shows t +@+ shows_nl +@+ e)
   }"

context poly_order
begin
definition coeffs_of_constraint :: "('f,'v)rule \<Rightarrow> ('a \<times> 'a) list"
  where "coeffs_of_constraint st \<equiv> [(a,b). a \<leftarrow> coeffs_of_lpoly (Pleft (fst st)), b \<leftarrow> coeffs_of_lpoly (Pright (snd st))]"

abbreviation check_polo_s_constrained :: "('a \<Rightarrow> 'a \<Rightarrow> bool) \<Rightarrow> ('f, 'v::show) rule \<Rightarrow> shows check"
where "check_polo_s_constrained weak_gt st \<equiv> if (\<forall> (a,b) \<in> set (coeffs_of_constraint st). wf_elem a \<longrightarrow> wf_elem b \<longrightarrow> weak_gt a b \<longrightarrow> gt a b) then check_polo_sI plus_single_mono zero one plus times max0 geq weak_gt pi st else error (shows ''internal error in check_polo_s'')" 

(* TODO: move *)
lemma wf_lpoly_coeff: assumes wf: "wf_lpoly p" and mem: "a \<in> set (coeffs_of_lpoly p)" shows "wf_elem a"
proof -
  obtain b as where p: "p = LPoly b as" by (cases p, auto)
  note wf = wf[unfolded p, simplified]
  show ?thesis
  proof (cases "a = b")
    case True 
    with wf show ?thesis by auto
  next
    case False
    with mem[unfolded p] have mem: "a \<in> set (coeffs_of_pvars as)" by auto
    from wf have "wf_pvars as" ..
    with mem show ?thesis
    proof (induct as)
      case Nil
      thus ?case unfolding coeffs_of_pvars_def by auto
    next
      case (Cons b bs)
      show ?case
      proof (cases "a = snd b")
        case True
        with Cons(3) show ?thesis unfolding wf_pvars_def  by auto
      next
        case False
        with Cons(2) have mem: "a \<in> set (coeffs_of_pvars bs)" unfolding coeffs_of_pvars_def by auto
        from False Cons(3) have wf: "wf_pvars bs" unfolding wf_pvars_def by auto
        from Cons(1)[OF mem wf] show ?thesis .
      qed
    qed
  qed
qed
        

lemma wf_elem_coeff_left: assumes "a \<in> set (coeffs_of_lpoly (Pleft t))" shows "wf_elem a"
proof -
  have "wf_ass default_ass" and "pos_ass default_ass" by auto
  from Pleft_both[OF this] have "wf_lpoly (Pleft t)" ..
  from wf_lpoly_coeff[OF this] show ?thesis using assms by auto
qed

lemma wf_elem_coeff_right: assumes "a \<in> set (coeffs_of_lpoly (Pright t))" shows "wf_elem a"
proof -
  have "wf_ass default_ass" and "pos_ass default_ass" by auto
  from Pright_both[OF this] have "wf_lpoly (Pright t)" by auto
  from wf_lpoly_coeff[OF this] show ?thesis using assms by auto
qed
end

sublocale poly_order \<subseteq> redtriple_impl "inter_s :: ('f :: show,'v :: show)trs" inter_ns inter_ns "check_polo_s_constrained weak_gt" check_polo_ns check_polo_ns
proof
  fix st :: "('f,'v)rule"
  assume check: "isOK(check_polo_s_constrained weak_gt st)"
  obtain s t where st: "st = (s,t)" by force
  note check = check[unfolded st]
  from check have weak_gt: "\<And> a b. (a,b) \<in> set (coeffs_of_constraint (s,t)) \<Longrightarrow> wf_elem a \<Longrightarrow> wf_elem b \<Longrightarrow> weak_gt a b \<Longrightarrow> a > b"  by auto
  from check have "isOK(check_polo_sI plus_single_mono 0 1 op + op * max0 geq weak_gt pi (s,t))" by simp
  show "st \<in> inter_s" unfolding st inter_s_def 
  proof (intro CollectI, simp, intro allI) 
    fix \<alpha>
    show "wf_ass \<alpha> \<and> pos_ass \<alpha> \<and> apos_ass \<alpha> \<longrightarrow> (pi\<guillemotleft>\<alpha>,s\<guillemotright> > pi\<guillemotleft>\<alpha>,t\<guillemotright>)" (is "?ass \<longrightarrow> ?goal")
    proof
      assume ?ass
      from check have check1: "isOK(check_lpoly_s weak_gt one (Pleft s) (Pright t))" by (auto simp: Let_def)
      have strict: "(Pleft s, Pright t) \<in> poly_s" 
        by (rule check_lpoly_s_sound[OF check1 _ _ weak_gt], auto simp: coeffs_of_constraint_def wf_elem_coeff_left wf_elem_coeff_right)  
      have ge1: "pi\<guillemotleft>\<alpha>,s\<guillemotright> \<ge> eval_lpoly \<alpha> (Pleft s)" by (rule Pleft_sound, auto simp: `?ass`)
      have gt: "\<dots> > eval_lpoly \<alpha> (Pright t)" using strict unfolding poly_s_def using `?ass` by auto
      have ge2: "\<dots> \<ge> pi\<guillemotleft>\<alpha>,t\<guillemotright>" by (rule Pright_sound, auto simp: `?ass`)
      show ?goal 
        by (rule compat_orig[OF ge1 compat2[OF gt ge2]], auto simp: `?ass` wf_eval_lpoly) 
    qed
  qed
next
  fix st :: "('f,'v)rule"
  assume check: "isOK(check_polo_ns st)"
  show "st \<in> inter_ns"
  proof (cases st)
    case (Pair s t)
    thus ?thesis unfolding inter_ns_def
    proof (simp, intro allI)
      fix \<alpha>
      show "wf_ass \<alpha> \<and> pos_ass \<alpha> \<and> apos_ass \<alpha> \<longrightarrow> (pi\<guillemotleft>\<alpha>,s\<guillemotright> \<ge> pi\<guillemotleft>\<alpha>,t\<guillemotright>)" (is "?ass \<longrightarrow> ?goal")
      proof
        assume ?ass
        from check Pair have check1: "isOK(check_lpoly_ns one (Pleft s) (Pright t))" by (auto simp: Let_def)
        have strict: "(Pleft s, Pright t) \<in> poly_ns" 
          by (rule check_lpoly_ns_sound, rule check1, auto)
        have ge1: "pi\<guillemotleft>\<alpha>,s\<guillemotright> \<ge> eval_lpoly \<alpha> (Pleft s)" by (rule Pleft_sound, auto simp: `?ass`)
        have gt: "\<dots> \<ge> eval_lpoly \<alpha> (Pright t)" using strict unfolding poly_ns_def using `?ass` by auto
        have ge2: "\<dots> \<ge> pi\<guillemotleft>\<alpha>,t\<guillemotright>" by (rule Pright_sound, auto simp: `?ass`)
        show ?goal by (rule geq_trans, rule ge1, rule geq_trans, rule gt, rule ge2, auto simp: `?ass` wf_eval_lpoly)
      qed
    qed
  qed
qed


record 'a poly_c =
  wf :: "'a \<Rightarrow> bool"
  zero :: 'a
  one :: 'a
  default :: 'a
  plus :: "'a \<Rightarrow> 'a \<Rightarrow> 'a"
  times :: "'a \<Rightarrow> 'a \<Rightarrow> 'a"
  max0 :: "'a \<Rightarrow> 'a"
  ge :: "'a \<Rightarrow> 'a \<Rightarrow> bool"
  gt :: "'a \<Rightarrow> 'a \<Rightarrow> bool"
  mono :: bool
  description :: string
  arc_pos :: "'a \<Rightarrow> bool"
  check_mono :: "'a \<Rightarrow> bool"

hide_const(open) wf zero one default plus times max0 ge gt mono description arc_pos check_mono

abbreviation
  check_polo_s_comp :: "('a::show) poly_c \<Rightarrow> ('f::show, 'a) lpoly_inter \<Rightarrow> ('f, 'v::show) rule \<Rightarrow> shows check"
where
  "check_polo_s_comp c \<equiv>
     check_polo_sI (poly_c.mono c) (poly_c.zero c) (poly_c.one c) (poly_c.plus c) (poly_c.times c) (poly_c.max0 c) (poly_c.ge c) (poly_c.gt c)"

abbreviation check_polo_ns_comp :: "('a::show)poly_c \<Rightarrow> ('f::show,'a)lpoly_inter \<Rightarrow> ('f,'v::show)rule \<Rightarrow> shows check"
where "check_polo_ns_comp c \<equiv> check_polo_nsI (poly_c.zero c) (poly_c.one c) (poly_c.plus c) (poly_c.times c) (poly_c.max0 c) (poly_c.ge c)"

abbreviation check_polo_coeffs_comp :: "('a::show)poly_c \<Rightarrow> ('f::show,'a)lpoly_interL \<Rightarrow> shows check"
where "check_polo_coeffs_comp c \<equiv> check_lpoly_coeffs (poly_c.wf c) (poly_c.arc_pos c) (\<lambda> x. (poly_c.ge c) x (poly_c.zero c))"

abbreviation check_polo_mono_comp :: "('a :: show)poly_c \<Rightarrow> ('f :: show, 'a)lpoly_interL \<Rightarrow> shows check" 
  where "check_polo_mono_comp c \<equiv> check_poly_monoI (poly_c.zero c) (poly_c.ge c) (poly_c.check_mono c)"

locale poly_order_impl = 
  fixes c :: "('a :: show) poly_c"
    and I :: "('f :: show, 'a)lpoly_interL"
    and d :: string
  assumes check_components: "isOK(check_polo_coeffs_comp c I)"
     and carrier: "\<And> as. \<exists> gt. poly_order_carrier (poly_c.wf c) (poly_c.zero c) (poly_c.plus c) (poly_c.times c) (poly_c.ge c) gt (poly_c.mono c) (poly_c.one c) (poly_c.default c) (poly_c.max0 c) (poly_c.arc_pos c) \<and> (\<forall> (a,b) \<in> set as. poly_c.wf c a \<longrightarrow> poly_c.wf c b \<longrightarrow> poly_c.gt c a b \<longrightarrow> gt a b) \<and> (poly_c.mono c \<longrightarrow>
    mono_poly_order_carrier (poly_c.wf c) (poly_c.zero c) (poly_c.plus c) (poly_c.times c) (poly_c.ge c) gt (poly_c.mono c) (poly_c.one c) (poly_c.default c) (poly_c.max0 c) (poly_c.arc_pos c) (poly_c.check_mono c))"
begin

abbreviation zero where "zero \<equiv> poly_c.zero c"
abbreviation one where "one \<equiv> poly_c.one c"
abbreviation default where "default \<equiv> poly_c.default c"
abbreviation times where "times \<equiv> poly_c.times c"
abbreviation plus where "plus \<equiv> poly_c.plus c"
abbreviation max0 where "max0 \<equiv> poly_c.max0 c"
abbreviation mono where "mono \<equiv> poly_c.mono c"
abbreviation arc_pos where "arc_pos \<equiv> poly_c.arc_pos c"
abbreviation wf where "wf \<equiv> poly_c.wf c"
abbreviation ge where "ge \<equiv> poly_c.ge c"
abbreviation check_mono where "check_mono \<equiv> poly_c.check_mono c"

abbreviation II :: "('f :: show,'a)lpoly_inter" where "II \<equiv> to_lpoly_inter default one I"

lemma poly_order: 
  shows "\<exists> S NS. ce_af_redtriple_order S NS NS (create_afI zero I) \<and> 
  (\<forall> st \<in> set sts. isOK(check_polo_s_comp c II st) \<longrightarrow> st \<in> S) \<and>
  (\<forall> st. isOK(check_polo_ns_comp c II st) \<longrightarrow> st \<in> NS) \<and>
  (mono \<longrightarrow> isOK(check_polo_mono_comp c I) \<longrightarrow> mono_ce_af_redtriple_order S NS NS (create_afI zero I) \<and> one_step_ctxt_closed S)"
proof -
  let ?pi = "create_afI zero I"
  let ?as = "[ab. st \<leftarrow> sts, ab \<leftarrow> poly_order.coeffs_of_constraint zero plus times one max0 II st]"
  from carrier[of ?as] obtain gt where carrier: 
    "poly_order_carrier wf zero plus times ge gt mono one default max0 arc_pos"
    and gt: "\<forall> (a,b) \<in> set ?as. wf a \<longrightarrow> wf b \<longrightarrow> poly_c.gt c  a b \<longrightarrow> gt a b" 
    and mono: "mono \<longrightarrow> mono_poly_order_carrier wf zero plus times ge gt mono one default max0 arc_pos check_mono" by auto
  interpret poly_order_carrier wf zero plus times ge gt mono one default max0 arc_pos by fact
  interpret poly_order wf zero plus times ge gt mono one default max0 arc_pos II
  proof
    show "wf_lpoly_inter II"
    proof (rule conjI, intro allI)
      fix f n 
      let ?goal = "\<lambda> f n. wf (fst (II f n)) \<and> (\<forall>a. a \<in> set (snd (II f n)) \<longrightarrow> wf a \<and> (ge a zero))"
      show "?goal f n"
      proof (cases "lookup (f,n) I")
        case None
        hence repl: "II f n = (default, replicate n one)" by auto
        have "\<forall> x \<in> set (replicate n one). x = one" by (induct n, auto)
        with repl show "?goal f n" by auto
      next
        case (Some aas)
        hence aas: "II f n = aas" by auto
        from Some have "((f,n),aas) \<in> set I" by (rule lookup_sound)
        with check_components have "wf (fst aas) \<and> (\<forall> a \<in> set (snd aas). wf a \<and> ge a zero)" by (cases aas, auto)
        with aas show "?goal f n" by auto
      qed
    next
      let ?goal2 = "\<lambda> f n. arc_pos (fst (II f n)) \<or>
        (\<exists>a\<in>set (take n (snd (II f n))). arc_pos a)"
      show "\<forall> f n. ?goal2 f n"
      proof (intro impI allI)
        fix f n
        show "?goal2 f n"
        proof (cases "lookup (f,n) I")
	  case None
	  hence "II f n = (default, replicate n one)" by auto
	  thus "?goal2 f n" by auto
        next
	  case (Some aas)
	  hence aas: "II f n = aas" by auto
	  from Some have "((f,n),aas) \<in> set I" by (rule lookup_sound)
	  with check_components have "arc_pos (fst aas) \<or> list_ex arc_pos (take n (snd aas))" by (cases aas, auto)
	  hence "arc_pos (fst aas) \<or> (\<exists> b \<in> set (take n (snd aas)) . arc_pos b)" by (simp add: list_ex_iff)
	  with aas show "?goal2 f n" by auto
        qed
      qed
    qed
  qed
  interpret ce_af_redtriple_order inter_s inter_ns inter_ns ?pi
  proof
    show "\<exists> n. \<forall> m\<ge>n . \<forall> c d. ce_trs (c,d,m) \<subseteq> inter_ns"
      by (rule inter_ns_ce_compat[OF refl])
  next
    show "af_compatible ?pi inter_ns"
      by (rule inter_ns_af_compat[OF refl])
  qed
  have order: "ce_af_redtriple_order inter_s inter_ns inter_ns ?pi" ..
  {
    fix st
    assume st: "st \<in> set sts" and ok: "isOK(check_polo_s_comp c II st)"
    have "st \<in> inter_s"
      by (rule checkSSound[where weak_gt = "poly_c.gt c"], insert ok gt st, auto)
  } note S = this
  show ?thesis
  proof (intro exI conjI allI ballI, 
      rule order, 
      (intro impI, rule S, auto)[1],
      (intro impI, rule checkNSSound, auto)[1],
      intro impI)
    assume mono and ok: "isOK(check_polo_mono_comp c I)"
    with mono interpret mono_poly_order_carrier wf zero plus times ge gt mono one default max0 arc_pos check_mono by auto
    interpret mono_poly_order wf zero plus times ge gt mono one default max0 arc_pos II check_mono
      by (unfold_locales, rule check_poly_mono_sound[OF ok])
    interpret mono_ce_af_redtriple_order inter_s inter_ns inter_ns ?pi
      by (unfold_locales, rule inter_s_ce_compat[OF refl])
    show "mono_ce_af_redtriple_order inter_s inter_ns inter_ns ?pi \<and> one_step_ctxt_closed inter_s"
      by (rule conjI[OF _ inter_s_mono], unfold_locales)
  qed
qed
end  

definition create_lpoly_repr :: "'a poly_c \<Rightarrow> ('f::show,'a::show)lpoly_interL \<Rightarrow> shows"
where "create_lpoly_repr c I = (let pi = to_lpoly_inter (poly_c.default c) (poly_c.one c) I in  (
  ''polynomial interpretration over '' +#+ poly_c.description c +#+ shows_nl +@+ (shows_sep (\<lambda>(f,n).
    let t = Fun f (map Var (fresh_strings_list ''x_'' 1 [] n)) in (
      ''Pol('' +#+ shows t +@+ '') = '' +#+
      shows_lpoly (poly_c.zero c) (poly_c.one c) (PleftI (poly_c.zero c) (poly_c.one c) (poly_c.plus c) (poly_c.times c) (poly_c.max0 c) pi t)
    )) shows_nl (remdups (map fst I)))))"

abbreviation create_polo_af_comp :: "('a :: show)poly_c \<Rightarrow> ('f,'a)lpoly_interL \<Rightarrow> 'f af"
  where "create_polo_af_comp c I \<equiv> create_afI (poly_c.zero c) I"

definition create_poly_redtriple :: "(('f,'a)lpoly_interL \<Rightarrow> shows check) \<Rightarrow> 'a poly_c \<Rightarrow> ('f :: show, 'a :: show)lpoly_interL \<Rightarrow> ('f,'v :: show)redtriple"
where "create_poly_redtriple cI c I = (let pi = to_lpoly_inter (poly_c.default c) (poly_c.one c) I; ns = check_polo_ns_comp c pi in 
    \<lparr>redtriple.valid = do {cI I; check_polo_coeffs_comp c I}, 
     s = check_polo_s_comp c pi, 
     ns = ns, nst = ns, 
     af = create_polo_af_comp c I,
     mono = do{ check (poly_c.mono c) (shows_string ''monotone polynomials not support for this domain''); check_polo_mono_comp c I}, 
     desc = create_lpoly_repr c I \<rparr>)"

lemma create_poly_redtriple:
  fixes cI :: "('f :: show, 'a :: show)lpoly_interL \<Rightarrow> shows check"
  assumes carrier: "\<And> I as. isOK(cI I) \<Longrightarrow> \<exists> gt. poly_order_carrier 
  (poly_c.wf c)
  (poly_c.zero c)
  (poly_c.plus c)
  (poly_c.times c)
  (poly_c.ge c)
  gt
  (poly_c.mono c)
  (poly_c.one c)
  (poly_c.default c)
  (poly_c.max0 c)
  (poly_c.arc_pos c) \<and> 
  (\<forall>(a, b)\<in>set as. poly_c.wf c a \<longrightarrow> poly_c.wf c b \<longrightarrow> poly_c.gt c a b \<longrightarrow> gt a b) \<and>
  (poly_c.mono c \<longrightarrow>
    mono_poly_order_carrier (poly_c.wf c) (poly_c.zero c) (poly_c.plus c) (poly_c.times c) (poly_c.ge c) gt (poly_c.mono c) (poly_c.one c) (poly_c.default c) (poly_c.max0 c) (poly_c.arc_pos c) (poly_c.check_mono c))"
  shows "generic_redtriple_impl (create_poly_redtriple cI c)"
proof
  fix I :: "('f,'a)lpoly_interL" and s_list ns_list nst_list :: "('f,'v :: show)rule list" and  \<pi>
  let ?c = "create_poly_redtriple cI c I :: ('f,'v)redtriple"
  let ?pi = "redtriple.af ?c :: 'f af"
  assume valid: "isOK (redtriple.valid ?c)" and 
    check_s: "isOK(check_allm (redtriple.s ?c) s_list)" and 
    check_ns: "isOK(check_allm (redtriple.ns ?c) ns_list)" and
    check_nst: "isOK(check_allm (redtriple.nst ?c) nst_list)"
  from valid have carrier_valid: "isOK(cI I)" and polo_valid: "isOK(check_polo_coeffs_comp c I)" unfolding create_poly_redtriple_def by (auto simp: Let_def)
  let ?I = "to_lpoly_inter (poly_c.default c) (poly_c.one c) I :: ('f,'a)lpoly_inter"
  interpret  poly_order_impl c I 
    by (unfold_locales, rule polo_valid, rule carrier[OF carrier_valid]) 
  from poly_order[of s_list] obtain S NS where order: "ce_af_redtriple_order S NS NS ?pi" 
    and S: "(\<forall> st \<in> set s_list. isOK(check_polo_s_comp c ?I st) \<longrightarrow> st \<in> S)" 
    and NS: "(\<forall> st. isOK(check_polo_ns_comp c ?I st) \<longrightarrow> st \<in> NS)" 
    and mono: "mono \<Longrightarrow> isOK(check_polo_mono_comp c I) \<Longrightarrow> mono_ce_af_redtriple_order S NS NS ?pi \<and> one_step_ctxt_closed S" unfolding create_poly_redtriple_def Let_def redtriple.simps by blast
  from check_s[unfolded create_poly_redtriple_def Let_def] S have S: "set s_list \<subseteq> S" by auto
  from check_nst[unfolded create_poly_redtriple_def Let_def] NS have NST: "set nst_list \<subseteq> NS" by auto
  from check_ns[unfolded create_poly_redtriple_def Let_def] NS have NS: "set ns_list \<subseteq> NS" by auto
  show "\<exists> S NS NST. ce_af_redtriple_order S NS NST ?pi \<and> set s_list \<subseteq> S \<and> set ns_list \<subseteq> NS \<and> set nst_list \<subseteq> NST \<and> (isOK (redtriple.mono ?c) \<longrightarrow> mono_ce_af_redtriple_order S NS NST ?pi \<and> one_step_ctxt_closed S)"
  proof (intro exI conjI, rule order, rule S, rule NS, rule NST, intro impI)
    assume "isOK(redtriple.mono ?c)"
    from this[unfolded create_poly_redtriple_def Let_def] have mono 
      and ok: "isOK(check_polo_mono_comp c I)" by auto
    from mono[OF this] 
    show "mono_ce_af_redtriple_order S NS NS ?pi \<and> one_step_ctxt_closed S" .
  qed
qed
    
context poly_order_carrier
begin
declare poly_simps[simp del] poly_order_simps[simp del]
end



end
