(*
Copyright 2011 Christian Sternagel, René Thiemann

This file is part of IsaFoR/CeTA.

IsaFoR/CeTA is free software: you can redistribute it and/or modify it under the
terms of the GNU Lesser General Public License as published by the Free Software
Foundation, either version 3 of the License, or (at your option) any later
version.

IsaFoR/CeTA is distributed in the hope that it will be useful, but WITHOUT ANY
WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR A
PARTICULAR PURPOSE.  See the GNU Lesser General Public License for more details.

You should have received a copy of the GNU Lesser General Public License along
with IsaFoR/CeTA. If not, see <http://www.gnu.org/licenses/>.
*)
theory List_Order
imports Abstract_Rewriting
begin

(* TODO: move *)

lemma trans_refl_imp_rtrancl_id:
  assumes "trans r" "refl r"
  shows "r^* = r"
proof
  show "r^* \<subseteq> r"
  proof
    fix x y
    assume "(x,y) \<in> r^*"
    thus "(x,y) \<in> r"
      by (induct, insert assms, unfold refl_on_def trans_def, blast+)
  qed
qed regexp

locale pre_order_pair = 
  fixes S :: "'a ars"
    and NS :: "'a ars"
  assumes refl_NS: "refl NS"
    and trans_S: "trans S"
    and trans_NS: "trans NS"
begin

lemma rtrancl_NS: "NS^* = NS" 
  by (rule trans_refl_imp_rtrancl_id[OF trans_NS refl_NS])

lemma trans_S_point: "\<And> x y z. (x,y) \<in> S \<Longrightarrow> (y,z) \<in> S \<Longrightarrow> (x,z) \<in> S"
  using trans_S unfolding trans_def by blast

lemma trans_NS_point: "\<And> x y z. (x,y) \<in> NS \<Longrightarrow> (y,z) \<in> NS \<Longrightarrow> (x,z) \<in> NS"
  using trans_NS unfolding trans_def by blast
end

locale compat_pair = 
  fixes S NS :: "'a ars" 
  assumes compat_NS_S: "NS O S \<subseteq> S"
    and compat_S_NS: "S O NS \<subseteq> S"
begin
lemma compat_NS_S_point: "\<And> x y z. (x,y) \<in> NS \<Longrightarrow> (y,z) \<in> S \<Longrightarrow> (x,z) \<in> S"
  using compat_NS_S by blast

lemma compat_S_NS_point: "\<And> x y z. (x,y) \<in> S \<Longrightarrow> (y,z) \<in> NS \<Longrightarrow> (x,z) \<in> S"
  using compat_S_NS by blast
end

locale order_pair = pre_order_pair S NS + compat_pair S NS 
  for S NS :: "'a ars"

locale SN_rel = fixes S :: "'a ars"
  assumes SN: "SN S"

locale SN_pair = compat_pair S NS + SN_rel S for S NS :: "'a ars"

locale SN_order_pair = order_pair S NS + SN_rel S for S NS :: "'a ars"

sublocale SN_order_pair \<subseteq> SN_pair ..

type_synonym 'a list_ext = "'a ars \<Rightarrow> 'a ars \<Rightarrow> 'a list ars"

locale list_order_extension =
  fixes s_list :: "'a list_ext"
   and ns_list :: "'a list_ext"
  assumes extension: "SN_order_pair S NS \<Longrightarrow> SN_order_pair (s_list S NS) (ns_list S NS)"
     and s_map: "\<lbrakk>\<And> a b. (a,b) \<in> S \<Longrightarrow> (f a,f b) \<in> S; \<And> a b. (a,b) \<in> NS \<Longrightarrow> (f a,f b) \<in> NS\<rbrakk> \<Longrightarrow> (as,bs) \<in> s_list S NS \<Longrightarrow> (map f as, map f bs) \<in> s_list S NS" 
     and ns_map: "\<lbrakk>\<And> a b. (a,b) \<in> S \<Longrightarrow> (f a,f b) \<in> S; \<And> a b. (a,b) \<in> NS \<Longrightarrow> (f a,f b) \<in> NS\<rbrakk> \<Longrightarrow> (as,bs) \<in> ns_list S NS \<Longrightarrow> (map f as, map f bs) \<in> ns_list S NS" 
     and all_ns_imp_ns: "length as = length bs \<Longrightarrow> \<lbrakk>\<And> i. i < length bs \<Longrightarrow>  (as ! i, bs ! i) \<in> NS\<rbrakk> \<Longrightarrow> (as,bs) \<in> ns_list S NS"

end
