(*
Copyright 2011 Christian Sternagel, René Thiemann

This file is part of IsaFoR/CeTA.

IsaFoR/CeTA is free software: you can redistribute it and/or modify it under the
terms of the GNU Lesser General Public License as published by the Free Software
Foundation, either version 3 of the License, or (at your option) any later
version.

IsaFoR/CeTA is distributed in the hope that it will be useful, but WITHOUT ANY
WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR A
PARTICULAR PURPOSE.  See the GNU Lesser General Public License for more details.

You should have received a copy of the GNU Lesser General Public License along
with IsaFoR/CeTA. If not, see <http://www.gnu.org/licenses/>.
*)
theory "Size-Change_Termination_Processors_Impl"
imports
  QDP_Framework_Impl 
  "Size-Change_Termination_Processors"
  Generic_Usable_Rules_Impl
  "Q-Restricted_Rewriting_Impl"
  Edg_Impl
  Subterm_Criterion_Impl
begin

subsection {*Definitions of Processors*}

definition check_sct_entry where
  "check_sct_entry is_def S NST s t stri nstri \<equiv> do {
    check_no_var s;
    check_no_var t;
    check_no_defined_root is_def t;
    let m  = num_args t;
    let n  = num_args s;
    check_allm
      (\<lambda>i. check (i \<le> n) (''left-index to large'' +#+ shows i +@+ shows_nl))
      (remdups (map fst (stri @ nstri)));
    check_allm
      (\<lambda>j. check (j \<le> m)
        (''right-index to large or argument violates usable-rules condition'' +#+ shows j))
      (remdups (map snd (stri @ nstri)));
    let ss = args s;
    let ts = args t;
    check_allm
      (\<lambda>(i, j). check (isOK(S (get_arg s i, get_arg t j)))
                      (''problem with edge '' +#+ shows i +@+ '' -S-> '' +#+ shows j)) stri;
    check_allm
      (\<lambda>(i, j). check (isOK(NST (get_arg s i, get_arg t j)))
                      (''problem with edge '' +#+ shows i +@+ '' -NS-> '' +#+ shows j)) nstri
  } <+? (\<lambda>m. ''problems with DP '' +#+ shows_rule (s, t) +@+ shows_nl +@+ m)"

definition sct_entry_to_sts where
  "sct_entry_to_sts s t stri nstri \<equiv> let
      js = remdups (map snd (stri @ nstri))
      in map (\<lambda> j. (s, get_arg t j)) js"

definition
  sct_ur_af_proc ::
    "('dpp, 'f, string) dpp_ops \<Rightarrow> ('f::{linorder,show}, string) redtriple \<Rightarrow> 
     (('f, string) rule \<times> ((nat \<times> nat) list \<times> (nat \<times> nat) list)) list  \<Rightarrow> 
     ('f, string) rules option \<Rightarrow> 'dpp \<Rightarrow>
     shows check"
where
  "sct_ur_af_proc I rp Gs U_opt dpp \<equiv> do {
       redtriple.valid rp;
       let is_def = dpp_spec.is_defined I dpp;
       let \<pi> = redtriple.af rp;
       let S = redtriple.s rp;
       let NS = redtriple.ns rp;
       let NST = redtriple.nst rp;
       check_allm (\<lambda>((s,t),(stri,nstri)). check_sct_entry is_def S NST s t stri nstri) Gs;
       let sts = concat (map (\<lambda>((s,t),(stri,nstri)). sct_entry_to_sts s t stri nstri) Gs);
       U \<leftarrow> smart_usable_rules_checker_impl I dpp \<pi> U_opt sts;
       check_allm NS U
          <+? (\<lambda>s. ''problem when orienting usable rules'' +#+ shows_nl +@+ s);
       let P = dpp_ops.pairs I dpp;
       let cap1 = capRM_dpp I dpp;
       let cap2 = reverse_capRM_dpp I dpp;
       check_subseteq P (map fst Gs)
         <+? (\<lambda>s. ''there is no size-change graph for DP '' +#+ shows_rule s);
       check (check_SCT (\<lambda>(s,t,cs,ct) (u,v,cu,cv).
         Ground_Context.match ct u \<and> Ground_Context.match cu t) (map (\<lambda>((s,t),(stri,nstri)). let e = (s,t,cap2 s,cap1 t) in Scg e e stri nstri) Gs))
         (''size-change analysis failed'' +#+ shows_nl)
     } <+? (\<lambda>s. ''could not apply the size-change processor with the following '' +#+ shows_nl +@+
       redtriple.desc rp +@+ shows_nl +@+ ''for the following reason'' +#+ shows_nl +@+ s)
     "

definition
  sct_subterm_proc ::
    "('dpp, 'f::show, 'v::show) dpp_ops \<Rightarrow>
     (('f, 'v) rule \<times> ((nat \<times> nat) list \<times> (nat \<times> nat) list)) list \<Rightarrow>
     'dpp \<Rightarrow> shows check"
where
  "sct_subterm_proc I Gs dpp = do {
    let P      = dpp_ops.pairs I dpp;
    let is_def = dpp_spec.is_defined I dpp;
    let cap1   = capRM_dpp I dpp;
    let cap2   = reverse_capRM_dpp I dpp;
    check_subseteq P (map fst Gs)
      <+? (\<lambda>s. ''there is no size-change graph for the pair '' +#+ shows_rule s);
    check_allm (\<lambda>((s, t), (stri, nstri)). do {
      check_no_var s;
      check_no_var t;
      check_no_defined_root is_def t;
      let m = num_args t;
      let n = num_args s;
      check_allm (\<lambda>(i, j). check
        (i \<le> n \<and> j \<le> m \<and> isOK (check_supt (get_arg s i) (get_arg t j)))
        (''problem with edge '' +#+ shows i +@+ '' |> '' +#+ shows j +@+ shows_nl)) stri;
      check_allm (\<lambda>(i, j). check
        (i \<le> n \<and> j \<le> m \<and> isOK (check_supteq (get_arg s i) (get_arg t j)))
        (''problem with edge '' +#+ shows i +@+ '' |>= '' +#+ shows j +@+ shows_nl)) nstri
    } <+? (\<lambda>m. ''problem with pair '' +#+ shows_rule (s, t) +@+ shows_nl +@+ m)) Gs;
    check
      (check_SCT
        (\<lambda>(s, t, cs, ct) (u, v, cu, cv). Ground_Context.match ct u \<and> Ground_Context.match cu t)
        (map (\<lambda>((s, t), (stri, nstri)).
          let e = (s, t, cap2 s, cap1 t)
          in Scg e e stri nstri) Gs))
      (''size-change analysis failed'' +#+ shows_nl)
  } <+? (\<lambda>s. ''could not apply the size-change processor based on the subterm-relation''
          +#+ shows_nl +@+ s)"


subsection {*Soundness Proofs*}

context dpp_spec
begin

lemma sct_subterm_proc_sound:
  assumes check: "isOK (sct_subterm_proc I Gs d)"
  shows "finite_dpp (dpp_ops.dpp I d)"
proof (rule ccontr)
  assume not_finite: "\<not> finite_dpp (dpp_ops.dpp I d)"
  obtain P where P: "P = dpp_ops.pairs I d" by auto
  obtain R where R: "R = dpp_ops.rules I d" by auto
  obtain Q where Q: "Q = dpp_ops.Q I d" by auto
  from not_finite obtain s t \<sigma>
    where mchain: "min_ichain (dpp_ops.dpp I d) s t \<sigma>" by (auto simp: finite_dpp_def)
  let ?rm    = "dpp_ops.rules_map I d"
  let ?cap   = "capRM_dpp I d"
  let ?capRM = "reverse_capRM_dpp I d"  
  let ?entry = "\<lambda>s t. (s, t, Edg.cap ((set R)^-1) s, Edg.cap (set R) t)"
  let ?Gs    = "map (\<lambda>((s, t), (stri, nstri)). Scg (?entry s t) (?entry s t) stri nstri) Gs"
  let ?conn  = "\<lambda>(s, t, cs, ct) (u, v, cu, cv). Ground_Context.match ct u \<and> Ground_Context.match cu t"
  from check
    have 1: "set P \<subseteq> fst ` set Gs \<and> check_SCT ?conn ?Gs"
    unfolding sct_subterm_proc_def Let_def reverse_capRM_dpp capRM_dpp P R by simp
  let ?check2'' = "\<lambda>f g s t stri nstri.
    is_Fun s
    \<and> is_Fun t
    \<and> \<not> defined (set R) (the (root t))
    \<and> (\<forall>(i, j)\<in>set stri.
      i \<le> length (args s) \<and> j \<le> length (args t) \<and> f (get_arg s i, get_arg t j))
    \<and> (\<forall>(i, j)\<in>set nstri.
      i \<le> length (args s) \<and> j \<le> length (args t) \<and> g (get_arg s i, get_arg t j))"
  let ?check2' = "?check2'' (\<lambda>r. isOK (check_supt (fst r) (snd r)))
                            (\<lambda>r. isOK (check_supteq (fst r) (snd r)))"
  let ?check2  = "?check2'' (\<lambda>r. r \<in> op \<rhd>) (\<lambda>r. r \<in> op \<unrhd>)"
  from check have "\<forall>((s, t), (stri, nstri)) \<in> set Gs. ?check2' s t stri nstri"
    by (force simp: sct_subterm_proc_def Let_def
                    check_no_defined_root_sound[OF is_defined_sound,
                      unfolded rules_map_defined_conv, of d] R)
  hence 2: "\<forall>((s, t), (stri, nstri))\<in>set Gs. ?check2 s t stri nstri"
    unfolding split_def unfolding isOK_check_supt isOK_check_supteq
    unfolding fst_conv snd_conv .
  let ?tuple = "\<lambda>s t. is_Fun s \<and> is_Fun t \<and> (the (root t)) \<notin> defs (set R)"
  let ?graph = "\<lambda>s t. \<exists>stri nstri. (
    Scg (?entry s t) (?entry s t) stri nstri \<in> set ?Gs
    \<and> (\<forall>(i, j)\<in>set stri.
      i \<le> length (args s) \<and> j \<le> length (args t) \<and> get_arg s i \<rhd> get_arg t j)
    \<and> (\<forall>(i, j)\<in>set nstri.
      i \<le> length (args s) \<and> j \<le> length (args t) \<and> get_arg s i \<unrhd> get_arg t j))"
  {
    fix s t
    assume "(s, t) \<in> set P"
    with 1 obtain stri nstri
      where mem: "((s, t), (stri, nstri)) \<in> set Gs" by auto
    hence memGs: "Scg (?entry s t) (?entry s t) stri nstri \<in> set ?Gs" by force
    from mem and 2 have ch: "?check2 s t stri nstri" by auto
    from ch have tuple: "?tuple s t" unfolding defs_def by force
    from ch memGs have "?graph s t" unfolding split_def by auto
    with tuple have "?tuple s t \<and> ?graph s t" by auto
  }
  hence tuple_graph: "\<forall>(s, t)\<in>set P. ?tuple s t \<and> ?graph s t" by auto
  from tuple_graph have tuple: "\<forall>(s, t)\<in>set P. ?tuple s t" by auto
  from tuple_graph have graph: "\<forall>(s, t)\<in>set P. ?graph s t" by auto
  have "min_ichain (set P, {}, set Q, set R, {}) s t \<sigma>"
    unfolding P Q R
    by (rule min_ichain_mono[OF mchain[unfolded dpp_spec_sound]], unfold dpp_spec_sound, auto)
  from sct_with_subterm[OF this tuple graph]
    show False using 1 by simp
qed
end

lemma sct_ur_af_proc: assumes I: "dpp_spec I"
  and grp: "generic_redtriple_impl grp"
  and check: "isOK(sct_ur_af_proc I (grp rp) Gs U_opt d)"
  shows "finite_dpp (dpp_ops.dpp I d)"
proof -
  interpret dpp_spec I by fact
  let ?Pd = P
  let ?Rd = R
  interpret generic_redtriple_impl grp by fact
  show ?thesis
  proof (rule finite_under_var_cond_impl)
    let ?R = "dpp_ops.rules I d"
    assume var: "\<And> lr. lr \<in> set ?R \<Longrightarrow> is_Fun (fst lr)" 
    let ?rp = "grp rp"
    obtain strict nstrict nstrict_top \<pi> where ids: "redtriple.s ?rp = strict" and idn: "redtriple.ns ?rp = nstrict" and idnt: "redtriple.nst ?rp = nstrict_top" and idp: "redtriple.af ?rp = \<pi>" by auto
    obtain P where P: "P = dpp_ops.pairs I d" by auto
    obtain R where R: "R = dpp_ops.rules I d" by auto
    obtain Q where Q: "Q = dpp_ops.Q I d" by auto
    let ?cap = "capRM_dpp I d"
    let ?capRM = "reverse_capRM_dpp I d"
    let ?entry = "\<lambda> s t. (s,t,Edg.cap ((set ?R)^-1) s, Edg.cap (set ?R) t)"
    note check = check[unfolded sct_ur_af_proc_def, simplified, unfolded ids idn idnt idp]
    let ?sts = "concat (map (\<lambda>((s,t),(stri,nstri)). sct_entry_to_sts s t stri nstri) Gs)"
    let ?uchecker = "smart_usable_rules_checker_impl I d \<pi> U_opt ?sts"
    from check have "isOK(?uchecker)" by simp
    then obtain ur where uchecker: "?uchecker = return ur" by (cases ?uchecker, auto)
    note check = check[unfolded uchecker, simplified]
    let ?GGs = "Gs"
    let ?Gs = "(map (\<lambda> ((s,t),(stri,nstri)). Scg (?entry s t) (?entry s t) stri nstri) ?GGs)"
    let ?conn = "\<lambda>(s,t,cs,ct) (u,v,cu,cv). Ground_Context.match ct u \<and> Ground_Context.match cu t"
    let ?lhss = "concat (map (\<lambda> ((s,t),_). map (\<lambda> i. get_arg s i) [0 ..< Suc (length (args s))]) ?GGs)" 
    let ?rhss = "concat (map (\<lambda> ((s,t),_). map (\<lambda> i. get_arg t i) [0 ..< Suc (length (args t))]) ?GGs)" 
    let ?pairs = "concat (map (\<lambda> t. map (\<lambda> s. (s,t)) ?lhss) ?rhss)"
    let ?stri = "filter (\<lambda> r. isOK(strict r)) ?pairs"
    let ?nstri = "filter (\<lambda> r. isOK(nstrict r)) ?pairs"
    let ?nstri_top = "filter (\<lambda> r. isOK(nstrict_top r)) ?pairs"
    have stri: "isOK(check_allm strict ?stri)" by auto
    from check have valid: "isOK(redtriple.valid ?rp)" and check1: "set P \<subseteq> fst ` set ?GGs"
      "isOK(check_allm nstrict ur)"
      "isOK(check_allm nstrict_top ?nstri_top)"
      "check_SCT ?conn ?Gs"
      by (auto simp: Let_def P)
    from generate_redtriple[OF valid, simplified ids idn idnt, OF stri check1(2) check1(3)] obtain S NS NST where
      redtriple: "ce_af_redtriple_order S NS NST \<pi>" and
      S: "set ?stri \<subseteq> S" and NS: "set ur \<subseteq> NS" and NST: "set ?nstri_top \<subseteq> NST" by (auto simp: idp)
    interpret ce_af_redtriple_order S NS NST \<pi> by fact
    have ce: "ce_af_redtriple S NS NST \<pi>" ..
    from smart_usable_rules_checker_impl[OF I uchecker var] have ur: "smart_usable_rules_checker \<pi> (set Q) (rules d) U_opt (set ?sts) = Some ur" unfolding Q by auto
    note sct = generic_sct_redtriple[OF _ smart_usable_rules_checker NS ce _ _ check1(4)]
    let ?check2b = "\<lambda> s t stri nstri. is_Fun s \<and> is_Fun t \<and> \<not> defined (set R) (the (root t)) \<and> 
      (\<forall> i \<in> fst ` set (stri @ nstri).  i \<le> length (args s)) \<and>
      (\<forall> j \<in> snd ` set (stri @ nstri).  j \<le> length (args t)) \<and> 
      (\<forall> (i,j) \<in> set stri.  isOK(strict (get_arg s i,get_arg t j))) \<and>
      (\<forall> (i,j) \<in> set nstri. isOK(nstrict_top (get_arg s i,get_arg t j)))"
    let ?check2a = "\<lambda> s t stri nstri. is_Fun s \<and> is_Fun t \<and> \<not> defined (set R) (the (root t)) \<and> 
      (\<forall> i \<in> fst ` set (stri @ nstri).  i \<le> length (args s)) \<and>
      (\<forall> j \<in> snd ` set (stri @ nstri).  j \<le> length (args t)) \<and> 
      (\<forall> (i,j) \<in> set stri. (get_arg s i,get_arg t j) \<in> S) \<and>
      (\<forall> (i,j) \<in> set nstri. (get_arg s i,get_arg t j) \<in> NST)"
    let ?check2 = "\<lambda> s t stri nstri. is_Fun s \<and> is_Fun t \<and> \<not> defined (set R) (the (root t)) \<and> 
      (\<forall> i \<in> fst ` set (stri @ nstri).  i \<le> length (args s)) \<and>
      (\<forall> j \<in> snd ` set (stri @ nstri).  j \<le> length (args t)) \<and> 
      (\<forall> (i,j) \<in> set stri. (get_arg s i, get_arg t j) \<in> S) \<and>
      (\<forall> (i,j) \<in> set nstri. (get_arg s i, get_arg t j) \<in> NST)"
    let ?is_def = "dpp_spec.is_defined I d"
    note is_def = check_no_defined_root_sound[OF is_defined_sound, of d,
      unfolded R[symmetric] rules_map_defined_conv]
    { 
      fix s t stri nstri
      assume g: "((s,t),(stri,nstri)) \<in> set ?GGs"
      with check
      have "isOK(check_sct_entry ?is_def strict nstrict_top s t stri nstri)" by auto
      hence check2b: "?check2b s t stri nstri"
        unfolding check_sct_entry_def by (force simp: is_def Let_def simp del: set_append map_append)
      hence i: "\<And> ij. ij \<in> set stri \<union> set nstri \<Longrightarrow> fst ij \<le> length (args s)" 
        and j: "\<And> ij. ij \<in> set stri \<union> set nstri \<Longrightarrow> snd ij \<le> length (args t)" 
        by auto
      {
        fix i j
        assume "(i,j) \<in> set stri \<union> set nstri"
        from i[OF this] j[OF this] have i: "i \<le> length (args s)" and j: "j \<le> length (args t)" 
          by auto
        from i have "i \<in> set [0 ..< Suc (length (args s))]" by auto
        hence "get_arg s i \<in> set (map (get_arg s) [0..<Suc (length (args s))])" by auto
        hence i: "get_arg s i \<in> set ?lhss" using g by force 
        from j have "j \<in> set [0 ..< Suc (length (args t))]" by auto
        hence "get_arg t j \<in> set (map (get_arg t) [0..<Suc (length (args t))])" by auto
        hence j: "get_arg t j \<in> set ?rhss" using g by force
        from i j have "(get_arg s i, get_arg t j) \<in> set ?pairs" by (simp,blast)
      } note ij_pairs = this
      have stri: "\<forall> (i,j) \<in> set stri. (get_arg s i, get_arg t j) \<in> S"
      proof (intro ballI, clarify)
        fix i j
        assume ij: "(i,j) \<in> set stri"
        from ij check2b have ok: "isOK(strict (get_arg s i,get_arg t j))" by auto
        from ij have ij: "(i,j) \<in> set stri \<union> set nstri" by auto
        show "(get_arg s i, get_arg t j) \<in> S"
          by (rule set_mp[OF S], insert ok ij_pairs[OF ij], simp)        
      qed
      have nstri: "\<forall> (i,j) \<in> set nstri. (get_arg s i, get_arg t j) \<in> NST"
      proof (intro ballI, clarify)
        fix i j
        assume ij: "(i,j) \<in> set nstri"
        from NST have NST: "set ?nstri_top \<subseteq> NST" by blast
        from ij check2b have ok: "isOK(nstrict_top (get_arg s i,get_arg t j))" by auto
        from ij have ij: "(i,j) \<in> set stri \<union> set nstri" by auto
        show "(get_arg s i, get_arg t j) \<in> NST"
          by (rule set_mp[OF NST], insert ok ij_pairs[OF ij], simp)
      qed
      from check2b stri nstri
      have check2a: "?check2a s t stri nstri" by blast
      hence "?check2 s t stri nstri" by blast
    } note check2 = this
    let ?tuple = "\<lambda> s t. is_Fun s \<and> is_Fun t \<and> (the (root t)) \<notin> defs (set R)"
    let ?pgraph = "\<lambda> s t stri non_stri. (
      Scg (?entry s t) (?entry s t) stri non_stri \<in> set ?Gs \<and>
      (\<forall> (i,j) \<in> set stri. i \<le> length (args s) \<and> j \<le> length (args t) \<and> (get_arg s i, get_arg t j) \<in> S) \<and> 
      (\<forall> (i,j) \<in> set non_stri. i \<le> length (args s) \<and> j \<le> length (args t) \<and> (get_arg s i, get_arg t j) \<in> NST))"
    let ?graph = "\<lambda> s t. \<exists> stri non_stri. ?pgraph s t stri non_stri"
    {
      fix s t
      assume "(s,t) \<in> set P"
      with check1 obtain stri nstri where mem: "((s,t),(stri,nstri)) \<in> set ?GGs" by auto
      hence memGs: "Scg (?entry s t) (?entry s t) stri nstri \<in> set ?Gs" by force
      from check2[OF mem] have ch: "?check2 s t stri nstri" by blast
      from ch have stri_ur: "\<forall> (i,j) \<in> set stri. i \<le> length (args s) \<and> j \<le> length (args t)"
        unfolding is_Fun_Fun_conv by force
      from ch have nstri_ur: "\<forall> (i,j) \<in> set nstri. i \<le> length (args s) \<and> j \<le> length (args t)"
        unfolding is_Fun_Fun_conv by force
      from ch have tuple: "?tuple s t" unfolding defs_def by force
      from ch memGs stri_ur nstri_ur have "?pgraph s t stri nstri" by auto
      hence "?graph s t" by blast
      with tuple have "?tuple s t \<and> ?graph s t" by auto
    }
    hence  tuple_graph:  "\<forall> (s,t) \<in> set P. ?tuple s t \<and> ?graph s t" by auto
    from tuple_graph have tuple: "\<forall> (s,t) \<in> set P. ?tuple s t" by blast
    note sct = sct[OF tuple] 
    from tuple_graph have graph: "\<forall> (s,t) \<in> set P. ?graph s t" by blast
    note sct = sct[unfolded R, OF _ graph]
    note sct = sct[where U_opt = U_opt]
    show ?thesis unfolding dpp_spec_sound
    proof (rule sct, unfold ur[symmetric] Q[symmetric], 
      rule arg_cong[where f = "smart_usable_rules_checker \<pi> (set Q) (rules d) U_opt"],
      unfold set_map)
      note sts = sct_entry_to_sts_def[unfolded Let_def]
      let ?one_cond = "\<lambda> s t j.      \<exists>cs ct st ns.
        Scg (s, t, cs, ct) (s, t, cs, ct) st ns
        \<in> (\<lambda>((s, t), x, y).
              Scg (?entry s t) (?entry s t) x y) `
          set Gs \<and>
        j \<in> snd ` set st \<union> snd ` set ns"
      let ?one = "{(s, get_arg t j) | s t j . ?one_cond s t j}"
      {
        fix s t j
        assume "?one_cond s t j"
        then obtain cs ct st ns
          where scg: "Scg (s,t,cs,ct) (s,t,cs,ct) st ns \<in> (\<lambda>((s,t), st, ns). Scg (?entry s t) (?entry s t) st ns) ` set Gs" and j: "j \<in> snd ` set st \<union> snd ` set ns" by blast
        from scg have scg: "((s,t),st,ns) \<in> set Gs" by auto
        have "(s, get_arg t j) \<in> set ?sts" unfolding set_concat set_map
          by (rule, rule, rule scg, unfold sts, insert j, auto)
      }
      hence one: "?one \<subseteq> set ?sts" by blast
      {
        fix s tj
        assume "(s,tj) \<in> set ?sts"
        from this[unfolded sts] obtain t st ns j where mem: "((s,t),st,ns) \<in> set Gs" and j: "j \<in> snd ` set st \<union> snd ` set ns" and tj: "tj = get_arg t j" by force
        have "(s,tj) \<in> ?one" unfolding tj
          by (rule, intro exI conjI, rule refl, rule image_eqI[OF _ mem], (simp, intro conjI, (rule refl)+), rule j)
      }
      hence two: "set ?sts \<subseteq> ?one" ..
      from one two show "?one = set ?sts" by blast
    next
      show "set P = set (?Pd d) \<union> set (Pw d)" unfolding P by simp
    next
      show "set (rules d) = set (?Rd d) \<union> set (Rw d)" by simp
    qed
  qed
qed


end