(*
Copyright 2011 Christian Sternagel, René Thiemann

This file is part of IsaFoR/CeTA.

IsaFoR/CeTA is free software: you can redistribute it and/or modify it under the
terms of the GNU Lesser General Public License as published by the Free Software
Foundation, either version 3 of the License, or (at your option) any later
version.

IsaFoR/CeTA is distributed in the hope that it will be useful, but WITHOUT ANY
WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR A
PARTICULAR PURPOSE.  See the GNU Lesser General Public License for more details.

You should have received a copy of the GNU Lesser General Public License along
with IsaFoR/CeTA. If not, see <http://www.gnu.org/licenses/>.
*)
theory Tree_Automata_Impl
imports
  Trs_Impl
  Tree_Automata
  Linear_Orders
begin


datatype ('q,'f)ta_rule_impl = TA_rule_impl 'f "'q list" 'q "'q rs" 

fun ta_rule_conv :: "('q,'f)ta_rule_impl \<Rightarrow> ('q,'f)ta_rule" 
  where "ta_rule_conv (TA_rule_impl f qs q qs') = (TA_rule f qs q)"

fun conv_ta_rule :: "('q \<Rightarrow> 'q rs) \<Rightarrow> ('q,'f)ta_rule \<Rightarrow> ('q,'f)ta_rule_impl" 
  where "conv_ta_rule eps (TA_rule f qs q) = (TA_rule_impl f qs q (eps q))"

record ('q,'f)ta_impl =
  ta_final_impl :: "'q rs"
  ta_rules_impl :: "('f \<times> nat, ('q,'f) ta_rule_impl list) rm"
  ta_rqs_impl :: "'q list"
  ta_rqs_set :: "'q rs"
  ta_eps_impl :: "('q \<times> 'q) list"
  ta_epss_impl :: "'q \<Rightarrow> 'q rs"
  ta_epsrs_impl :: "'q \<Rightarrow> 'q rs"


definition ta_of :: "('q :: linorder,'f :: linorder)ta_impl \<Rightarrow> ('q,'f)ta"
  where "ta_of TA = \<lparr> ta_final = set (rs_to_list (ta_final_impl TA)), 
  ta_rules = \<Union>{ta_rule_conv ` (set (rm_set_lookup (ta_rules_impl TA) fn)) | fn. True},
  ta_eps = set (ta_eps_impl TA)\<rparr>"

fun rqs_impl where "rqs_impl (TA_rule_impl f qs q qs') = qs"
fun rq_impl where "rq_impl (TA_rule_impl f qs q qs') = q"
fun rqss_impl where "rqss_impl (TA_rule_impl f qs q qs') = qs'"
fun rsym_impl where "rsym_impl (TA_rule_impl f qs q qs') = (f,length qs)"

fun ta_res_impl :: "('f \<times> nat,('q :: linorder,'f :: linorder)ta_rule_impl list)rm \<Rightarrow> ('q \<Rightarrow> 'q rs) \<Rightarrow> ('f,'q)term \<Rightarrow> 'q rs"
where "ta_res_impl TA eps (Var q) = eps q"
  | "ta_res_impl TA eps (Fun f ts) = (
        let rec = map (ta_res_impl TA eps) ts;
            n = length ts;
            rules = rm_set_lookup TA (f,n);
            arules = filter (\<lambda> rule. let qs = rqs_impl rule in list_all (\<lambda> i. rs_memb (qs ! i) (rec ! i)) [0 ..< n]) rules;
            qs = map (\<lambda> rule. rqss_impl rule) arules   
           in rs_Union qs)"



locale ta_inv = rm_set "ta_rules_impl TA" rsym_impl "conv_ta_rule (ta_epss_impl TA) ` ta_rules (ta_of TA)" for
  TA :: "('q :: linorder,'f :: linorder)ta_impl" +
  assumes ta_rqs_set: "rs_\<alpha> (ta_rqs_set TA) = ta_rqs (ta_of TA)"
  and     ta_rqs_impl: "set (ta_rqs_impl TA) = ta_rqs (ta_of TA)"
  and     ta_epss: "\<And> q. rs_\<alpha> (ta_epss_impl TA q) = {q'. (q,q') \<in> (ta_eps (ta_of TA))^*}"
  and     ta_epsrs: "\<And> q. rs_\<alpha> (ta_epsrs_impl TA q) = {q'. (q',q) \<in> (ta_eps (ta_of TA))^*}"
  and     ta_eps: "ta_eps (ta_of TA) = set (ta_eps_impl TA)"
begin
abbreviation ta_rm_set_lookup where "ta_rm_set_lookup \<equiv> rm_set_lookup (ta_rules_impl TA)"
abbreviation TA' :: "('q,'f)ta" where "TA' \<equiv> ta_of TA"

lemma rm_set_lookup1: assumes "rule \<in> set (ta_rm_set_lookup (f,n))"
  shows "\<exists> qs q qs'. rule = (TA_rule_impl f qs q qs') \<and> n = length qs \<and> rs_\<alpha> qs' = {q'. (q,q') \<in> (ta_eps TA')^*} \<and> (f qs \<rightarrow> q) \<in> ta_rules TA'"
proof -
  from assms have "rsym_impl rule = (f,n)" and mem: "rule \<in> conv_ta_rule (ta_epss_impl TA) ` ta_rules TA'"
    using assms unfolding rm_set_lookup by auto
  then obtain qs q qs' where n: "n = length qs" and rule: "rule = TA_rule_impl f qs q qs'" by (cases rule, auto)
  from mem obtain rrule where id: "rule = conv_ta_rule (ta_epss_impl TA) rrule" and rrule: "rrule \<in> ta_rules TA'" by auto
  from id[unfolded rule] have idd: "rrule = (f qs \<rightarrow> q)" by (cases rrule, auto) 
  from id[unfolded rule idd] have "rs_\<alpha> qs' = rs_\<alpha> (ta_epss_impl TA q)" by auto
  from this[unfolded ta_epss] rrule
  show ?thesis unfolding idd rule n
    by auto
qed

lemma rm_set_lookup2: assumes "rule \<in> conv_ta_rule (ta_epss_impl TA) ` (ta_rules TA')"
  shows "rule \<in> set (ta_rm_set_lookup (rsym_impl rule))"
  using assms unfolding rm_set_lookup
  by (cases rule, auto)

lemma rm_set_lookup3: assumes "rule \<in> conv_ta_rule (ta_epss_impl TA) ` (ta_rules TA')"
  shows "\<exists> rules. Some rules = rm_\<alpha> (ta_rules_impl TA) (rsym_impl rule) \<and> rule \<in> set rules"
  using rm_set_lookup2[OF assms] by (auto simp: rm_set_lookup_def)


lemma ta_res_impl: 
  shows "rs_\<alpha> (ta_res_impl (ta_rules_impl TA) (ta_epss_impl TA) t) = ta_res TA' t"
proof (induct t)
  case (Var x) thus ?case by (simp add: ta_epss)
next
  case (Fun f ts)
  show ?case
  proof (rule set_eqI)
    fix q
    show "(q \<in> rs_\<alpha> (ta_res_impl (ta_rules_impl TA) (ta_epss_impl TA) (Fun f ts))) = (q \<in> ta_res (ta_of TA) (Fun f ts))"
      (is "(q \<in> ?L) = (q \<in> ?S)")
    proof
      assume "q \<in> ?L"
      then obtain qs' rule where mem: "rule \<in> set (ta_rm_set_lookup (f,length ts))"
        and q: "rqss_impl rule = qs'"
        and q': "q \<in> rs_\<alpha> qs'"
        and args: "\<forall> i \<in> {0 ..< length ts}.
             rs_memb (rqs_impl rule ! i) (map (ta_res_impl (ta_rules_impl TA) (ta_epss_impl TA)) ts ! i)"
        by (auto simp: Let_def list_all_iff rs_correct)
      from rm_set_lookup1[OF mem] 
      obtain qs qs'' q' where id: "rule = TA_rule_impl f qs q' qs''" and
        len: "length ts = length qs" and 
        qs'': "rs_\<alpha> qs'' = {q. (q',q) \<in> (ta_eps TA')^*}" and
        rule: "f qs \<rightarrow> q' \<in> ta_rules TA'" by auto
      from q id have qs': "qs' = qs''" by auto
      from q'[unfolded qs' qs''] have qq': "(q',q) \<in> (ta_eps TA')^*" by simp
      show "q \<in> ?S"
      proof (unfold ta_res.simps, rule, rule exI[of _ q], rule exI[of _ q'], rule exI[of _ qs], intro conjI, simp, simp add: rule, rule qq', simp add: len, intro allI impI)
        fix i
        assume i: "i < length ts"
        hence mem: "ts ! i \<in> set ts" by auto
        from args[unfolded id, simplified] i len
        have "rs_memb (qs ! i) (ta_res_impl (ta_rules_impl TA) (ta_epss_impl TA) (ts ! i))" by auto
        with Fun[OF mem] i 
        show "qs ! i \<in> map (ta_res TA') ts ! i" by (simp add: rs_correct)
      qed
    next
      assume "q \<in> ?S"
      then obtain q' qs where rule: "f qs \<rightarrow> q' \<in> ta_rules TA'"
        and len: "length qs = length ts"
        and args: "\<forall> i < length ts. qs ! i \<in> map (ta_res TA') ts ! i"
        and q: "(q',q) \<in> (ta_eps TA')^*"
        by auto
      let ?rule = "TA_rule_impl f qs q' (ta_epss_impl TA q')"
      have rrule: "?rule = conv_ta_rule (ta_epss_impl TA) (f qs \<rightarrow> q')" by auto
      have rrule: "?rule \<in> conv_ta_rule (ta_epss_impl TA) ` ta_rules TA'"
        unfolding rrule using rule by blast
      have rule: "?rule \<in> set (ta_rm_set_lookup (f,length qs))"
        using rm_set_lookup3[OF rrule]
        unfolding rm_set_lookup_def by auto
      {
                fix i
        assume i: "i < length ts"
        hence mem: "ts ! i \<in> set ts" by simp
        have "qs ! i \<in> rs_\<alpha> (ta_res_impl (ta_rules_impl TA) (ta_epss_impl TA) (ts ! i))"
          unfolding Fun[OF mem] using args i by auto
      } note args = this
      show "q \<in> ?L"
        by (simp add: Let_def list_all_iff rs_correct, intro exI conjI, rule rule[unfolded len], insert args q, simp add: len, simp add: ta_epss)
    qed
  qed
qed
end



(* the following performs a brute force enumeration over
  all possible substitutions; this might be a bottle-neck and can be improved *)
definition ta_rule_compatible_list :: "('q :: linorder,'f :: linorder)ta_impl \<Rightarrow> ('f,'v)rule \<times> 'v list \<Rightarrow> (('f,'q)rule \<times> 'q) check"
where "ta_rule_compatible_list TA \<equiv> let rm = ta_rules_impl TA; eps = ta_epss_impl TA in (\<lambda> ((l,r),xs). 
          (check_allm (\<lambda> vec. let \<tau> = fun_of vec;
                               l\<tau> = map_vars \<tau> l;
                               r\<tau> = map_vars \<tau> r;
                               rhs = ta_res_impl rm eps r\<tau>
                            in  (check_allm (\<lambda> q. check (rs_memb q rhs) ((l\<tau>,r\<tau>),q)) (rs_to_list (ta_res_impl rm eps l\<tau>))))
                        (enum_vectors (ta_rqs_impl TA) xs)))"

context ta_inv
begin

(* I believe that also equivalence holds but was to lazy to prove it *)
lemma ta_rule_compatible_list: assumes wf: "vars_term r \<subseteq> vars_term l"
  and comp: "isOK(ta_rule_compatible_list TA ((l,r),vars_term_list l))"
  shows "ta_rule_compatible TA' (l,r)"
proof (cases "ta_rqs TA' = {}")
  case False
  show ?thesis
    unfolding ta_rule_compatible_def 
  proof(clarify)
    fix \<tau> q
    assume q: "q \<in> ta_res TA' (map_vars \<tau> l)"
      and \<tau>: "\<tau> ` vars_term l \<subseteq> ta_rqs TA'"
    let ?Q = "ta_rqs_impl TA"
    from False have "?Q \<noteq> []" using ta_rqs_impl by auto
    from enum_vectors_complete[OF this, of "vars_term_list l"]
    obtain vec where vec: "vec \<in> set (enum_vectors ?Q (vars_term_list l))" and \<tau>2: "\<And> x c. x \<in> set (vars_term_list l) \<Longrightarrow> c \<in> set ?Q \<Longrightarrow>  \<tau> x = c \<longrightarrow> fun_of vec x = c" by blast
    let ?\<tau> = "fun_of vec"
    {
      fix x
      assume x: "x \<in> vars_term l"
      with \<tau> have "\<tau> x \<in> ta_rqs TA'" by auto    
      from \<tau>2[unfolded vars_term_list_sound ta_rqs_impl, OF x this] have "\<tau> x = ?\<tau> x"
        by simp
    } note \<tau>2 = this
    have l: "map_vars \<tau> l = map_vars ?\<tau> l"
      by (rule map_vars_vars_term[OF \<tau>2])
    have r: "map_vars \<tau> r = map_vars ?\<tau> r"
      by (rule map_vars_vars_term, insert \<tau>2 wf, auto)
    from comp[unfolded ta_rule_compatible_list_def Let_def, simplified, THEN bspec[OF _ vec]] q 
    show "q \<in> ta_res TA' (map_vars \<tau> r)" unfolding l r by (auto simp: rs_correct ta_res_impl)
  qed
next
  case True
  show "ta_rule_compatible TA' (l,r)"
    unfolding ta_rule_compatible_def
  proof (clarify)
    fix \<tau> q
    assume q: "q \<in> ta_res TA' (map_vars \<tau> l)"
      and \<tau>: "\<tau> ` vars_term l \<subseteq> ta_rqs TA'"    
    show "q \<in> ta_res TA' (map_vars \<tau> r)"
    proof (cases l)
      case (Fun f ls)
      from q[unfolded Fun] True
      show ?thesis unfolding ta_rqs_def by auto
    next
      case (Var x)
      from \<tau>[unfolded Var ta_rqs_def]  True
      show ?thesis unfolding ta_rqs_def by simp
    qed
  qed    
qed
end

(* TODO: perhaps integrate in ta_match and here an optimization of concat / fun_merge that deals with 
   non-linear terms. Currently conflicts might occur (x / a and x / b) and duplicates may occur (x / a 
   and x / a). This is only important for efficiency, it will not effect power *)
declare size_simps[simp]
fun ta_match_pre_impl :: "('f \<times> nat,('q :: linorder,'f :: linorder)ta_rule_impl list)rm \<Rightarrow> ('q \<Rightarrow> 'q rs) \<Rightarrow> ('f,'v)term \<Rightarrow> 'q \<Rightarrow> ('v \<times> 'q)list set"
  where "ta_match_pre_impl TA eps (Var x) q = { [(x,q')] | q'. q' \<in> rs_\<alpha> (eps q)}"
   | "ta_match_pre_impl TA eps (Fun f ts) q = { concat \<sigma>s | \<sigma>s qs' qs q'. 
          (TA_rule_impl f qs q' qs') \<in> set (rm_set_lookup TA (f,length ts)) \<and> 
          q' \<in> rs_\<alpha> (eps q) \<and> 
          length \<sigma>s = length ts \<and>
          (\<forall> i < length ts. \<sigma>s ! i \<in> ta_match_pre_impl TA eps (ts ! i) (qs ! i)) }"
declare size_simps[simp del]

context ta_inv
begin
lemma ta_match_pre_impl': "\<sigma> \<in> ta_match TA' t q \<Longrightarrow> \<exists> \<tau> \<in> ta_match_pre_impl (ta_rules_impl TA) (ta_epsrs_impl TA) t q. (set (map fst \<tau>) = vars_term t \<and> (\<forall> x \<in> vars_term t. fun_of \<tau> x = \<sigma> x))"
proof (induct t arbitrary: q \<sigma>)
  case (Var x q)
  thus ?case  unfolding fun_of_def using ta_epsrs
    by auto
next
  case (Fun f ts q \<sigma>)
  let ?I = "ta_match_pre_impl (ta_rules_impl TA) (ta_epsrs_impl TA)"
  from Fun(2) obtain \<sigma>s qs q' where rule: "f qs \<rightarrow> q' \<in> ta_rules TA'"
    and q'q: "(q',q) \<in> (ta_eps TA')^*"
    and \<sigma>: "\<sigma> = fun_merge \<sigma>s (map vars_term ts)"
    and len1: "length qs = length ts"
    and len2: "length \<sigma>s = length ts" 
    and rec: "\<And> i. i < length ts \<Longrightarrow> \<sigma>s ! i \<in> ta_match TA' (ts ! i) (qs ! i)" by auto
  let ?prop = "\<lambda> \<tau> i. \<tau> \<in> ?I (ts ! i) (qs ! i) \<and> set (map fst \<tau>) = vars_term (ts ! i) \<and> (\<forall> x \<in> vars_term (ts ! i). fun_of \<tau> x = (\<sigma>s ! i) x)"
  {
    fix i      
    assume i: "i < length ts"
    let ?t = "ts ! i"
    from i have t: "?t \<in> set ts" by auto
    from Fun(1)[OF t rec[OF i]] have "\<exists> \<tau>. ?prop \<tau> i" by auto
  }
  hence "\<forall> i. \<exists> \<tau>. i < length ts \<longrightarrow> ?prop \<tau> i" by auto
  from choice[OF this] obtain \<tau>s where \<tau>s: "\<And> i. i < length ts \<Longrightarrow> ?prop (\<tau>s i) i" by auto
  obtain \<tau> where \<tau>: "\<tau> = concat (map (\<lambda> i. \<tau>s i) [0 ..< length ts])" by auto
  have vars_term: "set (map fst \<tau>) = vars_term (Fun f ts)" (is "?L = ?R")
  proof (unfold set_eq_iff, rule)
    fix x
    {
      assume "x \<in> ?L"
      from this[unfolded \<tau>] obtain i where i: "i < length ts" and x: "x \<in> set (map fst (\<tau>s i))" by force
      from \<tau>s[OF i] x have x: "x \<in> vars_term (ts ! i)" by auto
      hence "x \<in> ?R" using i 
        by (auto simp: set_conv_nth[of ts])
    }
    moreover
    {
      assume "x \<in> ?R"
      then obtain i where i: "i < length ts" and x: "x \<in> vars_term (ts ! i)"
        by (auto simp: set_conv_nth[of ts])
      from \<tau>s[OF i] x have x: "x \<in> set (map fst (\<tau>s i))" by force
      with i have "x \<in> ?L" unfolding \<tau> by force
    }
    ultimately show "(x \<in> ?L) = (x \<in> ?R)" by auto
  qed
  show ?case
  proof(intro bexI[of _ \<tau>] conjI, rule vars_term, intro ballI)
    show "\<tau> \<in> ?I (Fun f ts) q"
    proof (unfold ta_match_pre_impl.simps, rule, unfold ta_epsrs, intro exI conjI allI impI, rule \<tau>)
      fix i
      assume i: "i < length ts"
      show "map \<tau>s [0 ..< length ts] ! i \<in> ?I (ts ! i) (qs ! i)"
        using \<tau>s[OF i, THEN conjunct1] i by auto
    next
      show "q' \<in> {q'. (q',q) \<in> (ta_eps TA')^*}" using q'q by simp
    next
      show "length (map \<tau>s [0..< length ts]) = length ts" by simp
    next
      let ?e = "ta_epss_impl TA q'"
      show "TA_rule_impl f qs q' ?e \<in> set (ta_rm_set_lookup (f, length ts))"
      proof (unfold rm_set_lookup, rule, intro conjI)
        show "rsym_impl (TA_rule_impl f qs q' ?e) = (f, length ts)"
          by (simp add: len1)
      next
        show "TA_rule_impl f qs q' ?e \<in> conv_ta_rule (ta_epss_impl TA) ` ta_rules TA'"
          by (rule, insert rule, auto)
      qed
    qed
  next
    fix x
    assume "x \<in> ?R"
    then obtain i where i: "i < length ts" and x: "x \<in> vars_term (ts ! i)"
      by (auto simp: set_conv_nth[of ts])
    show "fun_of \<tau> x = \<sigma> x" unfolding \<sigma> \<tau>
    proof (rule fun_of_concat_merge)
      fix i
      assume i: "i < length ts" and x: "x \<in> vars_term (ts ! i)"
      show "fun_of (\<tau>s i) x = (\<sigma>s ! i) x"
        using \<tau>s[OF i] x i by auto
    next
      show "length ts = length \<sigma>s" using len1 len2 by auto
    next
      fix i
      assume i: "i < length ts"
      show "(lookup x (\<tau>s i) \<noteq> None) = (x \<in> vars_term (ts ! i))"
      proof -
        have "(x \<in> vars_term (ts ! i)) = (x \<in> set (map fst (\<tau>s i)))" (is "?l = ?m")
          using \<tau>s[OF i]  by simp
        also have "... = (lookup x (\<tau>s i) \<noteq> None)" (is "_ = ?r")
        proof
          assume ?m
          then obtain y where "(x,y) \<in> set (\<tau>s i)" by auto
          with lookup_complete[of x "\<tau>s i"] show ?r by auto
        next
          assume ?r
          then obtain y where "lookup x (\<tau>s i) = Some y" by (cases "lookup x (\<tau>s i)", auto)
          from lookup_sound[OF this] show ?m by force
        qed
        finally show "?r = ?l" by simp
      qed
    qed (insert i x, auto)
  qed
qed

lemma ta_match_pre_impl: "\<sigma> \<in> ta_match TA' t q \<Longrightarrow> \<exists> \<tau> \<in> ta_match_pre_impl (ta_rules_impl TA) (ta_epsrs_impl TA) t q. \<forall> x \<in> vars_term t. fun_of \<tau> x = \<sigma> x"
  using ta_match_pre_impl' by blast
end

declare size_simps[simp]
fun ta_match_impl :: "('f \<times> nat,('q :: linorder,'f :: linorder)ta_rule_impl list)rm \<Rightarrow> ('q \<Rightarrow> 'q rs) \<Rightarrow> ('f,'v)term \<Rightarrow> 'q \<Rightarrow> ('v :: linorder \<times> 'q)list rs"
  where "ta_match_impl TA eps (Var x) q = list_to_rs (map (\<lambda>q'. [(x,q')]) (rs_to_list (eps q)))" 
   | "ta_match_impl TA eps (Fun f ts) q = (let 
          n = length ts;
          rules = rm_set_lookup TA (f, n);
          ep = eps q;
          f = (\<lambda> rule. list_to_rs (case rule of TA_rule_impl _ qs q' qs' \<Rightarrow>
                 (if rs_memb q' ep then 
                   (let rec = map (\<lambda> (tsi,qsi). rs_to_list (ta_match_impl TA eps tsi qsi)) (zip ts qs)
                     in map concat (concat_lists rec))
                 else [])))
        in rs_Union (map f rules))" 
declare size_simps[simp del]

context ta_inv
begin
lemma ta_match_impl_conv_pre: "rs_\<alpha> (ta_match_impl (ta_rules_impl TA) eps t q) = ta_match_pre_impl (ta_rules_impl TA) eps t q"
proof (induct t arbitrary: q)
  case (Var x)
  thus ?case by (auto simp: rs_correct)
next
  case (Fun f ts q)
  let ?TA = "ta_rules_impl TA"
  show ?case (is "?I = ?P")
  proof -
    {
      fix \<sigma>
      assume "\<sigma> \<in> ?P"
      from this
      obtain \<sigma>s qs qs' q' where \<sigma>: "\<sigma> = concat \<sigma>s"
        and rule: "TA_rule_impl f qs q' qs' \<in> set (ta_rm_set_lookup (f, length ts))"
        and q'q: "q' \<in> rs_\<alpha> (eps q)"
        and len1: "length \<sigma>s = length ts"
        and rec: "\<And> i. i < length ts \<Longrightarrow> \<sigma>s ! i \<in> ta_match_pre_impl ?TA eps (ts ! i) (qs ! i)" by auto
      from rm_set_lookup1[OF rule] have len2: "length qs = length ts" by auto
      from q'q have q'q: "rs_memb q' (eps q)" by (simp add: rs_correct)
      have " \<exists>a\<in>set (ta_rm_set_lookup (f, length ts)).
       \<exists>uu q' qs.
          (\<exists> f. TA_rule_impl f qs q' uu = a) \<and>
          rs_memb q' (eps q) \<and>
          concat \<sigma>s
          \<in> concat `
            {as. length as = min (length ts) (length qs) \<and>
                 (\<forall>i. i < length ts \<and> i < length qs \<longrightarrow>
                      as ! i \<in> rs_\<alpha> (ta_match_impl ?TA eps (ts ! i) (qs ! i)))} "
      proof (rule bexI[OF _ rule], intro exI conjI, rule refl, rule q'q, rule, rule refl,
        unfold mem_def[of \<sigma>s] Collect_def len1 len2, intro conjI allI impI allI, simp)
        fix i
        assume "i < length ts \<and> i < length ts" 
        hence i: "i < length ts" by auto
        hence "ts ! i \<in> set ts" by auto
        from Fun[OF this] rec[OF i] show "\<sigma>s ! i \<in> rs_\<alpha> (ta_match_impl ?TA eps (ts ! i) (qs ! i))" by auto
      qed
(*
        show "\<forall> \<sigma>s \<in> set (map (\<lambda>(x,y).rs_to_list (ta_match_impl ?TA eps x y)) (zip ts qs)). \<sigma>s \<noteq> []"
        proof
          fix \<sigma>s'
          assume "\<sigma>s' \<in> set (map (\<lambda>(x,y).rs_to_list (ta_match_impl ?TA eps x y)) (zip ts qs))"
          from this[unfolded set_conv_nth] obtain i where 
            "\<sigma>s' = map (\<lambda>(x,y). rs_to_list (ta_match_impl ?TA eps x y)) (zip ts qs) ! i"
            and i: "i < length ts" by auto          
          hence \<sigma>s': "\<sigma>s' = rs_to_list (ta_match_impl ?TA eps (ts ! i) (qs ! i))"
            using len2 by auto
          from i have tsi: "ts ! i \<in> set ts" by auto
          show "\<sigma>s' \<noteq> []" 
          proof 
            assume "\<sigma>s' = []"
            hence "set \<sigma>s' = {}" by simp
            thus False
              unfolding \<sigma>s' using  rec[OF i, unfolded Fun[OF tsi, symmetric]]
              by (simp add: rs_correct)
          qed
        qed 
      qed *)
      hence "\<sigma> \<in> ?I" unfolding ta_match_impl.simps Let_def set_concat rs_correct \<sigma>
        by (simp add: rs_correct)
    }
    moreover
    {
      fix \<sigma>
      assume "\<sigma> \<in> ?I"
      from this
      obtain rule g \<sigma>s qs' qs q' where
         rule: "TA_rule_impl g qs q' qs' \<in> set (ta_rm_set_lookup (f, length ts))"
        and q': "rs_memb q' (eps q)"
        and \<sigma>: "\<sigma> \<in> concat ` {\<sigma>s. length \<sigma>s = min (length ts) (length qs) \<and> (\<forall> i. i < length ts \<and> i < length qs \<longrightarrow> \<sigma>s ! i \<in> rs_\<alpha> (ta_match_impl ?TA eps (ts ! i) (qs ! i)))}" by (simp add: rs_correct) blast+
      from rm_set_lookup1[OF rule] have gf: "g = f" and len2: "length qs = length ts" by auto
      from \<sigma> obtain \<sigma>s where \<sigma>: "\<sigma> = concat \<sigma>s" and len1: "length \<sigma>s = length ts" and rec: "\<And> i. i < length ts \<Longrightarrow> \<sigma>s ! i \<in> rs_\<alpha> (ta_match_impl ?TA eps (ts ! i) (qs ! i))" unfolding len2 by auto
      have "\<sigma> \<in> ?P"
      proof (simp, intro exI conjI, rule \<sigma>, rule rule[unfolded gf], rule q'[simplified rs_correct], rule len1,
        intro allI impI)
        fix i
        assume i: "i < length ts" 
        hence "ts ! i \<in> set ts" by auto
        from Fun[OF this] rec[OF i] show "\<sigma>s ! i \<in> ta_match_pre_impl ?TA eps (ts ! i) (qs ! i)" by auto
      qed
    }
    ultimately show ?thesis by blast
  qed
qed        

lemma ta_match_impl: "\<sigma> \<in> ta_match TA' t q \<Longrightarrow> \<exists> \<tau> \<in> rs_\<alpha> (ta_match_impl (ta_rules_impl TA) (ta_epsrs_impl TA) t q). \<forall> x \<in> vars_term t. fun_of \<tau> x = \<sigma> x"
  using ta_match_pre_impl unfolding ta_match_impl_conv_pre .
end

declare ta_match_impl.simps[simp del]

fun ta_rule_compatible_eff_list :: "('q :: linorder,'f :: linorder)ta_impl \<Rightarrow> ('f,'v :: linorder)rule \<Rightarrow> (('f,'q)rule \<times> 'q) check"
where "ta_rule_compatible_eff_list TA (Fun f ls,r) =  
           (let rm = ta_rules_impl TA; 
                eps = ta_epss_impl TA;
                eps' = ta_epsrs_impl TA;
                n = length ls;
                ta_res' = ta_res_impl rm eps
             in check_allm (\<lambda> rule. case rule of  TA_rule_impl f qs q qs' \<Rightarrow>
                   check_allm (\<lambda> \<sigma>. 
                       let \<sigma>' = fun_of \<sigma>;
                           l\<sigma> = map_vars \<sigma>' (Fun f ls);
                           r\<sigma> = map_vars \<sigma>' r;
                           qsl = ta_res' l\<sigma>;
                           qsr = ta_res' r\<sigma>
                         in (case rs_subset qsl qsr of None \<Rightarrow> OK | Some q' \<Rightarrow> error ((l\<sigma>,r\<sigma>),q'))
                   ) (rs_to_list (ta_match_impl rm eps' (Fun f ls) q)))
                 (rm_set_lookup rm (f,n)))" |
    "ta_rule_compatible_eff_list TA (Var x,r) = 
           (let rm = ta_rules_impl TA; 
                eps = ta_epss_impl TA;
                ta_res' = ta_res_impl rm eps
             in 
                check_allm (\<lambda> q. 
                  (let r\<sigma> = map_vars (\<lambda> _. q) r
                    in check (rs_memb q (ta_res' r\<sigma>)) ((Var q, r\<sigma>),q)))
                (ta_rqs_impl TA))"

lemma (in ta_inv) ta_rule_compatible_eff_list: assumes ok: "isOK(ta_rule_compatible_eff_list TA (l,r))"
  and vars: "vars_term r \<subseteq> vars_term l"
  shows "ta_rule_compatible TA' (l,r)"
proof (rule ta_rule_compatible_eff[OF vars])
  show "ta_rule_compatible_eff TA' (l, r)" 
  proof (cases l)
    case (Fun f ls)
    show ?thesis unfolding Fun ta_rule_compatible_eff.simps
    proof (intro allI impI, clarify)
      fix \<sigma> qs q
      assume rule: "f qs \<rightarrow> q \<in> ta_rules TA'" and 
        len: "length qs = length ls" and 
        match: "\<sigma> \<in> ta_match TA' (Fun f ls) q" and
        q: "q \<in> ta_res TA' (map_vars \<sigma> (Fun f ls))"
      from ta_match_impl[OF match]
      obtain \<tau> where \<tau>1: "\<tau> \<in> rs_\<alpha> (ta_match_impl (ta_rules_impl TA) (ta_epsrs_impl TA) l q)"
        and \<tau>2: "\<And> x. x \<in> vars_term l \<Longrightarrow> fun_of \<tau> x = \<sigma> x" unfolding Fun by blast
      let ?\<tau> = "fun_of \<tau>"
      have "map_vars \<sigma> (Fun f ls) = map_vars ?\<tau> l" unfolding Fun[symmetric]
        by (rule map_vars_vars_term, simp add: \<tau>2)
      from q[unfolded this] have q: "q \<in> ta_res TA' (map_vars ?\<tau> l)" by auto
      let ?e = "ta_epss_impl TA q"
      have rule: "TA_rule_impl f qs q ?e \<in> set (ta_rm_set_lookup (f,length ls))"
        unfolding rm_set_lookup
        by (rule, intro conjI, rule image_eqI[OF _ rule], simp_all add: len)
      let ?rl = "\<lambda> x. ta_res_impl (ta_rules_impl TA) (ta_epss_impl TA) (map_vars (fun_of x) l)"
      let ?rr = "\<lambda> x. ta_res_impl (ta_rules_impl TA) (ta_epss_impl TA) (map_vars (fun_of x) r)"
      from ok[unfolded Fun ta_rule_compatible_eff_list.simps Let_def] rule
      have " \<And>x. x\<in> rs_\<alpha> (ta_match_impl (ta_rules_impl TA) (ta_epsrs_impl TA) l q) \<Longrightarrow>
        isOK (case rs_subset (?rl x) (?rr x) of
           None \<Rightarrow> Inr () | Some q' \<Rightarrow> Inl ((map_vars (fun_of x) l, map_vars (fun_of x) r), q'))" 
        unfolding Fun by (auto simp: rs_correct)
      from this[OF \<tau>1[unfolded Fun[symmetric]]]
      have "rs_subset (?rl \<tau>) (?rr \<tau>) = None"
        by (cases "rs_subset (?rl \<tau>) (?rr \<tau>)", auto)
      from this[unfolded rs_subset ta_res_impl] q have q: "q \<in> ta_res TA' (map_vars ?\<tau> r)" by auto
      have "map_vars \<sigma> r = map_vars ?\<tau> r"
        by (rule map_vars_vars_term, insert vars \<tau>2, auto)
      thus "q \<in> ta_res TA' (map_vars \<sigma> r)" using q by simp
    qed
  next
    case (Var x)
    show ?thesis
      unfolding Var ta_rule_compatible_eff.simps
    proof (clarify)
      fix q 
      assume q: "q \<in> ta_rqs TA'"
      thus "q \<in> ta_res TA' (map_vars (\<lambda>_. q) r)"
        using ok[unfolded Var] by (auto simp: Let_def ta_rqs_impl rs_correct ta_res_impl)
    qed
  qed
qed


fun ta_res_impl_all :: "'q rs \<Rightarrow> ('f \<times> nat,('q :: linorder,'f :: linorder)ta_rule_impl list)rm \<Rightarrow> ('f,'v)term \<Rightarrow> 'q rs"
where "ta_res_impl_all Q TA (Var _) = Q"
  | "ta_res_impl_all Q TA (Fun f ts) = (
        let rec = map (ta_res_impl_all Q TA) ts;
            n = length ts;
            rules = rm_set_lookup TA (f,n);
            arules = filter (\<lambda> rule. let qs = rqs_impl rule in list_all (\<lambda> i. rs_memb (qs ! i) (rec ! i)) [0 ..< n]) rules;
            qs = map (\<lambda> rule. rqss_impl rule) arules   
           in rs_Union qs)"

context ta_inv
begin
lemma ta_res_impl_all: assumes "\<tau> ` vars_term t \<subseteq> ta_rqs TA'"
  shows "rs_\<alpha> (ta_res_impl (ta_rules_impl TA) (ta_epss_impl TA) (map_vars \<tau> t)) \<subseteq> rs_\<alpha> (ta_res_impl_all (ta_rqs_set TA) (ta_rules_impl TA) t)"
  using assms
proof (induct t)
  case (Var q)
  with assms  have q: "\<tau> q \<in> ta_rqs TA'" by auto
  { 
    fix q'
    assume q': "(\<tau> q, q') \<in> (ta_eps TA')^*"
    from q q' 
    have "q' \<in> ta_rqs TA'" unfolding ta_rqs_def
      by auto
  }
  thus ?case by (auto simp: ta_epss ta_rqs_set)
next
  case (Fun f ts)
  {
    fix rule 
    assume 
      args: "\<forall> i \<in> {0..<length ts}. rs_memb (rqs_impl rule ! i) (map (ta_res_impl (ta_rules_impl TA) (ta_epss_impl TA)) (map (map_vars \<tau>) ts) ! i)"
    have args: "\<forall> i \<in> {0..<length ts}. rs_memb (rqs_impl rule ! i)  (map (ta_res_impl_all (ta_rqs_set TA) (ta_rules_impl TA)) ts ! i)" (is "\<forall> i \<in> _. ?c i")
    proof -
      {
        fix i
        assume i: "i < length ts"
        from i have mem: "ts ! i \<in> set ts" by auto
        with Fun(2) have "\<tau> ` vars_term (ts ! i) \<subseteq> ta_rqs TA'" by auto
        from Fun(1)[OF mem this] args i
        have "?c i" by (auto simp: rs_correct)
      }
      thus ?thesis by auto
    qed
  }
  thus ?case by (auto simp: Let_def rs_correct list_all_iff)
qed
end

definition ta_rule_compatible_heuristic :: "('q :: linorder,'f :: linorder)ta_impl \<Rightarrow> ('f,'v :: linorder)term \<Rightarrow> bool" where
"ta_rule_compatible_heuristic TA l \<equiv> rs_isEmpty (ta_res_impl_all (ta_rqs_set TA) (ta_rules_impl TA) l)"

context ta_inv
begin

lemma ta_rule_compatible_heuristic_subteq: assumes compat: "ta_rule_compatible_heuristic TA u"
    and subt: "l \<unrhd> u"
  shows "ta_rule_compatible TA' (l,r)"
proof -
  {
    fix \<tau>
    assume \<tau>: "\<tau> ` vars_term l \<subseteq> ta_rqs TA'"
    {
      fix x
      assume "x \<in> vars_term u"
      from subt supteq_Var[OF this] have "l \<unrhd> Var x" by (rule supteq_trans)
      from var_subt_imp_subt_in_vars[OF this] have "x \<in> vars_term l" .
    }
    with \<tau> have \<tau>u: "\<tau> ` vars_term u \<subseteq> ta_rqs TA'" by auto
    from assms[unfolded ta_rule_compatible_heuristic_def]
    have empty: "rs_\<alpha> (ta_res_impl_all (ta_rqs_set TA) (ta_rules_impl TA) u) = {}" by (simp add: rs_correct)
    from ta_res_impl_all[OF \<tau>u] empty
    have empty: "ta_res TA' (map_vars \<tau> u) = {}" 
      unfolding ta_res_impl by auto
    from subt obtain C where l: "l = C\<langle>u\<rangle>" ..
    have "ta_res TA' (map_vars \<tau> l) = {}"
      unfolding l
    proof (induct C)
      case (More f bef C aft)
      { 
        fix qs
        have "\<And> g. \<exists> i < Suc (length bef + length aft). qs ! i \<notin> (map g bef @ ta_res TA' (map_vars \<tau> C\<langle>u\<rangle>) # map g aft) ! i"
          by (rule exI[of _ "length bef"], simp add: nth_append, insert More, auto)
      }
      thus ?case by (auto simp: o_def)
    qed (insert empty, auto)
  }
  thus ?thesis
    unfolding ta_rule_compatible_def by auto
qed
end


definition ta_trs_compatible_list :: "('q :: linorder,'f :: linorder)ta_impl \<Rightarrow> (('f,'v)rule \<times> 'v list)list \<Rightarrow> (('f,'v)rule \<times> ('f,'q)rule \<times> 'q) check"
where "ta_trs_compatible_list TA R \<equiv> let check = ta_rule_compatible_list TA in
     check_allm (\<lambda> lr. check lr <+? (\<lambda> lq. (fst lr,lq))) R"

lemma (in ta_inv) ta_trs_compatible_list: assumes wf_trs: "\<And> l r. (l,r) \<in> fst ` set R \<Longrightarrow> vars_term r \<subseteq> vars_term l"
  and vars: "\<And> l r xs. ((l,r),xs) \<in> set R \<Longrightarrow> xs = vars_term_list l"
  and comp: "isOK(ta_trs_compatible_list TA R)"
  shows "ta_trs_compatible TA' (fst ` set R)" 
proof -
  {
    fix l r xs
    assume rule: "((l,r),xs) \<in> set R"
    from comp[unfolded ta_trs_compatible_list_def Let_def list_all_iff]
      rule have comp: "isOK(ta_rule_compatible_list TA ((l,r),xs))"
      by auto
    from ta_rule_compatible_list[OF wf_trs comp[unfolded vars[OF rule]]] rule
    have "ta_rule_compatible TA' (l,r)" by force
  }
  thus ?thesis unfolding ta_trs_compatible_def 
    by force
qed

definition ta_trs_compatible_eff_list :: "('q :: linorder,'f :: linorder)ta_impl \<Rightarrow> ('f,'v :: linorder)rule list \<Rightarrow> (('f,'v)rule \<times> ('f,'q)rule \<times> 'q) check"
where "ta_trs_compatible_eff_list TA R \<equiv> let check = ta_rule_compatible_eff_list TA in
     check_allm (\<lambda> lr. check lr <+? (\<lambda> lq. (lr,lq))) R"

lemma (in ta_inv) ta_trs_compatible_eff_list: assumes wf_trs: "\<And> l r. (l,r) \<in> set R \<Longrightarrow> vars_term r \<subseteq> vars_term l"
  and comp: "isOK(ta_trs_compatible_eff_list TA R)"
  shows "ta_trs_compatible TA' (set R)" 
proof -
  {
    fix l r 
    assume rule: "(l,r) \<in> set R"
    from comp[unfolded ta_trs_compatible_eff_list_def Let_def list_all_iff]
      rule have comp: "isOK(ta_rule_compatible_eff_list TA (l,r))"
      by auto
    from ta_rule_compatible_eff_list[OF comp wf_trs[OF rule]]
    have "ta_rule_compatible TA' (l,r)" .
  }
  thus ?thesis unfolding ta_trs_compatible_def 
    by force
qed


definition ta_contains_impl :: "('f \<times> nat) list \<Rightarrow> ('q :: linorder,'f :: linorder)ta_impl \<Rightarrow> 'q list \<Rightarrow> ('q + ('f,'q)term)check"
  where "ta_contains_impl F TA qs \<equiv> 
       do { 
         let fin = ta_final_impl TA;
         let look = rm_set_lookup (ta_rules_impl TA);
         let eps = ta_epss_impl TA;
         check_allm (\<lambda> q. check (rs_memb q fin) q) qs
           <+? Inl;
         check_allm (\<lambda> (f,n). let rules = look (f,n) 
              in (check_all (\<lambda> qs'. list_ex (\<lambda> rule. rqs_impl rule = qs' \<and> (let qq = rqss_impl rule in list_ex (\<lambda>q. rs_memb q qq) qs)) rules)
          (concat_lists (replicate n qs))) <+? (\<lambda>qs'. Fun f (map Var qs'))) F
            <+? Inr
          } "

context ta_inv
begin
(* I believe that also the other direction is correct, but currently this
   is not needed *)
lemma ta_contains_impl: 
  assumes contain: "isOK(ta_contains_impl F TA qs)"
  shows "ta_contains (set F) TA' (set qs)"
proof -
  note contain = contain[unfolded ta_contains_impl_def Let_def list_ex_iff concat_lists, simplified, simplified rs_correct, simplified]
  have final: "set qs \<subseteq> ta_final TA'" using contain[THEN conjunct1]
    unfolding ta_of_def by (auto simp: rs_correct)
  show ?thesis unfolding ta_contains_def
  proof (rule conjI[OF final], intro allI impI)
    fix f qs'
    assume "(f,length qs') \<in> set F \<and> set qs' \<subseteq> set qs"
    hence f: "(f,length qs') \<in> set F" and i: "\<forall> i < length qs'. qs' ! i \<in> set qs" unfolding set_conv_nth[of qs'] by auto
    from contain[THEN conjunct2, THEN bspec[OF _ f], simplified, THEN spec[of _ qs']] i obtain rule q where q: "q \<in> set qs" and rule: "rule \<in> set (ta_rm_set_lookup (f, length qs'))" 
      and q2: "q \<in> rs_\<alpha> (rqss_impl rule)" and qs': "rqs_impl rule = qs'" by auto
    from rm_set_lookup1[OF rule] 
    obtain qsr1 qr qsr2 where rule: "rule = TA_rule_impl f qsr1 qr qsr2"
      and len: "length qs' = length qsr1"
      and eps: "rs_\<alpha> qsr2 = {q'. (qr,q') \<in> (ta_eps TA')^*}"
      and rrule: "f qsr1 \<rightarrow> qr \<in> ta_rules TA'" by auto
    from rule qs' have qs': "qs' = qsr1" by auto
    show "\<exists> q q'. f qs' \<rightarrow> q \<in> ta_rules TA' \<and> q' \<in> set qs \<and> (q,q') \<in> (ta_eps TA')^*"
      by (intro exI conjI, unfold qs', rule rrule, rule q, insert q2[unfolded rule] eps, auto)
  qed
qed
end

datatype ('q,'f)tree_automaton = Tree_Automaton "'q list" "('q,'f)ta_rule list" "('q \<times> 'q)list"

fun generate_ta :: "('q :: linorder,'f :: linorder)tree_automaton \<Rightarrow> ('q,'f)ta_impl"
  where "generate_ta (Tree_Automaton fin rules eps) = (let ep = memo_rbt_rtrancl eps;
     epr = memo_rbt_rtrancl (map (\<lambda> (q,q'). (q',q)) eps);
     rqs_rs = rs_Union (map (\<lambda> rule. (ep (rq rule))) rules);
     rrules = map (conv_ta_rule ep) rules
     in
       \<lparr> ta_final_impl = list_to_rs fin, 
         ta_rules_impl = elem_list_to_rm rsym_impl rrules,
         ta_rqs_impl = rs_to_list rqs_rs,
         ta_rqs_set = rqs_rs,
         ta_eps_impl = eps,
         ta_epss_impl = ep,
         ta_epsrs_impl = epr
          \<rparr>)"

lemma generate_ta_rules: "ta_rules (ta_of (generate_ta (Tree_Automaton fin rules eps))) = set rules"
  (is "ta_rules ?TA' = _")
proof -
  {
    fix eps
    let ?rules = "\<Union> { ta_rule_conv ` {d \<in> set (map (conv_ta_rule eps) rules). rsym_impl d = fn} | fn. True}"
    {
      fix r
      assume "r \<in> ?rules"
      from this[simplified]
      obtain rl rs where r: "r = ta_rule_conv rl"
        and          rl: "rl = conv_ta_rule eps rs"
        and           rs: "rs \<in> set rules"
        by auto
      hence "r \<in> set rules" by (cases rs, auto)
    }
    moreover
    {
      fix r 
      assume "r \<in> set rules"
      then obtain f qs q where r: "r = (f qs \<rightarrow> q)" and rr: "f qs \<rightarrow> q \<in> set rules" by (cases r, auto)
      have "conv_ta_rule eps r = TA_rule_impl f qs q (eps q)" (is "_ = ?rl") unfolding r  by simp
      with rr have rl: "?rl \<in> set (map (conv_ta_rule eps) rules)" (is "_ \<in> ?L") by force
      have r: "r = ta_rule_conv ?rl" unfolding r by auto
      have "r \<in> ta_rule_conv ` ?L" unfolding r using rl by blast
      hence "r \<in> ?rules" by blast
    }
    ultimately have "?rules = set rules" by blast
  }
  thus ?thesis
    unfolding generate_ta.simps  Let_def
    unfolding ta_of_def ta.simps ta_impl.simps elem_list_to_rm.rm_set_lookup
    by auto
qed

lemma generate_ta: "ta_inv(generate_ta TA)"
proof(cases TA)
  case (Tree_Automaton fin rules epsilon)
  obtain eps where eps: "eps = memo_rbt_rtrancl epsilon" by auto
  obtain epsr where epsr: "epsr = memo_rbt_rtrancl (map (\<lambda>(q,q'). (q',q)) epsilon)" by auto
  let ?rqs = "rs_Union (map (\<lambda> rule. (eps (rq rule))) rules)"
  let ?rrules = "map (conv_ta_rule eps) rules"
  from Tree_Automaton eps epsr have A: "generate_ta TA = 
       \<lparr> ta_final_impl = list_to_rs fin, 
         ta_rules_impl = elem_list_to_rm rsym_impl ?rrules,
         ta_rqs_impl = rs_to_list ?rqs,
         ta_rqs_set = ?rqs,
         ta_eps_impl = epsilon,
         ta_epss_impl = eps,
         ta_epsrs_impl = epsr
          \<rparr>" (is "_ = ?TA")
    by (auto simp: Let_def)
  let ?TA' = "ta_of ?TA"
  let ?rules = "\<Union> { ta_rule_conv ` {d \<in> set (map (conv_ta_rule eps) rules). rsym_impl d = fn} | fn. True}"
  have ta_rules: "ta_rules ?TA' = set rules"
    using generate_ta_rules[of fin rules epsilon] 
    by (simp add: eps epsr Let_def)
  have qq1: "(\<lambda> (q,q'). (q,q') \<in> (set epsilon)^*) = (set epsilon)^*" unfolding mem_def by auto
  have qq2: "((\<lambda> (q,q'). (q',q)) ` set epsilon)^* = ((set epsilon)^*)^-1" unfolding rtrancl_converse[symmetric] 
    by (rule arg_cong[where f = "\<lambda> x. x^*"], auto)
  have one: "ta_rqs ?TA' = rs_\<alpha> (ta_rqs_set ?TA)" 
    unfolding ta_rqs_def ta_rules 
    unfolding ta_of_def ta.simps ta_impl.simps rs_Union
    by (simp add: eps memo_rbt_rtrancl qq1, auto)
  have two: "set (ta_rqs_impl ?TA) = ta_rqs ?TA'" unfolding one 
    by (auto simp: rs_correct Let_def) 
  show ?thesis unfolding A
    by (unfold_locales, unfold one two, unfold ta_impl.simps elem_list_to_rm.rm_set_lookup ta_rules, 
      auto simp: ta_of_def eps epsr memo_rbt_rtrancl qq1 qq2)
qed

definition tree_aut_trs_compat :: "('q :: {linorder,show},'f :: {linorder,show})tree_automaton \<Rightarrow> ('f,'q)rules \<Rightarrow> shows check"
  where "tree_aut_trs_compat ta R \<equiv> do {
     let TA = generate_ta ta;
     check_wf_trs R;
     check_left_linear_trs R;
     ta_trs_compatible_eff_list TA R
        <+? (\<lambda> (lr,lrq,q). shows ''TA is not compatible with R'' +@+ shows_nl 
             +@+ shows ''for rule '' +@+ shows_rule lr +@+ shows_nl
             +@+ shows ''which is instantiated by states to '' +@+ shows_rule lrq +@+ shows_nl
             +@+ shows ''the state '' +@+ shows q +@+ shows '' is only reachable from the lhs'' +@+ shows_nl)
 
  }"

fun ta_of_ta :: "('q,'f)tree_automaton \<Rightarrow> ('q,'f)ta"
  where "ta_of_ta (Tree_Automaton fin rules eps) = 
  \<lparr> ta_final = set fin,
    ta_rules = set rules,
    ta_eps   = set eps
  \<rparr>"

lemma tree_aut_trs_compat:
  assumes "isOK(tree_aut_trs_compat ta R)"
  shows "{t | s t. s \<in> ta_lang (ta_of_ta ta) \<and> (s,t) \<in> (rstep (set R))^*} \<subseteq> ta_lang (ta_of_ta ta)"
proof (cases ta)
  case (Tree_Automaton fin rules eps) note ta = this
  from assms[unfolded tree_aut_trs_compat_def Let_def]
  have wf: "wf_trs (set R)" and 
    ll: "left_linear_trs (set R)" and 
    compat: "isOK( ta_trs_compatible_eff_list (generate_ta ta) R)" by auto
  {
    fix l r
    assume "(l,r) \<in> set R"
    hence "vars_term r \<subseteq> vars_term l" using wf  unfolding wf_trs_def by auto
  } note var_cond = this
  show ?thesis unfolding ta ta_of_ta.simps
  proof (rule ta_trs_compatible[OF _ ll var_cond subset_refl])
    let ?TA = "\<lparr>ta_final = set fin, ta_rules = set rules, ta_eps = set eps\<rparr>"
    interpret ta_inv "generate_ta ta" by (rule generate_ta)
    let ?T = "ta_of (generate_ta ta)"    
    have one: "ta_rules ?T = set rules" unfolding ta using generate_ta_rules[of fin rules eps] by auto
    have two: "ta_final ?T = set fin" unfolding ta by (simp add: Let_def ta_of_def rs_correct)
    have three: "ta_eps ?T = set eps" unfolding ta by (simp add: Let_def ta_of_def)
    from one two three have id: "?T = ?TA" by auto
    show "ta_trs_compatible ?TA (set R)" 
      by (rule ta_trs_compatible_eff_list[OF var_cond compat, unfolded id])
  qed
qed

datatype ('q,'f)tree_aut_problem = tree_aut_problem "('q,'f)tree_automaton" "('f,'q)rules" "('q,'f)tree_automaton"

fun certify_tree_aut_problem :: "('q :: {linorder,show},'f :: {linorder,show})tree_aut_problem \<Rightarrow> shows check"
  where "certify_tree_aut_problem (tree_aut_problem A R B) =
    tree_aut_trs_compat B R"

end


